/* ========================================================================== */
/* === Demo/napheap_demo.c ================================================== */
/* ========================================================================== */

/* simple stand-alone demo program for napheap */

#include "napheap.h"
#include <stdlib.h>
#include <stdio.h>

#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define TRUE 1
#define FALSE 0

/* portable replacement for rand, for repeatable results */
static unsigned long next = 1 ;
#define MY_RAND_MAX 32767
static int myrand ( void )
{
    next = next * 1103515245 + 12345;
    return ((unsigned) (next / 65536) % (MY_RAND_MAX + 1)) ;
}


/* generate a random number between lower and upper */
static double rnum
(
    double lower,
    double upper
)
{
    double diff, t ;
    t = ((double) myrand ()) / MY_RAND_MAX ;
    diff = upper - lower ;
    t = t*diff + lower ;
    return (t) ;
}

static void print_vector (char *name, double *x, NAPHEAP_INT n)
{
    NAPHEAP_INT j ;
    printf ("\n%s:\n", name) ;
    for (j = 0; j < n; j++)
    {
        printf ("%s [%i] = %g\n", name, j, x [j]) ;
    }
}


int main (void)
{
    double s, t, cumlo, cumhi, l, u, lambda, err, err1,
           *x, *y, *d, *a, *lo, *hi, *Work ;
    NAPHEAP_INT j, n, *iWork ;
    int status, ok ;
    napheap_parameter Parm ;
    napheap_stats Stats ;

    printf ("\nnapheap demo.  Version %d.%d.%d, %s\n", NAPHEAP_MAIN_VERSION,
        NAPHEAP_SUB_VERSION, NAPHEAP_SUBSUB_VERSION, NAPHEAP_DATE) ;

    /* randomly generate a problem and solve with default parameter values */
    n = 20 ;
    x = (double *) malloc (n*sizeof (double)) ;
    d = (double *) malloc (n*sizeof (double)) ;
    a = (double *) malloc (n*sizeof (double)) ;
    y = (double *) malloc (n*sizeof (double)) ;
    lo = (double *) malloc (n*sizeof (double)) ;
    hi = (double *) malloc (n*sizeof (double)) ;
    cumlo = 0 ;
    cumhi = 0 ;
    for (j = 0; j < n; j++)
    {
        /* dj = random number on (5, 10) */
        d [j] = rnum (5, 10) ;
        /* yj = random number on (-10, 10) */
        y [j] = rnum (-10, 10) ;
        a [j] = rnum (-10, 10) ;
        lo [j] = -1 ;
        hi [j] = +1 ;
        t = a [j]*lo [j] ;
        s = a [j]*hi [j] ;
        /* cumlo = min a'x */
        cumlo += MIN (t, s) ;
        /* cumhi = max a'x */
        cumhi += MAX (t, s) ;
    }
    l = rnum (cumlo, cumhi) ;
    u = l + 10. ;

    printf ("minimize 0.5x'Dx - y'x subject to lo <= x <= hi and "
        "blo <= a'x <= bhi\n") ;

    printf ("blo = %g, bhi = %g\n", l, u) ;
    printf ("D is diagonal, with d = diag(D):\n") ;
    print_vector ("d", d, n) ;
    print_vector ("y", y, n) ;
    print_vector ("lo", lo, n) ;
    print_vector ("hi", hi, n) ;
    print_vector ("a", a, n) ;

    status = napheap (x, y, d, a, l, u, lo, hi, n,
                      &lambda, NULL, NULL, NULL, NULL) ;
    ok = (status == NAPHEAP_STATUS_OK) ;
    err = napheap_check (NULL, NULL, NULL, x, y, d, a, l, u, lo, hi, n, lambda);
    printf ("status: %d\n", status) ;
    printf ("The solution is:\n") ;
    print_vector ("x", x, n) ;

    printf ("relative error: %g\n", err) ;

    /* resolve the problem and print statistics, parameters, and status */
    napheap_defaults (&Parm) ;
    printf ("\n=============================================\n") ;
    printf ("Resolve the problem and print statistics,\n") ;
    printf ("parameters, and status:\n") ;
    status = napheap (x, y, d, a, l, u, lo, hi, n,
                      &lambda, &Stats, &Parm, NULL, NULL) ;
    napheap_print_parms (&Parm) ;
    napheap_print_stats (&Stats) ;
    ok = ok && (status == NAPHEAP_STATUS_OK) ;
    err1 = napheap_check (NULL, NULL, NULL, x, y, d, a, l, u, lo, hi, n,lambda);
    printf ("relative error: %g\n", err1) ;
    err = MAX (err, err1) ;

    /* resolve the problem using the previously computed multiplier lambda
       as the starting guess */
    printf ("\n=============================================\n") ;
    printf ("Resolve the problem using the previously computed\n") ;
    printf ("multiplier as the starting guess:\n") ;
    Parm.use_lambda = TRUE ;
    status = napheap (x, y, d, a, l, u, lo, hi, n,
                      &lambda, &Stats, &Parm, NULL, NULL) ;
    napheap_print_parms (&Parm) ;
    napheap_print_stats (&Stats) ;
    ok = ok && (status == NAPHEAP_STATUS_OK) ;
    err1 = napheap_check (NULL, NULL, NULL, x, y, d, a, l, u, lo, hi, n,lambda);
    printf ("relative error: %g\n", err1) ;
    err = MAX (err, err1) ;

    /* resolve the problem using the previously computed multiplier lambda
       as the starting guess and purely the break point searching algorithm */
    Parm.K = 0 ;
    printf ("\n=============================================\n") ;
    printf ("Resolve the problem using the previously computed\n") ;
    printf ("multiplier as the starting guess and purely the break point\n") ;
    printf ("searching algorithm. Note that the number of break\n") ;
    printf ("to reach the solution is 0:\n") ;
    status = napheap (x, y, d, a, l, u, lo, hi, n,
                      &lambda, &Stats, &Parm, NULL, NULL) ;
    napheap_print_parms (&Parm) ;
    napheap_print_stats (&Stats) ;
    ok = ok && (status == NAPHEAP_STATUS_OK) ;
    err1 = napheap_check (NULL, NULL, NULL, x, y, d, a, l, u, lo, hi, n,lambda);
    printf ("relative error: %g\n", err1) ;
    err = MAX (err, err1) ;

    /* set a diagonal element to 0 to generate an error message */
    printf ("\n=============================================\n") ;
    printf ("Set a diagonal element to 0 to generate an error message\n") ;
    d [4] = 0 ;
    status = napheap (x, y, d, a, l, u, lo, hi, n,
                      &lambda, &Stats, &Parm, NULL, NULL) ;
    napheap_print_stats (&Stats) ;
    ok = ok && (status != NAPHEAP_STATUS_OK) ;

    /* now set the diagonal to 0 and solve the problem using own storage */
    Work = (double *) malloc (5*n*sizeof (double)) ;
    iWork = (NAPHEAP_INT *) malloc ((3*n+3)*sizeof (NAPHEAP_INT)) ;
    Parm.use_lambda = FALSE ;
    printf ("\n=============================================\n") ;
    printf ("Now change the diagonal to be zero and resolve using own \n") ;
    printf ("storage arrays:\n") ;
    status = napheap (x, y, NULL, a, l, u, lo, hi, n,
                      &lambda, &Stats, &Parm, Work, iWork) ;
    napheap_print_parms (&Parm) ;
    napheap_print_stats (&Stats) ;
    ok = ok && (status == NAPHEAP_STATUS_OK) ;
    err1 = napheap_check (NULL, NULL, NULL, x, y, NULL, a, l, u, lo, hi,
                          n, lambda);
    printf ("relative error: %g\n", err1) ;
    err = MAX (err, err1) ;

    free (x) ;
    free (d) ;
    free (a) ;
    free (y) ;
    free (lo) ;
    free (hi) ;
    free (Work) ;
    free (iWork) ;

    fprintf (stderr, "worst-case relative error: %g\n", err) ;
    printf ("worst-case relative error: %g\n", err) ;
    if (ok && err < 1e-6)
    {
        fprintf (stderr, "All tests passed.\n") ;
        printf ("\nAll tests passed.\n") ;
    }
    else
    {
        fprintf (stderr, "TEST FAILURE.\n") ;
        printf ("\nTEST FAILURE.\n") ;
    }
    return (0) ;
}
