/* ========================================================================== */
/* === Include/napheap.h ==================================================== */
/* ========================================================================== */

/* include file for user programs that use napheap */

#ifndef NAPHEAP_H
#define NAPHEAP_H

#include <limits.h>
#include <float.h>
#include <math.h>
#include <stdlib.h>

/* -------------------------------------------------------------------------- */
/* NAPHEAP_INT: the default integer for napheap */
/* -------------------------------------------------------------------------- */

/* NAPHEAP_INT can be defined at compile time (e.g. -DNAPHEAP_INT=long).
   For a MATLAB mexFunction, NAPHEAP_INT is always mwSignedIndex, which is
   32 bits on a 32 bit MATLAB, and 64 bits on a 64 bit MATLAB.  Most versions
   of MATLAB as of 2013 are 64 bits.

   To change the integer definition, do not edit this file.  Edit the
   OPTFLAGS string in Lib/napheap.mk instead, and add -DNAPHEAP_INT=long
   (or whatever you wish).
*/

#ifdef MATLAB_MEX_FILE
#include "mex.h"
#define NAPHEAP_INT mwSignedIndex
#else
#ifndef NAPHEAP_INT
#define NAPHEAP_INT int
#endif
#endif

/* -------------------------------------------------------------------------- */
/* NAPHEAP_INF: the notion of infinity used by napheap */
/* -------------------------------------------------------------------------- */

/* ANSI C99 has a clean definition of IEEE infinity, defined in <math.h>.
   MATLAB has this as well.  With the ANSI (C90) version of C, there is no
   well-defined infinity, so DBL_MAX is used instead.

   You can override these defaults and define your own version of infinity
   with (for example):

   cc -ansi -DNAPHEAP_INF=1e200 napheap.c ...

   To do this, edit the OPTFLAGS string in Lib/napheap.mk (editting this
   file is not recommended).
*/

#ifndef NAPHEAP_INF
#ifdef INFINITY
/* ANSI C99 (gcc -std=c99) */
#define NAPHEAP_INF INFINITY
#else
/* ANSI C90 (gcc -ansi) */
#define NAPHEAP_INF DBL_MAX
#endif
#endif

/* -------------------------------------------------------------------------- */
/* user-controllable parameters */
/* -------------------------------------------------------------------------- */

typedef struct napheap_parameter_struct
{

     /* T => perform an iterative refinement step to improve solution accurary
        F => no refinement */
    int refine ;

    /* The solution is refined if |a'x - b| > err (only if refine is true) */
    double err ;

    /* T => use the Newton to update lambda when K > 0
       F => use variable fixing to update lambda when K > 0 */
    int newton ;

    /* T => use input lambda as a starting guess for multiplier
       F => ignore input lambda (final lambda is returned in lambda argument) */
    int use_lambda ;

    /* T => diagonal of objective Hessian is all positive, and the
            MATLAB expression all(d>0) is assumed to be true.
       F => there could be zeros on the diagonal of the Hessian */
    int d_is_pos ;

    /* T => diagonal of objective Hessian is zero, and the
            MATLAB expression all(d==0) is assumed to be true.
       F => there could be nonzeros on the diagonal of the Hessian
            Note: If d is NULL is passed in to napheap, then it is interpretted
            as all (d == 0) */
    int d_is_zero ;

    /* Number of variable fixing iterations or the number of Newton
       iterations.  These K iterations are then followed by the the break point
       searching algorithm.  K = 4 seems to perform well for a wide range of
       problems. */
    int K ;

} napheap_parameter ;

/* -------------------------------------------------------------------------- */
/* statistics returned to the user */
/* -------------------------------------------------------------------------- */

typedef struct napheap_stats_struct
{
    NAPHEAP_INT nfree ;     /* number of free variables in initial heap */
    NAPHEAP_INT nbound ;    /* number of bound variables in initial heap */
    NAPHEAP_INT nbrks ;     /* number break points to reach initial solution */
    NAPHEAP_INT nrefine ;   /* number of break points during refinement */
    NAPHEAP_INT nvarfix ;   /* number of variable fixing iterations */
    NAPHEAP_INT nnewton ;   /* number of Newton iterations performed */
    NAPHEAP_INT nsecant ;   /* number of secant steps performed */

    /* for error reporting */
    NAPHEAP_INT status ;    /* return value from napheap */
    NAPHEAP_INT jerror ;    /* invalid n, or index where error found */
    double lobad ;          /* bad value of lo, if any */
    double hibad ;          /* bad value of hi, if any */
    double dbad ;           /* bad value of d, if any */

} napheap_stats ;

/* -------------------------------------------------------------------------- */
/* status return value from napheap */
/* -------------------------------------------------------------------------- */

#define NAPHEAP_STATUS_OK                (0)
#define NAPHEAP_STATUS_INFEASIBLE        (1)
#define NAPHEAP_STATUS_UNBOUNDED         (2)
#define NAPHEAP_STATUS_OUT_OF_MEMORY     (3)
#define NAPHEAP_STATUS_INVALID_N         (4)
#define NAPHEAP_STATUS_INVALID_BOUNDS    (5)
#define NAPHEAP_STATUS_INVALID_NEWTON    (6)
#define NAPHEAP_STATUS_INVALID_D         (7)
#define NAPHEAP_STATUS_INVALID_ARG       (8)

/* -------------------------------------------------------------------------- */
/* prototypes for user-callable functions */
/* -------------------------------------------------------------------------- */

int napheap /* return status of solution process:
                   0 (solution found)
                   1 (infeasible, return x closest to feasible set)
                   2 (primal is -infinity)
                   3 (out of memory)
                   4 (n < 0)
                   5 (lo [j] > hi [j] for some j)
                   6 (Newton's method chosen but d_is_pos FALSE )
                   7 (Either d [j] < 0 or d [j] = 0 and d_is_pos TRUE) */
(
    /* output, not defined on input: */
    double        *x,   /* size n. solution */

    /* inputs, not modified on output: */
    double const  *y,   /* size n. linear term in objective function */
    double const  *d,   /* size n. diagonal of objective Hessian (can be NULL)*/
    double const  *a,   /* size n. linear constraint vector */
    double       Blo,   /* lower bound for a'x */
    double       Bhi,   /* upper bound for a'x */
    double const *lo,   /* size n. lower bounds for x */
    double const *hi,   /* size n. upper bounds for x */
    NAPHEAP_INT    n,   /* problem dimension */

    /* input/output: */
    double   *Lambda,   /* input: starting multiplier guess if parameter
                           use_lambda = TRUE, otherwise ignored on input.
                           output: final multiplier (scalar) */

    /* output: */
    napheap_stats *UStats,  /* structure with statistics (can be NULL) */

    /* input: */
    napheap_parameter *UParm, /*parameters, NULL => use defaults */

    /* workspace: */
    double     *Work,   /* NULL => let code allocate double memory
                           not NULL => use array Work for double memory.
                           size at most 5*n */

    NAPHEAP_INT *iWork  /* NULL => let code allocate NAPHEAP_INT memory
                           not NULL => use array Work for NAPHEAP_INT memory.
                           size at most 3*(n+1) */
) ;

/* set default parameter values */
void napheap_defaults
(
    napheap_parameter *Parm
) ;

/* print the return value and statistics from napheap */
void napheap_print_stats
(
    napheap_stats *Stats
) ;

/* print the parameters */
void napheap_print_parms
(
    napheap_parameter *Parm
) ;

/* check the solution from napheap */
double napheap_check /* returns largest of the 3 error values */
(
    double     *errb, /* relative error in b (only return if not NULL) */
    double     *errB, /* relative error in bounds (only return if not NULL) */
    double     *erry, /* relative error in y (only return if not NULL) */
    double const  *x, /* size n, solution */
    double const  *y, /* size n, linear term in objective function */
    double const  *d, /* size n, diagonal of objective Hessian (zero if NULL) */
    double const  *a, /* size n, linear constraint vector */
    double       Blo, /* lower bound for a'x */
    double       Bhi, /* upper bound for a'x */
    double const *lo, /* size n, lower bounds for x */
    double const *hi, /* size n, upper bounds for x */
    NAPHEAP_INT    n, /* problem dimension */
    double    lambda  /* multiplier (scalar) */
) ;

/* -------------------------------------------------------------------------- */
/* napheap version information */
/* -------------------------------------------------------------------------- */

#define NAPHEAP_DATE "October 21, 2013"
#define NAPHEAP_VER_CODE(main,sub) ((main) * 1000 + (sub))
#define NAPHEAP_MAIN_VERSION 1
#define NAPHEAP_SUB_VERSION 1
#define NAPHEAP_SUBSUB_VERSION 0
#define NAPHEAP_VERSION \
    NAPHEAP_VER_CODE(NAPHEAP_MAIN_VERSION,NAPHEAP_SUB_VERSION)

#endif
