/* ========================================================================== */
/* === MATLAB/napheap_check_mex.c =========================================== */
/* ========================================================================== */

/* napheap_check mexFunction */

#include "napheap_mex.h"

static void napheap_check_details ( void )
{
    printf (
"\n NAPHEAP_CHECK check the knapsack solution found by napheap\n"
"\n"
"   Determines KKT error in a solution to the separable convex quadratic\n"
"   knapsack problem:\n"
"\n"
"   min .5*x'*D*x - y'*x  subject to lo <= x <= hi, blo <= b <= bhi, a'*x = b\n"
"\n"
"   Input is the solution x and the Lagrange multiplier lambda associated\n"
"   with the constraint a'*x = b.  napheap_check determines the smallest\n"
"   perturbations in b, y, lo, hi, and lambda with the property that x and\n"
"   lambda satisfy the KKT conditions for the perturbed problem. The KKT\n"
"   conditions are the following:\n"
"\n"
"   1. If b = a'*x, then b = bhi if lambda > 0\n"
"                        b = blo if lambda < 0\n"
"                        b in [blo, bhi] if lambda = 0\n"
"\n"
"   2. lo (j) <= x (j) <= hi (j)\n"
"\n"
"   3. If d (j) = 0, then x (j) = hi (j) if a (j)*lambda - y (j) < 0\n"
"                         x (j) = lo (j) if a (j)*lambda - y (j) > 0\n"
"      If d (j) > 0, then\n"
"      x (j) = mid {lo (j), (y (j) - a (j)*lambda)/d (j), hi (j)}\n"
"      Here 'mid' denotes the mean (or middle) of the three quantities.\n"
"\n"
"   The code returns the relative 1-norm of the minimum perturbations\n"
"   in b, y, lo, hi, and lambda that are needed to satisfy these conditions.\n"
"   These three terms are computed, and the overall error is the largest of\n"
"   the three:\n"
"\n"
"   errb: (|pert b|+|pert lambda|)/ sum |a (j)*x (j)|\n"
"   errB: ||pert lo||_1 + ||pert hi||_1 over the 1-norm of the corresponding\n"
"       original components\n"
"   erry: ||pert y||_1 over the 1-norm of the corresponding original\n"
"       components\n"
"\n"
"   Example:\n"
"       [x lambda] = napheap (y, d, a, blo, bhi, lo, hi) ;\n"
"       [err, errb, errB, erry] = napheap_check ...\n"
"           (x, lambda, y, d, a, blo, bhi, lo, hi)\n"
"\n"
"   See also NAPHEAP, NAPHEAP_INSTALL, NAPHEAP_TEST\n"
"\n"
"   Copyright 2013, T. A. Davis, W. W. Hager, and J. T. Hungerford\n"
"   http://www.suitesparse.com\n"
"   For a paper on napheap, see napheap/Doc/napheap.pdf\n\n\n") ;
}

void mexFunction
(
    int nargout,
    mxArray *pargout [ ],
    int nargin,
    const mxArray *pargin [ ]
)
{
    NAPHEAP_INT n, j ;
    double err, errb, errB, erry, blo, bhi, lambda,
           *x, *y, *d, *a, *lo, *hi, *t ;
    double *a_default, *lo_default, *hi_default, *blo_default, *bhi_default,
           *d_default ;
    double inf = mxGetInf ( ) ;

    /* ---------------------------------------------------------------------- */
    /* get inputs */
    /* ---------------------------------------------------------------------- */

    if (nargin == 0 && nargin == 0)
    {
        napheap_check_details ( ) ;
        return ;
    }

    if (nargout > 4 || nargin > 9 || nargin < 3)
    {
        mexErrMsgIdAndTxt ("napheap:usage", "Usage: [err, errb, errB, erry] = "
            "napheap_check (x, lambda, y, d, a, blo, bhi, lo, hi)");
    }

    /* get the problem size: n = length (y (:)) */
    n = mxGetNumberOfElements (pargin [2]) ;

    /* get x, a vector of length n */
    x = napheap_get_vector (ARG (0), n, "x", NULL, 0.) ;

    /* get lambda, a scalar */
    t = napheap_get_vector (ARG (1), 1, "lambda", NULL, 0.) ;
    if (t == NULL)
    {
        mexErrMsgIdAndTxt ("napheap:usage", "lambda must be a scalar") ;
    }

    /* get y, a vector of length n, which defines the problem size */
    y = napheap_get_vector (ARG (2), n, "y", NULL, 0.) ;

    if (x == NULL || t == NULL || y == NULL)
    {
        /* quick return for an empty problem */
        pargout [0] = mxCreateDoubleScalar (0) ;
        return ;
    }

    lambda = t [0] ;

    /* get d, a vector of length n.  If not present, treat as d=ones(n,1) */
    d = napheap_get_vector (ARG (3), n, "d", &d_default, 1.) ;

    /* get a, a vector of length n.  If not present, treat as a=ones(n,1) */
    a = napheap_get_vector (ARG (4), n, "a", &a_default, 1.) ;

    /* get blo, a scalar.  If not present, treat as blo = 0 */
    t = napheap_get_vector (ARG (5), 1, "blo", &blo_default, 0.) ;
    blo = t [0] ;

    /* get bhi, a scalar.  If not present, treat as bhi = inf */
    t = napheap_get_vector (ARG (6), 1, "bhi", &bhi_default, inf) ;
    bhi = t [0] ;

    /* get lo, a vector of length n.  If not present, treat as lo=zeros(n,1) */
    lo = napheap_get_vector (ARG (7), n, "lo", &lo_default, 0.) ;

    /* get hi, a vector of length n.  If not present, treat as hi=inf(n,1) */
    hi = napheap_get_vector (ARG (8), n, "hi", &hi_default, inf) ;

    /* ---------------------------------------------------------------------- */
    /* check the knapsack problem */
    /* ---------------------------------------------------------------------- */

    err = napheap_check (&errb, &errB, &erry,
                         x, y, d, a, blo, bhi, lo, hi, n, lambda) ;

    /* ---------------------------------------------------------------------- */
    /* return results */
    /* ---------------------------------------------------------------------- */

    pargout [0] = mxCreateDoubleScalar (err) ;
    if (nargout > 1) pargout [1] = mxCreateDoubleScalar (errb) ;
    if (nargout > 2) pargout [2] = mxCreateDoubleScalar (errB) ;
    if (nargout > 3) pargout [3] = mxCreateDoubleScalar (erry) ;
}
