/* ========================================================================== */
/* === MATLAB/napheap_mex.c ================================================= */
/* ========================================================================== */

/* napheap mexFunction */

#include "napheap_mex.h"

static void napheap_details ( )
{
    printf (
"\n NAPHEAP: Solve a separable convex quadratic knapsack problem\n"
"\n"
"   Usage:\n"
"   [x lambda stats] = napheap (y, d, a, blo, bhi, lo, hi, parm, guess)\n"
"\n"
"   Solve a separable convex quadratic knapsack problem of the form:\n"
"   min .5*x'*D*x - y'*x  subject to  lo <= x <= hi, blo <= a'*x <= bhi.\n"
"   where y, d, a, lo, hi are vectors of length n, blo and bhi are scalars,\n"
"   and D=diag(d) is a diagonal matrix with a nonnegative diagonal.  If not\n"
"   present or empty, lower/upper bounds default to 0 and inf, respectively,\n"
"   and a and d default to ones(n,1).  The arguments d, a, lo, and hi can be\n"
"   provided as scalars, in which case they are expanded to vectors of size n\n"
"   where each entry has that scalar value.\n"
"\n"
"   The constraints are rewritten as: lo <= x <= hi, blo <= b <= bhi, a'x=b.\n"
"   The code solves the dual problem obtained by introducing a multiplier\n"
"   lambda for the constraint a'*x = b.  The dual function is L(lambda) =\n"
"   inf {.5*x'*D*x - y'*x + lambda*(a'*x-b): lo <= x <= hi, blo <=b <= bhi}.\n"
"\n"
"   The dual function is concave, and the solution to the original (primal)\n"
"   problem can be constructed from the maximizer of the dual function.\n"
"   If blo < bhi, then there is a discontinuity at lambda = 0.  D is a\n"
"   diagonal matrix, and if d = diag(D) contains zeros, there could be\n"
"   additional points of discontinuity.  The algorithm iterates on lambda\n"
"   until 0 lies in the subdifferential of L(lambda). The optimal x is\n"
"   constructed from the minimizers of the dual function.  Three algorithms\n"
"   are implemented:  a break point searching algorithm, a variable fixing\n"
"   algorithm, and a Newton/secant algorithm.  A starting guess for the\n"
"   optimal multiplier can be provided.  If the user does not have a good\n"
"   starting guess, a few iterations of the variable fixing algorithm can be\n"
"   used to generate a starting guess.  The number of iterations of Newton's\n"
"   method or variable fixing is given by parm.k (default 4).\n"
"\n"
"   The napheap algorithm starts by performing up to parm.k iterations of\n"
"   either Newton's method or variable fixing (selected via parm.newton).\n"
"   This is followed by the breakpoint algorithm.\n"
"\n"
"   By default, variable fixing is used (parm.newton = false, by default).\n"
"   Newton's method reduces to variable fixing if K = 0 or 1, so even if\n"
"   requested, it is not used when K = 0 or 1.  Newton's method when K > 1\n"
"   requires all entries of d to be greater than zero (an error is returned\n"
"   if any entry of d is zero and Newton's method is attempted).\n"
"   No value of d can be negative, for any method.\n"
"\n"
"   parm is a struct for for changing default parameter settings.  If not\n"
"   present or empty, defaults are used.\n"
"   parm.print:    if nonzero, print parameters & solution stats (default 0)\n"
"   parm.refine:   if nonzero, do iterative refinement (default 0)\n"
"   parm.err:      error tolerance for iterative refinement (default 0)\n"
"   parm.newton:   if nonzero, use Newton's method (default 0)\n"
"   parm.k or parm.K:  number of iterations in Newton's method or\n"
"                  variable fixing algorithm (default 4)\n"
"\n"
"   stats is an optional output struct with statistics on the solution:\n"
"   stats.status:  ok, infeasible, or unbounded (a string)\n"
"   stats.nfree:   number of free variables in initial heap\n"
"   stats.nbound:  number of bound variables in initial heap\n"
"   stats.nbrks:   number break points to reach initial solution\n"
"   stats.nrefine: number of break points during refinement\n"
"   stats.nvarfix: number of variable fixing iterations\n"
"   stats.nnewton: number of Newton iterations performed\n"
"   stats.nsecant: number of secant steps performed\n"
"\n"
"   See also NAPHEAP_INSTALL, NAPHEAP_TEST\n"
"\n"
"   Copyright 2013, T. A. Davis, W. W. Hager, and J. T. Hungerford\n"
"   http://www.suitesparse.com\n"
"   For a paper on napheap, see napheap/Doc/napheap.pdf\n\n\n") ;
}


/* -------------------------------------------------------------------------- */
/* napheap */
/* -------------------------------------------------------------------------- */

void mexFunction
(
    int nargout,
    mxArray *pargout [ ],
    int nargin,
    const mxArray *pargin [ ]
)
{
    int status, print = 0 ;
    NAPHEAP_INT n, *iwork, j ;
    double *x, *y, *d, *a, blo, bhi, *lo, *hi, lambda, *t, *xwork ;
    double *a_default, *lo_default, *hi_default, *blo_default, *bhi_default,
           *d_default ;
    napheap_parameter Parm ;
    napheap_stats Stats ;
    const mxArray *Parm_matlab ;
    double inf = mxGetInf ( ) ;

    /* ---------------------------------------------------------------------- */
    /* get inputs */
    /* ---------------------------------------------------------------------- */

    if (nargin == 0 && nargout == 0)
    {
        /* print the detailed help, and nothing else */
        napheap_details ( ) ;
        return ;
    }

    status = NAPHEAP_STATUS_OK ;
    Stats.status = status ;
    Stats.jerror = 0 ;
    Stats.lobad = 0 ;
    Stats.hibad = 0 ;
    Stats.dbad = 0 ;
    Stats.nfree = 0 ;
    Stats.nbound = 0 ;
    Stats.nbrks = 0 ;
    Stats.nrefine = 0 ;
    Stats.nvarfix = 0 ;
    Stats.nnewton = 0 ;
    Stats.nsecant = 0 ;

    if (nargout > 3 || nargin > 9)
    {
        mexErrMsgIdAndTxt ("napheap:usage", "Usage: [x lambda stats] = "
            "napheap (y, d, a, blo, bhi, lo, hi, parm, lambda)");
    }

    /* set default parameters */
    napheap_defaults (&Parm) ;

    /* get the problem size: n = length (y (:)) */
    n = (nargin > 0) ? mxGetNumberOfElements (pargin [0]) : 0 ;

    /* get y, a vector of length n, which defines the problem size */
    y = napheap_get_vector (ARG (0), n, "y", NULL, 0.) ;

    /* get a, a vector of length n.  If not present, treat as a=ones(n,1) */
    a = napheap_get_vector (ARG (2), n, "a", &a_default, 1.) ;

    /* get blo, a scalar.  If not present, treat as blo = 0 */
    t = napheap_get_vector (ARG (3), 1, "blo", &blo_default, 0.) ;
    blo = t [0] ;

    /* get bhi, a scalar.  If not present, treat as bhi = inf */
    t = napheap_get_vector (ARG (4), 1, "bhi", &bhi_default, inf) ;
    bhi = t [0] ;

    /* get lo, a vector of length n.  If not present, treat as lo=zeros(n,1) */
    lo = napheap_get_vector (ARG (5), n, "lo", &lo_default, 0.) ;

    /* get hi, a vector of length n.  If not present, treat as hi=inf(n,1) */
    hi = napheap_get_vector (ARG (6), n, "hi", &hi_default, inf) ;

    /* get Parm, a struct.  If not present, use defaults (or settings above) */
    Parm_matlab = ARG (7) ;
    if (Parm_matlab != NULL && !mxIsEmpty (Parm_matlab))
    {
        mxArray *field ;
        if (!mxIsStruct (Parm_matlab)
            || mxGetNumberOfElements (Parm_matlab) != 1)
        {
            mexErrMsgIdAndTxt ("napheap:invalid", "parm must be a struct") ;
        }

        /* Parm.print: if true, print diagnostics; otherwise do not print */
        /* default is no printing */
        if ((field = mxGetField (Parm_matlab, 0, "print")) != NULL)
        {
            print = (((int) mxGetScalar (field)) != 0) ;
        }

        /* Parm.refine: if true, do iterative refinement. default: no refine */
        if ((field = mxGetField (Parm_matlab, 0, "refine")) != NULL)
        {
            Parm.refine = (((int) mxGetScalar (field)) != 0) ;
        }

        /* Parm.err: solution refined if |a'x - b| > err. default: 0 */
        /* ignored if Parm.refine is false */
        if ((field = mxGetField (Parm_matlab, 0, "err")) != NULL)
        {
            Parm.err = mxGetScalar (field) ;
        }

        /* Parm.newton: if true, use Newton method to update lambda when K > 0.
           if false, use variable fixing to update lambda when K > 0.
           default: false */
        if ((field = mxGetField (Parm_matlab, 0, "newton")) != NULL)
        {
            Parm.newton = (((int) mxGetScalar (field)) != 0) ;
        }

        /* see above for Parm.use_lambda, Parm.d_is_pos, and Parm.d_is_zero */

        /* Parm.K: max number of Newton or variable fixing iterations. After
           doing these K iterations, napheap switches to the break point
           searching algorithm.  default: 4 */
        if ((field = mxGetField (Parm_matlab, 0, "k")) != NULL)
        {
            Parm.K = (int) mxGetScalar (field) ;
        }
        if ((field = mxGetField (Parm_matlab, 0, "K")) != NULL)
        {
            Parm.K = (int) mxGetScalar (field) ;
        }
    }

    /* get lambda, a scalar.  If not present, treat as lambda = 0 */
    t = napheap_get_vector (ARG (8), 1, "lambda", NULL, 0.) ;
    if (t == NULL)
    {
        lambda = 0 ;
    }
    else
    {
        /* lambda is present on input, so use it as input to napheap */
        lambda = t [0] ;
        Parm.use_lambda = 1 ;
    }

    /* get d, a vector of length n.  If not present, treat as d=ones(n,1) */
    d = napheap_get_vector (ARG (1), n, "d", &d_default, 1.) ;
    if (d == d_default)
    {
        /* d is empty or not present; treated as d = ones (n,1) */
        Parm.d_is_zero = 0 ;
        Parm.d_is_pos = 1 ;
    }
    else
    {
        /* d is provided by the user, so check it */
        Parm.d_is_zero = 1 ;
        Parm.d_is_pos = 1 ;
        for (j = 0 ; j < n ; j++)
        {
            if (d [j] < 0)
            {
                Stats.status = NAPHEAP_STATUS_INVALID_D ;
                Stats.jerror = j ;
                Stats.dbad = d [j] ;
                if (print)
                {
                    napheap_print_parms (&Parm) ;
                    napheap_print_stats (&Stats) ;
                }
                mexErrMsgIdAndTxt ("napheap:invalid",
                    "all entries in d must be non-negative") ;
            }
            else if (d [j] == 0)
            {
                /* d has at least one zero on the diagonal of the Hessian */
                Parm.d_is_pos = 0 ;
            }
            else /* d [j] != 0) */
            {
                /* d has at least one nonzero on the diagonal of the Hessian */
                Parm.d_is_zero = 0 ;
            }
        }
    }

    /* ---------------------------------------------------------------------- */
    /* allocate the result, and workspace */
    /* ---------------------------------------------------------------------- */

    pargout [0] = mxCreateDoubleMatrix (n, 1, mxREAL) ;
    x = mxGetPr (pargout [0]) ;
    iwork = (NAPHEAP_INT *) mxMalloc (3 * (n+1) * sizeof (NAPHEAP_INT)) ;
    xwork = (double *)      mxMalloc (5 * (n+1) * sizeof (double)) ;

    /* ---------------------------------------------------------------------- */
    /* solve the knapsack problem */
    /* ---------------------------------------------------------------------- */

    status = napheap (x, y, d, a, blo, bhi, lo, hi, n, &lambda, &Stats, &Parm,
        xwork, iwork) ;

    /* ---------------------------------------------------------------------- */
    /* check for errors */
    /* ---------------------------------------------------------------------- */

    if (status == NAPHEAP_STATUS_INFEASIBLE)
    {
        /* this is just a warning */
        mexWarnMsgIdAndTxt ("napheap:infeasible",
            "napheap: problem is infeasible") ;
    }
    else if (status == NAPHEAP_STATUS_UNBOUNDED)
    {
        /* this is just a warning */
        mexWarnMsgIdAndTxt ("napheap:unbounded",
            "napheap: problem is unbounded") ;
    }
    else if (status == NAPHEAP_STATUS_INVALID_BOUNDS)
    {
        /* the rest are hard errors */
        if (print) napheap_print_stats (&Stats) ;
        mexErrMsgIdAndTxt ("napheap:invalid", "napheap: invalid bounds") ;
    }
    else if (status == NAPHEAP_STATUS_INVALID_NEWTON)
    {
        if (print) napheap_print_stats (&Stats) ;
        mexErrMsgIdAndTxt ("napheap:invalid",
            "napheap: Newton's method requires d>0") ;
    }
    else if (status == NAPHEAP_STATUS_INVALID_D)
    {
        if (print) napheap_print_stats (&Stats) ;
        mexErrMsgIdAndTxt ("napheap:invalid", "napheap: d invalid") ;
    }
    else if (status != NAPHEAP_STATUS_OK)
    {
        if (print) napheap_print_stats (&Stats) ;
        mexErrMsgIdAndTxt ("napheap:invalid", "napheap: failed") ;
    }

    /* ---------------------------------------------------------------------- */
    /* free workspace */
    /* ---------------------------------------------------------------------- */

    if (a_default   != NULL) mxFree (a_default) ;
    if (lo_default  != NULL) mxFree (lo_default) ;
    if (hi_default  != NULL) mxFree (hi_default) ;
    if (blo_default != NULL) mxFree (blo_default) ;
    if (bhi_default != NULL) mxFree (bhi_default) ;
    mxFree (iwork) ;
    mxFree (xwork) ;

    /* ---------------------------------------------------------------------- */
    /* return results */
    /* ---------------------------------------------------------------------- */

    if (nargout > 1)
    {
        pargout [1] = mxCreateDoubleScalar (lambda) ;
    }

    if (nargout > 2)
    {
        static const char *statnames [ ] = {
            "status",      /* problem ok / infeas / unbound */
            "nfree",       /* number of free variables in initial heap */
            "nbound",      /* number of bound variables in initial heap */
            "nbrks",       /* number break points to reach initial solution*/
            "nrefine",     /* number of break points during refinement */
            "nvarfix",     /* number of variable fixing iterations */
            "nnewton",     /* number of Newton iterations performed */
            "nsecant" } ;  /* number of secant steps performed */

        pargout [2] = mxCreateStructMatrix (1, 1, 8, statnames) ;

        /* warnings only, since errors would have terminated napheap already */
        if (status == NAPHEAP_STATUS_OK)
        {
            mxSetFieldByNumber (pargout [2], 0, 0,
                mxCreateString ("ok")) ;
        }
        else if (status == NAPHEAP_STATUS_INFEASIBLE)
        {
            mxSetFieldByNumber (pargout [2], 0, 0,
                mxCreateString ("problem is infeasible")) ;
        }
        else if (status == NAPHEAP_STATUS_UNBOUNDED)
        {
            mxSetFieldByNumber (pargout [2], 0, 0,
                mxCreateString ("problem is unbounded")) ;
        }

        mxSetFieldByNumber (pargout [2], 0, 1,
            mxCreateDoubleScalar ((double) Stats.nfree)) ;

        mxSetFieldByNumber (pargout [2], 0, 2,
            mxCreateDoubleScalar ((double) Stats.nbound)) ;

        mxSetFieldByNumber (pargout [2], 0, 3,
            mxCreateDoubleScalar ((double) Stats.nbrks)) ;

        mxSetFieldByNumber (pargout [2], 0, 4,
            mxCreateDoubleScalar ((double) Stats.nrefine)) ;

        mxSetFieldByNumber (pargout [2], 0, 5,
            mxCreateDoubleScalar ((double) Stats.nvarfix)) ;

        mxSetFieldByNumber (pargout [2], 0, 6,
            mxCreateDoubleScalar ((double) Stats.nnewton)) ;

        mxSetFieldByNumber (pargout [2], 0, 7,
            mxCreateDoubleScalar ((double) Stats.nsecant)) ;
    }

    /* ---------------------------------------------------------------------- */
    /* print results, if requested */
    /* ---------------------------------------------------------------------- */

    if (print)
    {
        napheap_print_parms (&Parm) ;
        napheap_print_stats (&Stats) ;
    }
}
