function napheap_test (reset_rand)
%NAPHEAP_TEST test the napheap function in MATLAB
%
% Note: this function resets the MATLAB random number generator, by default.
% To avoid this, pass in a '0' as the first argument.
%
% Example
%   napheap_test ;          % run the tests
%   napheap_test (0) ;      % run the tests without reseting rand
%
% See also NAPHEAP, NAPHEAP_CHECK, NAPHEAP_INSTALL.

% Copyright 2013, T. A. Davis, W. W. Hager, and J. T. Hungerford

if (nargin < 1)
    reset_rand = 1 ;
end
if (reset_rand)
    rng ('default') ;
    fprintf ('reset rand\n') ;
end

[x, lambda] = napheap ([ ]) ;
err = napheap_check (x, lambda, [ ]) ;
errmax = err ;
fprintf ('err0: %g\n', err) ;

[x, lambda] = napheap (0) ;
err = napheap_check (x, lambda, 0) ;
errmax = max (err, errmax) ;
fprintf ('err0: %g\n', err) ;

for n = [1 4 10 20]

    parm = [ ] ;

    fprintf ('\n----- n = %d\n', n) ;
    y = rand (n,1) ;
    [x, lambda] = napheap (y) ;
    err = napheap_check (x, lambda, y) ;
    errmax = max (err, errmax) ;
    fprintf ('err1: %g\n', err) ;

    [err, errb, errB, erry] = napheap_check (x, lambda, y) ;
    errmax = max (err, errmax) ;
    fprintf ('err1: %g (%g %g %g)\n', err, errb, errB, erry) ;

    % test default inputs (should be the same as solution above)
    d = ones (n,1) ;
    a = ones (n,1) ;
    blo = 0 ;
    bhi = inf ;
    lo = zeros (n,1) ;
    hi = inf (n,1) ;
    [x2, lambda2] = napheap (y, d, a, blo, bhi, lo, hi) ;
    err = norm (x-x2) + abs (lambda - lambda2) ;
    errmax = max (err, errmax) ;
    fprintf ('err1b: %g\n', err) ;

    d = 5 + 5*rand (n, 1) ;
    a = 20*rand (n, 1) - 10 ;
    y = 20*rand (n, 1) - 10 ;
    lo = -ones (n, 1) ;
    hi =  ones (n, 1) ;
    cumlo = 0 ;
    cumhi = 0 ;
    for j = 1:n
        t = a (j)*lo (j) ;
        s = a (j)*hi (j) ;
        cumlo = cumlo + min (t, s) ;
        cumhi = cumhi + max (t, s) ;
    end
    blo = cumlo + (cumhi-cumlo)*rand ;
    bhi = blo + 10 ;
    [x, lambda] = napheap (y, d, a, blo, bhi, lo, hi) ;
    [err, errb, errB, erry] = ...
        napheap_check (x, lambda, y, d, a, blo, bhi, lo, hi) ;
    errmax = max (err, errmax) ;
    fprintf ('err2: %g (%g %g %g)\n', err, errb, errB, erry) ;

    % try with scalar expansion for lo and hi
    [x2, lambda2] = napheap (y, d, a, blo, bhi, -1, 1) ;
    [err, errb, errB, erry] = ...
        napheap_check (x, lambda, y, d, a, blo, bhi, lo, hi) ;
    errmax = max (err, errmax) ;
    fprintf ('err2c: %g (%g %g %g)\n', err, errb, errB, erry) ;
    err = norm (x-x2) + abs (lambda - lambda2) ;
    errmax = max (err, errmax) ;
    fprintf ('err2d: %g\n', err) ;

    % try an input guess for lambda
    [x, lambda] = napheap (y, d, a, blo, bhi, lo, hi, parm, lambda) ;
    err = napheap_check (x, lambda, y, d, a, blo, bhi, lo, hi) ;
    errmax = max (err, errmax) ;
    fprintf ('err3: %g\n', err) ;

    % try an input guess for lambda again
    [x, lambda] = napheap (y, d, a, blo, bhi, lo, hi, parm, lambda) ;
    err = napheap_check (x, lambda, y, d, a, blo, bhi, lo, hi) ;
    errmax = max (err, errmax) ;
    fprintf ('err4: %g\n', err) ;

    % try hi = inf, with scalar expansion
    [x, lambda, stats] = napheap (y, d, a, blo, bhi, lo, inf) ;
    err = napheap_check (x, lambda, y, d, a, blo, bhi, lo, inf) ;
    display (stats) ;
    errmax = max (err, errmax) ;
    fprintf ('err5: %g\n', err) ;

    % try with non-default parameters
    for print = 0:1
        parm.print = print ;
        for refine = 0:1
            parm.refine = refine ;
            for e = [0 1e-6]
                parm.err = e ;
                for k = [1 3]
                    parm.k = k ;
                    for newt = 0:1
                        parm.newton = newt ;
                        [x, lambda] = napheap (y, d, a, blo, bhi, lo, hi, parm);
                        err = napheap_check (x, lambda, y, d, a, blo, bhi, ...
                            lo, hi) ;
                        errmax = max (err, errmax) ;
                        fprintf ('err6: %g\n', err) ;
                    end
                end
            end
        end
    end

end

%-------------------------------------------------------------------------------
% help
%-------------------------------------------------------------------------------

napheap_check
help napheap_check
napheap
help napheap

%-------------------------------------------------------------------------------
% error handling
%-------------------------------------------------------------------------------

ok = 1 ;

% d must be non-negative
try
    x = napheap (1, -1) ;                                                   %#ok
    ok = 0 ;
catch                                                                       %#ok
end

% too many output args
try
    [too, many, output, args, here] = napheap (y) ;                         %#ok
    ok = 0 ;
catch                                                                       %#ok
end

% lambda must be a scalar
gunk = rand (n,1) ;
try
    [x, lambda] = napheap (y, d, a, blo, bhi, lo, hi, parm, gunk) ;         %#ok
    ok = 0 ;
catch                                                                       %#ok
end

% sparse case not handled
try
    x = napheap (sparse (y)) ;                                              %#ok
    ok = 0 ;
catch                                                                       %#ok
end

% complex case not handled
try
    x = napheap (y * 1i) ;                                                  %#ok
    ok = 0 ;
catch                                                                       %#ok
end

% non-double case not handled
try
    x = napheap (int32 (y)) ;                                               %#ok
    ok = 0 ;
catch                                                                       %#ok
end

% length of d and y must match
try
    x = napheap (y, [d ; 1]) ;                                              %#ok
    ok = 0 ;
catch                                                                       %#ok
end

if (~ok)
    error ('error handling tests failed') ;
end

%-------------------------------------------------------------------------------
% wrapup and report results
%-------------------------------------------------------------------------------

fprintf ('\nworst-case relative error: %g\n', errmax) ;

if (errmax < 1e-6)
    fprintf ('All tests passed.  Napheap is compiled and working fine.\n') ;
else
    error ('TEST FAILURE') ;
end

