/* ========================================================================== */
/* === MATLAB/napheap_mex_util.c ============================================ */
/* ========================================================================== */

/* utility functions for napheap mexFunctions */

#include "napheap_mex.h"

/* -------------------------------------------------------------------------- */
/* default_vector */
/* -------------------------------------------------------------------------- */

/* allocate a default vector and populate it with a default_value */

double *napheap_default_vector
(
    NAPHEAP_INT n,
    double **default_result,
    double default_value
)
{
    double *x ;
    NAPHEAP_INT j ;
    x = mxMalloc (MAX (1,n) * sizeof (double)) ;
    *default_result = x ;
    for (j = 0 ; j < n ; j++)
    {
        x [j] = default_value ;
    }
    return (x) ;
}


/* -------------------------------------------------------------------------- */
/* napheap_get_vector */
/* -------------------------------------------------------------------------- */

/* Get a real vector of length n from the input parameter X.  X must be
   double-precision (not complex), and not sparse.  If X is a matrix or a row
   vector, then it is treated as a column vector, X(:).  If X is empty, return
   NULL, or allocate a default vector.  If X is a scalar, it is expanded to a
   vector of length n.
 */

double *napheap_get_vector
(
    const mxArray *X,
    NAPHEAP_INT n,
    char *name,
    double **default_result,
    double default_value
)
{
    NAPHEAP_INT nel ;

    if (default_result != NULL)
    {
        /* initialize the default vector */
        *default_result = NULL ;
    }

    if (X == NULL || mxIsEmpty (X))
    {
        /* X is empty or not present */
        if (default_result == NULL)
        {
            return (NULL) ;
        }
        else
        {
            /* allocate and return a default vector */
            return (napheap_default_vector (n, default_result, default_value)) ;
        }
    }

    /* make sure X is not sparse, not complex, and double */
    if (mxIsSparse (X))
    {
        mexErrMsgIdAndTxt ("napheap:invalid", "%s cannot be sparse", name) ;
    }

    if (mxIsComplex (X))
    {
        mexErrMsgIdAndTxt ("napheap:invalid", "%s cannot be complex", name) ;
    }

    if (!mxIsDouble (X))
    {
        mexErrMsgIdAndTxt ("napheap:invalid", "%s must be double", name) ;
    }

    /* make sure X is the right size */
    nel = mxGetNumberOfElements (X) ;
    if (nel != n)
    {
        if (nel == 1)
        {
            /* expand the input scalar into a vector of size n */
            return (napheap_default_vector (n, default_result, mxGetScalar(X)));
        }
        else if (n == 1)
        {
            mexErrMsgIdAndTxt ("napheap:invalid", "%s must be a scalar", name) ;
        }
        else
        {
            mexErrMsgIdAndTxt ("napheap:invalid",
                "%s must be a vector of length %d", name, n) ;
        }
    }

    /* return the contents of X */
    return (mxGetPr (X)) ;
}

