/* ========================================================================== */
/* === Source/napheap_check.c =============================================== */
/* ========================================================================== */

#include "napheap.h"

#include <math.h>
#include <stdlib.h>

#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define ZERO ((double) 0)

/*  Determine KKT error in a solution to the separable convex quadratic
    knapsack problem:

    min .5x'Dx - y'x  subject to  lo <= x <= hi, blo <= b <= bhi, a'x = b

    Input is the solution x and the Lagrange multiplier lambda associated
    with the constraint a'x = b. The code determines the smallest perturbations
    in b, y, lo, hi, and lambda with the property that x and lambda satisfy
    the KKT conditions for the perturbed problem. The KKT conditions are the
    following:

    1. If b = a'x, then b = bhi if lambda > 0
                        b = blo if lambda < 0
                        b in [blo, bhi] if lambda = 0

    2. lo [j] <= x [j] <= hi [j]

    3. If d [j] = 0, then x [j] = hi [j] if a [j]*lambda - y [j] < 0
                          x [j] = lo [j] if a [j]*lambda - y [j] > 0
       If d [j] > 0, then

       x [j] = mid {lo [j], (y [j] - a [j]*lambda)/d [j], hi [j]}

       Here `mid' denotes the mean (or middle) of the three quantities.

    The code returns the relative 1-norm of the minimum perturbations
    in b, y, lo, hi, and lambda that are needed to satisfy these conditions.
    These are output in 3 arguments:

    errb: (|pert b|+|pert lambda|)/ sum |a [j]*x [j]|

    errB: ||pert lo||_1 + ||pert hi||_1 over the 1-norm of the corresponding
          original components

    erry: ||pert y||_1 over the 1-norm of the corresponding original components
*/

double napheap_check /* returns largest of the 3 error values */
(
    double     *errb, /* relative error in b (only return if not NULL) */
    double     *errB, /* relative error in bounds (only return if not NULL) */
    double     *erry, /* relative error in y (only return if not NULL) */
    double const  *x, /* size n, solution */
    double const  *y, /* size n, linear term in objective function */
    double const  *d, /* size n, diagonal of objective Hessian (zero if NULL) */
    double const  *a, /* size n, linear constraint vector */
    double       Blo, /* lower bound for a'x */
    double       Bhi, /* upper bound for a'x */
    double const *lo, /* size n, lower bounds for x */
    double const *hi, /* size n, upper bounds for x */
    NAPHEAP_INT    n, /* problem dimension */
    double    lambda  /* multiplier (scalar) */
)
{
    NAPHEAP_INT j ;
    double ax, blo, bhi, t, s, err, normB, normy, pertb, pertB, perty,
           loj, hij, dj ;

    blo = Blo ;
    bhi = Bhi ;
    if ( blo > bhi )
    {
        t = blo ;
        blo = bhi ;
        bhi = t ;
    }

    /* determine b perturbation */
    t = ZERO ;
    ax = ZERO ;
    for (j = 0; j < n; j++)
    {
        s = a [j]*x [j] ;
        t += s ;
        ax += fabs (s) ;
    }

    if ( lambda > ZERO )
    {
        pertb = fabs (t-bhi) ;
        if ( lambda < pertb ) /* less change if lambda is perturbed */
        {
            pertb = lambda ;
        }
    }
    else if ( lambda < ZERO )
    {
        pertb = fabs (t-blo) ;
        if ( -lambda < pertb ) /* less change if lambda is perturbed */
        {
            pertb = -lambda ;
        }
    }
    else
    {
        pertb = MAX (blo-t, ZERO) ;
        pertb = MAX (t-bhi, pertb) ;
    }
    /* error relative to sum |aj*xj| */
    if ( ax > ZERO ) pertb /= ax ;

    /* determine perturbations in bounds and in y */
    pertB = ZERO ;
    normB = ZERO ;
    perty = ZERO ;
    normy = ZERO ;
    dj = ZERO ;
    for (j = 0; j < n; j++)
    {
        loj = lo [j] ;
        hij = hi [j] ;
        if      ( x [j] < loj ) loj = x [j] ;
        else if ( x [j] > hij ) hij = x [j] ;
        if ( d != NULL ) dj = d [j] ;
        if ( dj > ZERO )
        {
            t = dj*x [j] + a [j]*lambda ; /* y [j] that yield x [j] */
            s = fabs (t - y [j]) ;        /* perturbation in y */
            /* we have the choice of perturbing y, lo, or hi */
            if ( y [j] > t ) /*can eliminate y pert by moving hij to x[j]*/
            {
                if ( hij - x [j] < s ) /* pert in hi < pert in y */
                {
                    hij = x [j] ;
                    s = ZERO ;
                }
            }
            else /* y [j] <= t, can move loj to x [j] */
            {
                if ( x [j] - loj < s ) /* pert in lo < pert in y */
                {
                    loj = x [j] ;
                    s = ZERO ;
                }
            }
        }
        else /* d [j] = 0 */
        {
            /*  x [j] = hi [j] if a [j]*lambda - y [j] < 0
                x [j] = lo [j] if a [j]*lambda - y [j] > 0
                we have the choice of perturbing y, lo, or hi */
            t = a [j]*lambda - y [j] ;
            s = fabs (t) ;
            if ( t < ZERO )
            {
                if ( hij - x [j] < s ) /* pert hi < pert y */
                {
                    hij = x [j] ;
                    s = ZERO ;
                }
            }
            else /* t <= 0 */
            {
                if ( x [j] - loj < s ) /* pert lo < pert y */
                {
                    loj = x [j] ;
                    s = ZERO ;
                }
            }
        }

        /* accumulate the errors */
        if ( s > ZERO )
        {
            perty += s ;
            normy += fabs (y [j]) ;
        }
        if ( loj != lo [j] )
        {
            pertB += fabs (loj-lo [j]) ;
            normB += fabs (lo [j]) ;
        }
        if ( hij != hi [j] )
        {
            pertB += fabs (hij-hi [j]) ;
            normB += fabs (hi [j]) ;
        }
    }
    if ( normy > ZERO ) perty /= normy ;
    if ( normB > ZERO ) pertB /= normB ;

    if ( errb != NULL ) *errb = pertb ;
    if ( errB != NULL ) *errB = pertB ;
    if ( erry != NULL ) *erry = perty ;

    /* return the largest error */
    err = MAX (pertb, perty) ;
    err = MAX (err, pertB) ;
    return (err) ;
}
