/* ========================================================================== */
/* === Source/napheap_print.c =============================================== */
/* ========================================================================== */

/* user-callable functions for printing napheap parameters and statistics. */

#include "napheap.h"
#include <stdio.h>

#ifdef MATLAB_MEX_FILE
#define LPAREN "("
#define RPAREN ")"
#define OFFSET 1
#else
#define LPAREN "["
#define RPAREN "]"
#define OFFSET 0
#endif

/* ========================================================================== */
/* === napheap_print_stats ================================================== */
/* ========================================================================== */

/* print the return value and statistics from napheap */

void napheap_print_stats
(
    napheap_stats *Stats
)
{
    printf ("\nnapheap Version %d.%d.%d, %s.\n",
        NAPHEAP_MAIN_VERSION, NAPHEAP_SUB_VERSION, NAPHEAP_SUBSUB_VERSION,
        NAPHEAP_DATE) ;

    if (Stats == NULL)
    {
        return ;
    }

    printf ("Stats.status: (%d) ", Stats->status) ;
    if ( Stats->status == NAPHEAP_STATUS_OK )
    {
        printf ("no errors detected when solving the knapsack problem.\n") ;
    }
    else if ( Stats->status == NAPHEAP_STATUS_INFEASIBLE )
    {
        printf ("knapsack problem is infeasible.\n") ;
    }
    else if ( Stats->status == NAPHEAP_STATUS_UNBOUNDED )
    {
        printf ("optimal value in knapsack problem is -infinity.\n") ;
    }
    else if ( Stats->status == NAPHEAP_STATUS_OUT_OF_MEMORY )
    {
        printf ("out of memory.\n") ;
    }
    else if ( Stats->status == NAPHEAP_STATUS_INVALID_N )
    {
        printf ("dimension n = %g < 0.\n", (double) Stats->jerror) ;
    }
    else if ( Stats->status == NAPHEAP_STATUS_INVALID_BOUNDS )
    {
        printf ("invalid knapsack problem: lo %s%g%s = %e > hi %s%g%s = %e\n",
            LPAREN, (double) Stats->jerror + OFFSET, RPAREN, Stats->lobad,
            LPAREN, (double) Stats->jerror + OFFSET, RPAREN, Stats->hibad) ;
    }
    else if ( Stats->status == NAPHEAP_STATUS_INVALID_NEWTON )
    {
        printf ("Newton's method can only be used when "
                "Parm->d_is_pos is true.\n") ;
    }
    else if ( Stats->status == NAPHEAP_STATUS_INVALID_D )
    {
        if ( Stats->dbad < 0 )
        {
            printf ("d invalid (d %s%g%s = %e < 0)\n",
                LPAREN, (double) Stats->jerror + OFFSET, RPAREN, Stats->dbad) ;
        }
        else
        {
            printf ("d invalid (d %s%g%s = 0 but Parm.d_is_pos is true)\n",
                LPAREN, (double) Stats->jerror + OFFSET, RPAREN) ;
        }
    }
    else if ( Stats->status == NAPHEAP_STATUS_INVALID_ARG )
    {
        printf ("input argument(s): x, y, a, lo, hi, or Lambda missing\n") ;
    }

    printf ("Stats.nfree (free variables in initial heap):         %g\n",
        (double) Stats->nfree) ;
    printf ("Stats.nbound (bound variables in initial heap):       %g\n",
        (double) Stats->nbound) ;
    printf ("Stats.nbrks (break points to reach initial solution): %g\n",
        (double) Stats->nbrks) ;
    printf ("Stats.nrefine (break points during refinement):       %g\n",
        (double) Stats->nrefine) ;
    printf ("Stats.nvarfix (variable fixing iterations):           %g\n",
        (double) Stats->nvarfix) ;
    printf ("Stats.nnewton (number of Newton iterations):          %g\n",
        (double) Stats->nnewton) ;
    printf ("Stats.nsecant (number of secant iterations):          %g\n",
        (double) Stats->nsecant) ;
}

/* ========================================================================== */
/* === napheap_print_parms ================================================== */
/* ========================================================================== */

void napheap_print_parms
(
    napheap_parameter  *Parm        /* parameter structure */
)
{
    napheap_parameter Default_Parm ;

    printf ("\nnapheap Version %d.%d.%d, %s.  parameter settings:\n",
        NAPHEAP_MAIN_VERSION, NAPHEAP_SUB_VERSION, NAPHEAP_SUBSUB_VERSION,
        NAPHEAP_DATE) ;

    if (Parm == NULL)
    {
        /* initialize the default parameters */
        Parm = &Default_Parm ;
        napheap_defaults (Parm) ;
    }

    printf ("Parm.refine:  ") ;
    if ( Parm->refine )
    {
        printf ("true (perform refinement step)\n") ;
        printf ("Parm.err:  %e (solution refined if constraint violation"
            " exceeds this value)\n", Parm->err) ;
    }
    else
    {
        printf ("false (do not perform refinement step)\n") ;
        printf ("Parm.err:  ignored (only used for refinement)\n") ;
    }

    printf ("Parm.K:  ") ;
    if ( Parm->K <= 0 )
    {
        printf ("0 (only use break point searching algorithm)\n") ;
        printf ("Parm.newton: ignored (only used if K > 1)\n") ;
    }
    else if ( Parm->K == 1 )
    {
        printf ("1 (one iteration of variable fixing algorithm)\n") ;
        printf ("Parm.newton: ignored (only used if K > 1)\n") ;
    }
    else
    {
        printf ("%d (max number of iterations in Newton or variable"
            " fixing algorithm)\n", Parm->K) ;
        printf ("Parm.newton:  ") ;
        if ( Parm->newton )
        {
            printf ("true (use Newton's method)\n") ;
        }
        else
        {
            printf ("false (use variable fixing algorithm)\n") ;
        }
    }

#ifndef MATLAB_MEX_FILE
    printf ("Parm.use_lambda:  ") ;
    if ( Parm->use_lambda )
    {
        printf ("true (use input lambda as starting guess for multiplier)\n") ;
    }
    else
    {
        printf ("false (ignore input argument lambda)\n") ;
    }

    printf ("Parm.d_is_zero:  ") ;
    if ( Parm->d_is_zero )
    {
        printf ("true (diagonal of objective Hessian is zero)\n") ;
        printf ("Parm.d_is_pos:  ignored, since Parm.d_is_zero is true\n") ;
    }
    else
    {
        printf ("false (d, diagonal of objective Hessian is nonzero)\n");
        printf ("Parm.d_is_pos:  ") ;
        if ( Parm->d_is_pos )
        {
            printf ("true (diagonal of objective Hessian is > 0)\n") ;
        }
        else
        {
            printf ("false (diagonal of objective Hessian is >= 0)\n") ;
        }
        printf ("Note: if d=NULL, then these settings "
            "are ignored and d is treated as zero\n") ;
    }
#endif
}
