/* ========================================================================== */
/* === Test/naptest.c ======================================================= */
/* ========================================================================== */

/* exhaustive test for napheap */

#include "napheap.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/resource.h>

#define ZERO ((double) 0)
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define FALSE 0
#define TRUE 1

#define NDIM 53
#define INF NAPHEAP_INF

static void check_status (int status, int should_be)
{
    if (status != should_be)
    {
        printf ("got status %d, should be %d\n", status, should_be) ;
        fprintf (stderr, "TEST FAILURE\n") ;
        exit (0) ;
    }
}

/* portable replacement for rand, for repeatable results */
static unsigned long next = 1 ;
#define MY_RAND_MAX 32767
static int myrand ( void )
{
    next = next * 1103515245 + 12345;
    return ((unsigned) (next / 65536) % (MY_RAND_MAX + 1)) ;
}

/* generate a random number between lower and upper */
static double rnum
(
    double lower,
    double upper
)
{
    double diff, t ;
    t = ((double) myrand ()) / MY_RAND_MAX ;
    diff = upper - lower ;
    t = t*diff + lower ;
    return (t) ;
}

/* generate a random positive number between 0 and upper */
static double rnumpos
(
    double upper
)
{
    double t ;
    int k = myrand ( ) ;
    t = ((double) k + 1)/((double) MY_RAND_MAX + 1) ;

    if (t < 1e-6) printf ("myrand k %d t %g t*upper %g\n", k, t, t*upper) ;

    t = t*upper ;
    return (t) ;
}

/* generate a random permutation of the integers 0 through n-1 stored in perm */
static void randperm
(
    NAPHEAP_INT *perm,
    NAPHEAP_INT     n
)
{
    NAPHEAP_INT j, k, l ;
    for (j = n-1; j > 0; j--)
    {
        k = myrand () % (j+1) ;
        l = perm [j] ;
        perm [j] = perm [k] ;
        perm [k] = l ;
    }
    return ;
}

static double Timer (void)
{
    struct rusage ru ;
    double user, tsys ;

    (void) getrusage (RUSAGE_SELF, &ru) ;

    user =
    ru.ru_utime.tv_sec                  /* user time (seconds) */
    + 1e-6 * ru.ru_utime.tv_usec ;      /* user time (microseconds) */

    tsys =
    ru.ru_stime.tv_sec                  /* tsys time (seconds) */
    + 1e-6 * ru.ru_stime.tv_usec ;      /* tsys time (microseconds) */

    return (user + tsys) ;
}

/* -------------------------------------------------------------------------- */
/* check the solution returned by napheap */
/* -------------------------------------------------------------------------- */

static void check_solution
(
    int d_is_zero,      /* if true, pass d = NULL to napheap */
    double        *x,   /* size n. solution from napheap */
    double const  *y,   /* size n. linear term in objective function */
    double const  *d,   /* size n. diagonal of objective Hessian */
    double const  *a,   /* size n. linear constraint vector */
    double       Blo,   /* lower bound for a'x */
    double       Bhi,   /* upper bound for a'x */
    double const *lo,   /* size n. lower bounds for x */
    double const *hi,   /* size n. upper bounds for x */
    NAPHEAP_INT    n,   /* problem dimension */
    double     guess,   /* initial guess for lambda */
    double    lambda,   /* final multiplier */
    napheap_stats *Stats,   /* structure with statistics (can be NULL) */
    napheap_parameter *Parm,/*parameter structure, NULL => use default */

    double cpu,
    int status,         /* napheap return value */
    int refine,
    int set,
    int method,
    int trial,
    FILE *f
)
{
    double errb, errB, erry, tol, err ;

    if (status != 0)
    {
        return ;
    }

    err = napheap_check (&errb, &errB, &erry, x, y,
        d_is_zero ? NULL : d, a, Blo, Bhi, lo, hi, n, lambda) ;

    if (f != NULL)
    {
        /* diagnostic output */
        fprintf (f, "%2i %g %3i %6.3f %9.2e %9.2e %9.2e %1i",
             set, (double) n, trial, cpu, errb, errB, erry, status) ;

        if (Stats != NULL)
        {
            fprintf (f, "stats: %8i %8i %8i %8i %8i %8i %8i\n", Stats->nfree,
                Stats->nbound, Stats->nbrks, Stats->nrefine, Stats->nvarfix,
                Stats->nnewton, Stats->nsecant) ;
        }
        fprintf (f, "\n") ;
    }

    tol = (refine) ? 2e-6 : 1e-4 ;
    if ( set == 7 ) tol = 1.e-3 ;

    if ( err > tol )
    {
        /* test failure; save the problem to a MATLAB script, and abort */
        FILE *fbug ;
        NAPHEAP_INT i ;
        fprintf (stderr, "TEST FAILURE\n") ;
        printf ("\n\nTEST FAILURE\n") ;
        printf ("method %d set %d n %d trial %d refine %d use_lambda %d "
            "d_is_zero %d Blo %e Bhi %e\n",
            method, set, n, trial, refine,
            (Parm == NULL) ? FALSE : Parm->use_lambda,
            d_is_zero, Blo, Bhi) ;
        printf ("%d %g %d %6.3f err: %9.2e %9.2e %9.2e\n",
             set, (double) n, trial, cpu, errb, errB, erry) ;

        if (Stats != NULL)
        {
            printf ("stats: %8i %8i %8i %8i %8i %8i %8i\n", Stats->nfree,
                Stats->nbound, Stats->nbrks, Stats->nrefine, Stats->nvarfix,
                Stats->nnewton, Stats->nsecant) ;
        }

        fbug = fopen ("bugprob.m", "w") ;
        fprintf (fbug, "%% test failure:\n") ;
        fprintf (fbug, "n = %g ;\n", (double) n) ;
        fprintf (fbug, "blo = %32.18e ;\n", Blo) ;
        fprintf (fbug, "bhi = %32.18e ;\n", Bhi) ;
        fprintf (fbug, "y = [\n") ;
        for (i = 0 ; i < n ; i++) fprintf (fbug, "%32.18e\n", y [i]) ;
        fprintf (fbug, "] ;\nd = [\n") ;
        for (i = 0 ; i < n ; i++) fprintf (fbug, "%32.18e\n", d [i]) ;
        fprintf (fbug, "] ;\na = [\n") ;
        for (i = 0 ; i < n ; i++) fprintf (fbug, "%32.18e\n", a [i]) ;
        fprintf (fbug, "] ;\nlo = [\n") ;
        for (i = 0 ; i < n ; i++) fprintf (fbug, "%32.18e\n", lo [i]) ;
        fprintf (fbug, "] ;\nhi = [\n") ;
        for (i = 0 ; i < n ; i++) fprintf (fbug, "%32.18e\n", hi [i]) ;
        fprintf (fbug, "] ;\nxbad = [\n") ;
        for (i = 0 ; i < n ; i++) fprintf (fbug, "%32.18e\n", x [i]) ;
        fprintf (fbug, "] ;\n") ;
        fprintf (fbug, "lambda_bad = %32.18e ;\n", lambda) ;
        fprintf (fbug, "clear parm ;\n") ;
        fprintf (fbug, "parm.refine = %d ;\n",
            (Parm == NULL) ? FALSE : Parm->refine) ;
        fprintf (fbug, "parm.err = %32.18e ;\n",
            (Parm == NULL) ? ZERO : Parm->err) ;
        fprintf (fbug, "parm.newton = %d ;\n",
            (Parm == NULL) ? FALSE : Parm->newton) ;
        fprintf (fbug, "parm.k = %d\n",
            (Parm == NULL) ? 4 : Parm->K) ;
        fprintf (fbug, "guess = %32.18e\n", guess) ;
        fprintf (fbug, "[x lambda stats] = napheap (y, d, a, blo, bhi, lo,"
            " hi, parm") ;
        if (Parm != NULL && Parm->use_lambda)
        {
            fprintf (fbug, ", guess") ;
        }
        fprintf (fbug, ") ;\n") ;
        fprintf (fbug, "napheap_check (x, lambda, y, d, a, blo, bhi, lo,"
            " hi)\n") ;
        fprintf (fbug, "stats\n") ;
        fclose (fbug) ;

        /*
        fwrite (&n, 4, 1, fbug) ;
        fwrite (&Blo, 8, 1, fbug) ;
        fwrite (&Bhi, 8, 1, fbug) ;
        fwrite (y, 8, n, fbug) ;
        fwrite (d, 8, n, fbug) ;
        fwrite (a, 8, n, fbug) ;
        fwrite (lo, 8, n, fbug) ;
        fwrite (hi, 8, n, fbug) ;
        */

        if (f != NULL) fclose (f) ;
        exit (0) ;
    }
}


/* -------------------------------------------------------------------------- */

static int napheap_wrapper
(
    int d_is_zero,      /* if true, pass d = NULL to napheap */
    double        *x,   /* size n. solution (output) */
    double const  *y,   /* size n. linear term in objective function */
    double const  *d,   /* size n. diagonal of objective Hessian */
    double const  *a,   /* size n. linear constraint vector */
    double       Blo,   /* lower bound for a'x */
    double       Bhi,   /* upper bound for a'x */
    double const *lo,   /* size n. lower bounds for x */
    double const *hi,   /* size n. upper bounds for x */
    NAPHEAP_INT    n,   /* problem dimension */
    double   *Lambda,   /* input: starting multiplier guess if parameter
                           use_lambda = TRUE, otherwise ignored on input.
                           output: final multiplier (scalar) */
    napheap_stats *Stats,   /* structure with statistics (can be NULL) */
    napheap_parameter *Parm,/*parameter structure, NULL => use default */

    double *cpu,
    int redo_with_lambda,
    int print
)
{
    double tic ;
    double const *d2 ;
    int status ;

    d2 = d_is_zero ? NULL : d ;

    if (print)
    {
        napheap_print_parms (Parm) ;
    }

    tic = Timer () ;
    status = napheap (x, y, d2, a, Blo, Bhi, lo, hi, n,
        Lambda, Stats, Parm, NULL, NULL) ;
    (*cpu) = Timer () - tic ;

    if (print)
    {
        napheap_print_stats (Stats) ;
    }

    if (status != NAPHEAP_STATUS_OK)
    {
        return (status) ;
    }

    if ( redo_with_lambda && Parm->use_lambda )
    {
        /* rerun the code with lambda as the initial guess */
        /* also pass in user-provided work space, to test that feature */

        NAPHEAP_INT *myiwork ;
        double *myxwork ;

        myxwork = (double      *) malloc ((5*n+1)*sizeof (double)) ;
        myiwork = (NAPHEAP_INT *) malloc ((3*(n+1))*sizeof (NAPHEAP_INT)) ;

        status = napheap (x, y, d2, a, Blo, Bhi, lo, hi, n,
            Lambda, Stats, Parm, myxwork, myiwork) ;

        free (myxwork) ;
        free (myiwork) ;
    }

    return (status) ;
}

/* -------------------------------------------------------------------------- */

int main (void)
{
    double d1, y1min, y1max, a1min, a1max, l1min, l1max ;
    double d2min, d2max, y2min, y2max, a2min, a2max, l2min, l2max ;
    double y3, a3min, a3max, l3min, l3max ;
    double d4, y4min, y4max, a4, l4, u4 ;
    double d5, y5min, y5max, a5min, a5max, l5, u5 ;
    double d6, y6min, y6max, a6min, a6max, l6, u6 ;
    double d7, y7min, y7max, a7min, a7max, l7, u7 ;
    double d8, y8min, y8max, a8min, a8max, l8, u8 ;
    double r, s, t, cumlo, cumhi, l, u ;
    double lambda, cpu ;
    double losave, ysave, dsave, hisave, asave ;
    double *x, *y, *d, *a, *lo, *hi ;
    NAPHEAP_INT i, j, n, *perm ;
    int ntries, refine, status, method, use_lambda, d_is_zero, print,
        set, dim, trial, set_start, set_end, probdim [NDIM+1] ;
    FILE *f = NULL ;
    unsigned int rseed ;
    napheap_parameter Parm ;
    napheap_stats Stats ;

    #ifdef NAPSACK_BRUTAL
    fprintf (stderr, "testing napheap (brutal test), please wait ...\n") ;
    #else
    fprintf (stderr, "testing napheap (normal operation), please wait ...\n") ;
    #endif
    fprintf (stderr, "naptest inf: %g.\n", INF) ;

    for (method = 0 ; method <= 14 ; method++)
    {
        printf ("\n---- method: %d\n", method) ;
        fprintf (stderr, "test: %2d ", method) ;
        napheap_defaults (&Parm) ;

        /* ------------------------------------------------------------------ */
        /* try each variation of the method */
        /* ------------------------------------------------------------------ */

        switch (method)
        {

            case 0:     /* ==> Run/0 <== */
                Parm.newton = FALSE ;
                Parm.K = -1 ;   /* becomes K=0 in napheap */
                set_end = 18 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_0.txt", "w") ;
                #endif
                break ;

            case 1:     /* ==> Run/1 <== */
                Parm.newton = FALSE ;
                Parm.K = 1 ;
                set_end = 18 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_1.txt", "w") ;
                #endif
                break ;

            case 2:     /* ==> Run/2 <== */
                Parm.newton = FALSE ;
                Parm.K = 2 ;
                set_end = 18 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_2.txt", "w") ;
                #endif
                break ;

            case 3:     /* ==> Run/2N <== */
                Parm.newton = TRUE ;
                Parm.K = 2 ;
                set_end = 7 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_2N.txt", "w") ;
                #endif
                break ;

            case 4:     /* ==> Run/3 <== */
                Parm.newton = FALSE ;
                Parm.K = 3 ;
                set_end = 18 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_3.txt", "w") ;
                #endif
                break ;

            case 5:     /* ==> Run/3N <== */
                Parm.newton = TRUE ;
                Parm.K = 3 ;
                set_end = 7 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_3N.txt", "w") ;
                #endif
                break ;

            case 6:     /* ==> Run/4 <== */
                Parm.newton = FALSE ;
                Parm.K = 4 ;
                set_end = 18 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_4.txt", "w") ;
                #endif
                break ;

            case 7:     /* ==> Run/4N <== */
                Parm.newton = TRUE ;
                Parm.K = 4 ;
                set_end = 7 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_4N.txt", "w") ;
                #endif
                break ;

            case 8:     /* ==> Run/5 <== */
                Parm.newton = FALSE ;
                Parm.K = 5 ;
                set_end = 18 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_5.txt", "w") ;
                #endif
                break ;

            case 9:     /* ==> Run/5N <== */
                Parm.newton = TRUE ;
                Parm.K = 5 ;
                set_end = 7 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_5N.txt", "w") ;
                #endif
                break ;

            case 10:     /* ==> Run/10 <== */
                Parm.newton = FALSE ;
                Parm.K = 10 ;
                set_end = 18 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_10.txt", "w") ;
                #endif
                break ;

            case 11:     /* ==> Run/10N <== */
                Parm.newton = TRUE ;
                Parm.K = 10 ;
                set_end = 7 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_11.txt", "w") ;
                #endif
                break ;

            case 12:
                Parm.newton = FALSE ;
                Parm.K = 1000000 ;
                set_end = 18 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_inf.txt", "w") ;
                #endif
                break ;

            case 13:
                Parm.newton = TRUE ;
                Parm.K = 1000000 ;
                set_end = 7 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_infN.txt", "w") ;
                #endif
                break ;

            case 14:     /* ==> Run/1N <== */
                Parm.newton = TRUE ;
                Parm.K = 1 ;
                set_end = 8 ;
                #ifdef DIAGNOSTIC_OUTPUT
                f = fopen ("output_1N.txt", "w") ;
                #endif
                break ;

        }

        /* ------------------------------------------------------------------ */
        /* test the method on a range of problem sets, sizes, and parameters */
        /* ------------------------------------------------------------------ */

        set_start = 1 ;

        /* problem parameters */

        /* problem set 1 */
        d1 = 25. ;
        y1min = -25. ;
        y1max = +25. ;
        a1min = -25. ;
        a1max = +25. ;
        l1min = -15. ;
        l1max = +15. ;

        /* problem set 2 */
        d2min = .5 ;
        d2max = 1.5 ;
        y2min = -5. ;
        y2max = +5. ;
        a2min = -25. ;
        a2max = +25. ;
        l2min = -15. ;
        l2max = +15. ;

        /* problem set 3 */
        y3 = +5. ;
        a3min = -25. ;
        a3max = +25. ;
        l3min = -15. ;
        l3max = +15. ;

        /* problem set 4 */
        d4 = +1. ;
        y4min = -10. ;
        y4max = +10. ;
        a4 = 1. ;
        l4 = 0. ;
        u4 = 1. ;

        /* problem set 5 */
        d5 = +1. ;
        y5min = -10. ;
        y5max = +10. ;
        a5min = 1. ;
        a5max = 26. ;
        l5 = 0. ;
        u5 = 1. ;

        /* for problem sets 6, and 8 through 11 */
        d6 = +25. ;
        y6min = -25. ;
        y6max = +25. ;
        a6min = -10. ;
        a6max = +10. ;
        l6 = -1 ;
        u6 = INF ;

        /* problem set 7 */
        d7 = 1.e-6 ;
        y7min = -25. ;
        y7max = +25. ;
        a7min = -25. ;
        a7max = +25. ;
        l7 = 0. ;
        u7 = 1. ;

        /* for problem sets 8 through 11 */
        d8 = +25. ;
        y8min = -25. ;
        y8max = +25. ;
        a8min = -25. ;
        a8max = +25. ;
        l8 = -1. ;
        u8 = 1. ;

        /*
        i = 10 ;
        for (j = 0; j < NDIM; j++)
        {
            probdim [j] = i ;
            i *= 5 ;
        }
        */

        for (i = 0; i < 50; i++) probdim [i] = i + 1 ;
        probdim [50] = 250 ;
        probdim [51] = 1250 ;
        probdim [52] = 6250 ;

        ntries = 100 ;

        for (set = set_start; set <= set_end; set++)
        {

            fprintf (stderr, ".") ;

            for (dim = 0; dim < NDIM; dim++)
            {

                /* ---------------------------------------------------------- */
                /* allocate space for solution and problem */
                /* ---------------------------------------------------------- */

                n = probdim [dim] ;          /* problem dimension */
                x = (double *) malloc (n*sizeof (double)) ;
                d = (double *) malloc (n*sizeof (double)) ;
                a = (double *) malloc (n*sizeof (double)) ;
                y = (double *) malloc (n*sizeof (double)) ;
                lo = (double *) malloc (n*sizeof (double)) ;
                hi = (double *) malloc (n*sizeof (double)) ;
                perm = (NAPHEAP_INT *) malloc (n*sizeof (NAPHEAP_INT)) ;

                /* make ntries instances */
                for (trial = 0; trial < ntries; trial++)
                {

                    /* print the parameters for the first few trials */
                    print = (dim == 4 || dim == NDIM-1) && trial <= 6 ;

                    rseed = (n*10000) + (set*100) + trial ;

                    /* if ( rseed != 12500790 ) continue ;*/

                    printf ("method: %d set: %3d dim %d n: %g trial: %3d "
                            "seed: %10d\n", method, set, dim, (double) n,
                            trial, rseed) ;
                    fflush (stdout) ;
                    srand (rseed) ;

                    for (j = 0; j < n; j++) perm [j] = j ;
                    cumlo = ZERO ;
                    cumhi = ZERO ;

                    d_is_zero = FALSE ;
                    Parm.d_is_pos = TRUE ;

                    /* ------------------------------------------------------ */
                    /* create the test problem */
                    /* ------------------------------------------------------ */

                    if ( set == 1 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            d [j] = rnumpos (d1) ;
                            y [j] = rnum (y1min, y1max) ;
                            a [j] = rnum (a1min, a1max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            s = rnum (l1min, l1max) ;
                            t = rnum (l1min, l1max) ;
                            lo [j] = MIN (s, t) ;
                            hi [j] = MAX (s, t) ;
                            if (j == 1 && (trial % 5 == 4)) hi [j] = lo [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l ;
                    }
                    else if ( set == 2 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            a [j] = rnum (a2min, a2max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            y [j] = rnum (a[j]+y2min, a[j]+y2max) ;
                            d [j] = fabs (a [j])*rnum (d2min, d2max) ;
                            if ( !d [j] ) d [j] = rnum (d2min, d2max) ;
                            s = rnum (l2min, l2max) ;
                            t = rnum (l2min, l2max) ;
                            lo [j] = MIN (s, t) ;
                            hi [j] = MAX (s, t) ;
                            if (j == 1 && (trial % 3 == 0)) lo [j] = hi [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l ;
                    }
                    else if ( set == 3 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            a [j] = rnum (a3min, a3max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            y [j] = a [j] + y3 ;
                            d [j] = fabs (a [j]) + 1 ;
                            s = rnum (l3min, l3max) ;
                            t = rnum (l3min, l3max) ;
                            lo [j] = MIN (s, t) ;
                            hi [j] = MAX (s, t) ;
                            if (j == 1 && (trial % 3 == 0)) lo [j] = hi [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l ;
                    }
                    else if ( set == 4 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            a [j] = a4 ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            y [j] = rnum (y4min, y4max) ;
                            d [j] = d4 ;
                            lo [j] = l4 ;
                            hi [j] = u4 ;
                            if (j == 1 && (trial % 3 == 0)) lo [j] = hi [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l ;
                    }
                    else if ( set == 5 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            a [j] = floor (rnum (a5min, a5max)) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            y [j] = rnum (y5min, y5max) ;
                            d [j] = d5 ;
                            lo [j] = l5 ;
                            hi [j] = u5 ;
                            if (j == 1 && (trial % 3 == 0)) lo [j] = hi [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l ;
                    }
                    else if ( set == 6 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            d [j] = rnumpos (d6) ;
                            y [j] = rnum (y6min, y6max) ;
                            a [j] = rnum (a6min, a6max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            lo [j] = rnum (l6, -l6) ;
                            hi [j] = u6 ;
                            if (j == 1 && (trial % 3 == 0)) hi [j] = lo [j] ;
                            cumlo += a [j]*lo [j] ;
                        }
                        l = rnum (cumlo, MAX (100., cumlo+10)) ;
                        u = l + 10 ;
                    }
                    else if ( set == 7 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            d [j] = rnumpos (d7) ;
                            y [j] = rnum (y7min, y7max) ;
                            a [j] = rnum (a7min, a7max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            lo [j] = l7 ;
                            hi [j] = u7 ;
                            /*
                            if (j == 1 && (trial % 3 == 0)) lo [j] = hi [j] ;
                            */
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l + 1 ;
                    }
                    else if ( set == 8 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            d [j] = rnumpos (d8) ;
                            y [j] = rnum (y8min, y8max) ;
                            a [j] = rnum (a8min, a8max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            lo [j] = l8 ;
                            hi [j] = u8 ;
                            if (j == 1 && (trial % 3 == 0)) hi [j] = lo [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l + 10. ;
                        randperm (perm, n) ;
                        i = (NAPHEAP_INT) (.25*n) ;
                        for (j = 0; j < i; j++) d [perm [j]] = ZERO ;
                        Parm.d_is_pos = FALSE ;
                    }
                    else if ( set == 9 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            d [j] = rnumpos (d8) ;
                            y [j] = rnum (y8min, y8max) ;
                            a [j] = rnum (a8min, a8max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            lo [j] = l8 ;
                            hi [j] = u8 ;
                            if (j == 1 && (trial % 3 == 0)) hi [j] = lo [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l + 10. ;
                        randperm (perm, n) ;
                        i = (NAPHEAP_INT) (.50*n) ;
                        for (j = 0; j < i; j++) d [perm [j]] = ZERO ;
                        Parm.d_is_pos = FALSE ;
                    }
                    else if ( set == 10 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            d [j] = rnumpos (d8) ;
                            y [j] = rnum (y8min, y8max) ;
                            a [j] = rnum (a8min, a8max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            lo [j] = l8 ;
                            hi [j] = u8 ;
                            if (j == 1 && (trial % 3 == 0)) hi [j] = lo [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l + 10. ;
                        randperm (perm, n) ;
                        i = (NAPHEAP_INT) (.75*n) ;
                        for (j = 0; j < i; j++) d [perm [j]] = ZERO ;
                        Parm.d_is_pos = FALSE ;
                    }
                    else if ( set == 11 ) /* d = 0 */
                    {
                        for (j = 0; j < n; j++)
                        {
                            d [j] = ZERO ;
                            y [j] = rnum (y7min, y7max) ;
                            a [j] = rnum (a7min, a7max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            lo [j] = l7 ;
                            hi [j] = rnum (l7, u7) ;
                            if (j == 1 && (trial % 3 == 0)) hi [j] = lo [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l ;
                        if ( trial % 3 == 0 ) u += 1. ;
                        Parm.d_is_pos = FALSE ;
                        d_is_zero = TRUE ;
                    }
                    else if ( set == 12 ) /* d > 0 and lambda = 0 optimal */
                    {
                        for (j = 0; j < n; j++)
                        {
                            d [j] = rnumpos (1.001) ;
                            y [j] = rnum (y1min, y1max) ;
                            a [j] = rnum (a1min, a1max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            t = y [j]/d [j] ;
                            lo [j] = t - rnum (0., 1.) ;
                            hi [j] = t + rnum (0., 1.) ;
                            if (j == 1 && (trial % 5 == 4))
                            {
                                lo [j] = t + 1. ;
                                hi [j] = t + 1. ;
                                cumlo += (t+1)*a [j] ;
                            }
                            else
                            {
                                cumlo += t*a [j] ;
                            }
                        }
                        l = cumlo ;
                        if ( (j/2)*2 == j )
                        {
                            u = l + 1. ;
                        }
                        else
                        {
                            u = l ;
                        }

                        if ( trial % 23 == 0 )
                        {
                            l = -INF ;
                            u =  INF ;
                        }
                    }
                    else if ( set == 13 ) /* d = 0 and lambda = 0 optimal */
                    {
                        t = 0. ;
                        s = 0. ;
                        for (j = 0; j < n; j++)
                        {
                            d [j] = 0. ;
                            y [j] = rnum (y1min, y1max) ;
                            a [j] = rnum (a1min, a1max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            lo [j] = -rnum (0., 1.) ;
                            hi [j] =  rnum (0., 1.) ;
                            if (j == 3 && (trial % 2 == 0))
                            {
                                y [j] = 0. ;
                                t += MIN (a [j]*lo [j], a [j]*hi [j]) ;
                                s += MAX (a [j]*lo [j], a [j]*hi [j]) ;
                            }
                            if (j == 7 && (trial % 3 == 0))
                            {
                                y [j] = 0. ;
                                t += MIN (a [j]*lo [j], a [j]*hi [j]) ;
                                s += MAX (a [j]*lo [j], a [j]*hi [j]) ;
                            }
                            if      ( y [j] > 0. ) r = hi [j] ;
                            else if ( y [j] < 0. ) r = lo [j] ;
                            else                   r = 0. ;
                            cumlo += r*a [j] ;
                        }
                        d_is_zero = TRUE ;
                        t += cumlo ;
                        s += cumlo ;
                        /* right side must lie in [t, s] */
                        l = rnum (t-1, s) ;
                        u = rnum (t, s+1) ;
                        if ( trial == 15 )
                        {
                            l = -INF ;
                            u =  INF ;
                        }
                    }
                    /* d has 0 and nonzeros, lambda = 0 optimal */
                    else if ( set == 14 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            d [j] = rnumpos (.001) ;
                            y [j] = rnum (y1min, y1max) ;
                            a [j] = rnum (a1min, a1max) ;
                            if (j == 5 && (trial % 2 == 0)) a [j] = 0 ;
                            t = y [j]/d [j] ;
                            lo [j] = t - rnum (0., 1.) ;
                            hi [j] = t + rnum (0., 1.) ;
                            if (j == 1 && (trial % 5 == 4))
                            {
                                lo [j] = t + 1. ;
                                hi [j] = t + 1. ;
                            }
                        }
                        randperm (perm, n) ;
                        i = (NAPHEAP_INT) (.50*n) ;
                        for (j = 0; j < i; j++) d [perm [j]] = ZERO ;
                        Parm.d_is_pos = FALSE ;
                        t = 0. ;
                        s = 0. ;
                        for (j = 0; j < n; j++)
                        {
                            if ( d [j] == 0. )
                            {
                                if (j == 3 && (trial % 2 == 0))
                                {
                                    y [j] = 0. ;
                                    t += MIN (a [j]*lo [j], a [j]*hi [j]) ;
                                    s += MAX (a [j]*lo [j], a [j]*hi [j]) ;
                                }
                                if (j == 7 && (trial % 3 == 0))
                                {
                                    y [j] = 0. ;
                                    t += MIN (a [j]*lo [j], a [j]*hi [j]) ;
                                    s += MAX (a [j]*lo [j], a [j]*hi [j]) ;
                                }
                                if      ( y [j] > 0. ) r = hi [j] ;
                                else if ( y [j] < 0. ) r = lo [j] ;
                                else                   r = 0. ;
                                cumlo += r*a [j] ;
                            }
                            else
                            {
                                cumlo += (y [j]/d [j])*a [j] ;
                            }
                        }
                        t += cumlo ;
                        s += cumlo ;
                        /* right side must lie in [t, s] */
                        l = rnum (t-1, s) ;
                        u = rnum (t, s+1) ;
                        if ( trial == 15 )
                        {
                            l = -INF ;
                            u =  INF ;
                        }
                    }
                    /* d = 0 and domain of dual function is a point */
                    else if ( set == 15 )
                    {
                        for (j = 0; j < n; j++)
                        {
                            d [j] = ZERO ;
                            y [j] = rnum (y7min, y7max) ;
                            a [j] = rnum (a7min, a7max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            lo [j] = l7 ;
                            hi [j] = rnum (l7, u7) ;
                            if (j == 1 && (trial % 3 == 0)) hi [j] = lo [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        if ( n > 4 )
                        {
                            j = n/2 ;
                            t = y [j]/a [j] ;
                            for (j = 0; j < n; j++)
                            {
                                if ( a [j] != ZERO )
                                {
                                    s = y [j]/a [j] ;
                                    if ( s > t+.5 )
                                    {
                                        if ( a [j] > ZERO )
                                        {
                                            lo [j] = -INF ;
                                        }
                                        else
                                        {
                                            hi [j] =  INF ;
                                        }
                                    }
                                    else if ( s < t-.5 )
                                    {
                                        if ( a [j] < ZERO )
                                        {
                                            lo [j] = -INF ;
                                        }
                                        else
                                        {
                                            hi [j] =  INF ;
                                        }
                                    }
                                }
                            }
                            if ( trial % 7 == 0 )
                            {
                                j = n/2 ;
                                lo [j] = -INF ;
                                hi [j] =  INF ;
                            }
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l ;
                        if ( trial % 2 == 0 ) u += 1. ;
                        Parm.d_is_pos = FALSE ;
                        if ( trial % 3 ) d_is_zero = TRUE ;
                    }
                    else if ( set == 16 ) /* run set 1 but duplicate data */
                    {
                        i = (n/2) + 1 ;
                        for (j = 0; j < i; j++)
                        {
                            d [j] = rnumpos (d1) ;
                            y [j] = rnum (y1min, y1max) ;
                            a [j] = rnum (a1min, a1max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            s = rnum (l1min, l1max) ;
                            t = rnum (l1min, l1max) ;
                            lo [j] = MIN (s, t) ;
                            hi [j] = MAX (s, t) ;
                            if (j == 1 && (trial % 5 == 4)) hi [j] = lo [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        i = 0 ;
                        for (; j < n; j++)
                        {
                            d [j] = d [i] ;
                            y [j] = y [i] ;
                            a [j] = a [i] ;
                            lo [j] = lo [i] ;
                            hi [j] = hi [i] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                            i++ ;
                        }
                        l = rnum (cumlo, cumhi) ;
                        u = l ;
                        if ( trial % 2 == 0 ) u += 1 ;
                    }
                    else if ( set == 17 )
                    {
                        i = (n/2) + 1 ;
                        for (j = 0; j < i; j++)
                        {
                            d [j] = ZERO ;
                            y [j] = rnum (y7min, y7max) ;
                            a [j] = rnum (a7min, a7max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            lo [j] = l7 ;
                            hi [j] = rnum (l7, u7) ;
                            if (j == 1 && (trial % 3 == 0)) hi [j] = lo [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        if ( n > 4 )
                        {
                            j = n/4 ;
                            t = y [j]/a [j] ;
                            for (j = 0; j < i; j++)
                            {
                                if ( trial % 3 == 0 )
                                {
                                    if ( j % 2 ) d [j] = rnumpos (d1) ;
                                }
                                if ( a [j] != ZERO )
                                {
                                    s = y [j]/a [j] ;
                                    if ( s > t+.5 )
                                    {
                                        if ( a [j] > ZERO )
                                        {
                                            lo [j] = -INF ;
                                        }
                                        else
                                        {
                                            hi [j] =  INF ;
                                        }
                                    }
                                    else if ( s < t-.5 )
                                    {
                                        if ( a [j] < ZERO )
                                        {
                                            lo [j] = -INF ;
                                        }
                                        else
                                        {
                                            hi [j] =  INF ;
                                        }
                                    }
                                }
                            }
                            if ( trial % 4 == 0 )
                            {
                                lo [n/4] = -INF ;
                                hi [n/4] =  INF ;
                                d [n/4] = ZERO ;
                            }
                        }
                        i = 0 ;
                        for (; j < n; j++)
                        {
                            d [j] = d [i] ;
                            y [j] = y [i] ;
                            a [j] = a [i] ;
                            lo [j] = lo [i] ;
                            hi [j] = hi [i] ;
                            i++ ;
                        }
                        if ( (trial % 2 == 0) && (n > 4) )
                        {
                            if ( (trial/2)%3 == 0 )
                            {
                                lo [n/4] = rnum (-1, +1) ;
                            }
                            else if ( (trial/2)%3 == 1 )
                            {
                                hi [n/4] = rnum (-1, +1) ;
                            }
                            else
                            {
                                lo [n/4] = rnum (-1, +1) ;
                                hi [n/4] = rnum (lo [n/4], +1) ;
                            }
                        }
                        l = rnum (2*cumlo, 2*cumhi) ;
                        u = l ;
                        if ( trial % 3 == 0 ) u += 1. ;
                        Parm.d_is_pos = FALSE ;
                    }
                    else if ( set == 18 )
                    {
                        i = (n/2) + 1 ;
                        for (j = 0; j < i; j++)
                        {
                            d [j] = ZERO ;
                            y [j] = rnum (y7min, y7max) ;
                            a [j] = rnum (a7min, a7max) ;
                            if (j == 0 && (trial % 2 == 0)) a [j] = 0 ;
                            lo [j] = l7 ;
                            hi [j] = rnum (l7, u7) ;
                            if (j == 1 && (trial % 3 == 0)) hi [j] = lo [j] ;
                            t = a [j]*lo [j] ;
                            s = a [j]*hi [j] ;
                            cumlo += MIN (t, s) ;
                            cumhi += MAX (t, s) ;
                        }
                        if ( n > 4 )
                        {
                            j = n/4 ;
                            t = y [j]/a [j] ;
                            for (j = 0; j < i; j++)
                            {
                                if ( a [j] != ZERO )
                                {
                                    s = y [j]/a [j] ;
                                    if ( s > t+.5 )
                                    {
                                        if ( a [j] > ZERO )
                                        {
                                            lo [j] = -INF ;
                                        }
                                        else
                                        {
                                            hi [j] =  INF ;
                                        }
                                    }
                                    else if ( s < t-.5 )
                                    {
                                        if ( a [j] < ZERO )
                                        {
                                            lo [j] = -INF ;
                                        }
                                        else
                                        {
                                            hi [j] =  INF ;
                                        }
                                    }
                                }
                            }
                            if ( trial % 3 == 0 )
                            {
                                lo [n/4] = -INF ;
                                hi [n/4] =  INF ;
                            }
                        }
                        i = 0 ;
                        for (; j < n; j++)
                        {
                            d [j] = d [i] ;
                            y [j] = y [i] ;
                            a [j] = a [i] ;
                            lo [j] = lo [i] ;
                            hi [j] = hi [i] ;
                            i++ ;
                        }
                        if ( (trial % 3 == 0) && (n > 4) )
                        {
                            if ( (trial/3)%3 == 0 )
                            {
                                lo [n/4] = rnum (-1, +1) ;
                            }
                            else if ( (trial/3)%3 == 1 )
                            {
                                hi [n/4] = rnum (-1, +1) ;
                            }
                            else
                            {
                                lo [n/4] = rnum (-1, +1) ;
                                hi [n/4] = rnum (lo [n/4], +1) ;
                            }
                        }
                        l = rnum (2*cumlo, 2*cumhi) ;
                        u = l ;
                        if ( trial % 2 == 0 ) u += 1. ;
                        Parm.d_is_pos = FALSE ;
                        d_is_zero = TRUE ;
                    }

                    /* ------------------------------------------------------ */
                    /* try the test problem 4 times */
                    /* ------------------------------------------------------ */

                    /* with/without refinement, and with/without input lambda */

                    for (refine = 0 ; refine <= 1 ; refine++)
                    {
                        Parm.refine = refine ;

                        for (use_lambda = 0 ; use_lambda <= 1 ; use_lambda++)
                        {
                            double guess ;
                            Parm.use_lambda = use_lambda ;

                            lambda = rnum (-10., +10.) ;
                            guess = lambda ;

                            /* ---------------------------------------------- */
                            /* run the code */
                            /* ---------------------------------------------- */

                            status = napheap_wrapper (d_is_zero,
                                x, y, d, a, l, u, lo, hi, n, &lambda,
                                &Stats, &Parm, &cpu, trial % 34 == 0, print) ;

                            check_solution (d_is_zero,
                                x, y, d, a, l, u, lo, hi, n, guess, lambda,
                                &Stats, &Parm, cpu, status,
                                refine, set, method, trial, f) ;
                        }
                    }
                    printf ("\n") ;
                }

                /* ---------------------------------------------------------- */
                /* test error handling and alternate methods */
                /* ---------------------------------------------------------- */

                printf ("\n--- Testing error handling and special cases: [\n") ;
                losave = lo [0] ;
                ysave = y [0] ;
                dsave = d [0] ;
                hisave = hi [0] ;
                asave = a [0] ;

                Parm.use_lambda = 0 ;
                Parm.refine = 0 ;

                /* default parameters, no stats  */
                printf ("try defaults:\n") ;
                status = napheap_wrapper (d_is_zero,
                    x, y, d, a, l, u, lo, hi, n, &lambda,
                    NULL, NULL, &cpu, 0, print) ;
                check_solution (d_is_zero,
                    x, y, d, a, l, u, lo, hi, n, 0., lambda,
                    NULL, NULL, cpu, status,
                    0, set, method, trial, f) ;

                if (set == 12 || set == 13 || set == 14)
                {

                    /* unbounded case */
                    printf ("set %d, unbounded:\n", set) ;
                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, -INF, INF, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    check_solution (d_is_zero,
                        x, y, d, a, -INF, INF, lo, hi, n, 0., lambda,
                        &Stats, &Parm, cpu, status,
                        0, set, method, trial, f) ;

                    /* now try mangling the lo and hi */
                    printf ("\nset %d, unbounded, lo/hi mangled:\n", set) ;
                    lo [0] = hi [0] + 1 ;
                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, -INF, INF, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    check_status (status, NAPHEAP_STATUS_INVALID_BOUNDS) ;

                    /* try mangling the lo and hi, bounded case  */
                    printf ("\nset %d, lo/hi mangled:\n", set) ;
                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, l, u, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    check_status (status, NAPHEAP_STATUS_INVALID_BOUNDS) ;

                    /* try b unbounded, lo with INFs, but y zero */
                    printf ("\nset %d, b inf, lo/hi infs:\n", set) ;
                    lo [0] = -INF ;
                    y [0] = 0 ;
                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, -INF, INF, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    check_status (status, NAPHEAP_STATUS_OK) ;

                    y [0] = ysave ;
                    lo [0] = losave ;
                    hi [0] = hisave ;

                    /* now try mangling d */
                    if (!d_is_zero)
                    {
                        printf ("\nset %d, unbounded, d mangled:\n", set) ;
                        d [0] = -1 ;
                        status = napheap_wrapper (d_is_zero,
                            x, y, d, a, -INF, INF, lo, hi, n, &lambda,
                            &Stats, &Parm, &cpu, 0, print) ;
                        napheap_print_stats (&Stats) ;
                        check_status (status, NAPHEAP_STATUS_INVALID_D) ;
                        d [0] = dsave ;

                        /* try b unbounded, lo with INFs, but y zero */
                        printf ("\nset %d, unbounded, hi infs:\n", set) ;
                        d [0] = 0 ;
                        hi [0] = INF ;
                        y [0] = 1 ;
                        status = napheap_wrapper (d_is_zero,
                            x, y, d, a, -INF, INF, lo, hi, n, &lambda,
                            &Stats, &Parm, &cpu, 0, print) ;
                        napheap_print_stats (&Stats) ;
                        check_status (status, NAPHEAP_STATUS_UNBOUNDED) ;

                        y [0] = ysave ;
                        lo [0] = losave ;
                        hi [0] = hisave ;
                        d [0] = dsave ;

                        /* try b unbounded, lo with INFs, but y zero */
                        printf ("\nset %d, unbounded, lo infs:\n", set) ;
                        d [0] = 0 ;
                        lo [0] = -INF ;
                        y [0] = -1 ;
                        status = napheap_wrapper (d_is_zero,
                            x, y, d, a, -INF, INF, lo, hi, n, &lambda,
                            &Stats, &Parm, &cpu, 0, print) ;
                        napheap_print_stats (&Stats) ;
                        check_status (status, NAPHEAP_STATUS_UNBOUNDED) ;

                        y [0] = ysave ;
                        lo [0] = losave ;
                        hi [0] = hisave ;
                        d [0] = dsave ;

                    }
                }

                /* n is invalid */
                printf ("check n\n") ;
                status = napheap_wrapper (d_is_zero,
                    x, y, d, a, l, u, lo, hi, -1, &lambda,
                    &Stats, NULL, &cpu, 0, print) ;
                napheap_print_stats (&Stats) ;
                check_status (status, NAPHEAP_STATUS_INVALID_N) ;

                /* n is 0, which is OK */
                printf ("n is zero\n") ;
                status = napheap_wrapper (d_is_zero,
                    x, y, d, a, l, u, lo, hi, 0, &lambda,
                    &Stats, NULL, &cpu, 0, print) ;
                napheap_print_stats (&Stats) ;
                check_status (status, NAPHEAP_STATUS_OK) ;

                /* args are invalid */
                printf ("check arg\n") ;
                status = napheap_wrapper (d_is_zero,
                    NULL, y, d, a, l, u, lo, hi, n, &lambda,
                    &Stats, NULL, &cpu, 0, print) ;
                napheap_print_stats (&Stats) ;
                check_status (status, NAPHEAP_STATUS_INVALID_ARG) ;

                /* test invalid Newton parameters */
                if (!(Parm.d_is_pos) && Parm.K > 1)
                {
                    printf ("check newton\n") ;
                    int newt = Parm.newton ;
                    Parm.newton = TRUE ;
                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, l, u, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    check_status (status, NAPHEAP_STATUS_INVALID_NEWTON) ;
                    Parm.newton = newt ;
                }

                /* mangle the problem: d is invalid */
                if (!d_is_zero)
                {
                    printf ("d negative: %d %d\n", d_is_zero, Parm.d_is_pos) ;
                    d [0] = -1 ;
                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, l, u, lo, hi, n, &lambda,
                        &Stats, NULL, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    check_status (status, NAPHEAP_STATUS_INVALID_D) ;

                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, l, u, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    check_status (status, NAPHEAP_STATUS_INVALID_D) ;

                    if (Parm.d_is_pos)
                    {

                        printf ("d said to be positive but d < 0:\n") ;
                        d [0] = 0 ;
                        status = napheap_wrapper (d_is_zero,
                            x, y, d, a, l, u, lo, hi, n, &lambda,
                            &Stats, &Parm, &cpu, 0, print) ;
                        check_status (status, NAPHEAP_STATUS_INVALID_D) ;
                        napheap_print_stats (&Stats) ;
                        d [0] = dsave ;

                    }
                    else
                    {

                        printf ("d general, d = 0, unbounded hi:\n") ;
                        d [0] = 0 ;
                        a [0] = 0 ;
                        y [0] = 1 ;
                        hi [0] = INF ;
                        status = napheap_wrapper (d_is_zero,
                            x, y, d, a, l, u, lo, hi, n, &lambda,
                            &Stats, &Parm, &cpu, 0, print) ;
                        check_status (status, NAPHEAP_STATUS_UNBOUNDED) ;
                        napheap_print_stats (&Stats) ;

                        d [0] = dsave ;
                        y [0] = ysave ;
                        lo [0] = losave ;
                        hi [0] = hisave ;
                        a [0] = asave ;

                        printf ("d general, d = 0, unbounded lo:\n") ;
                        d [0] = 0 ;
                        a [0] = 0 ;
                        y [0] = -1 ;
                        lo [0] = -INF ;
                        status = napheap_wrapper (d_is_zero,
                            x, y, d, a, l, u, lo, hi, n, &lambda,
                            &Stats, &Parm, &cpu, 0, print) ;
                        check_status (status, NAPHEAP_STATUS_UNBOUNDED) ;
                        napheap_print_stats (&Stats) ;

                        d [0] = dsave ;
                        y [0] = ysave ;
                        lo [0] = losave ;
                        hi [0] = hisave ;
                        a [0] = asave ;

                    }
                }

                /* check d = 0 case, using Parm.d_is_zero parameter */
                if (d_is_zero)
                {
                    int d0save = Parm.d_is_zero ;
                    printf ("d zero:\n") ;
                    Parm.d_is_zero = TRUE ;
                    napheap_print_parms (&Parm) ;
                    status = napheap_wrapper (FALSE,
                        x, y, d, a, l, u, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    check_solution (FALSE,
                        x, y, d, a, l, u, lo, hi, n, 0., lambda,
                        &Stats, &Parm, cpu, status,
                        0, set, method, trial, f) ;
                    napheap_print_stats (&Stats) ;
                    Parm.d_is_zero = d0save ;

                    /* try b unbounded, lo with INFs, but y zero */
                    printf ("\nd zero, set %d, unbounded, hi infs:\n", set) ;
                    hi [0] = INF ;
                    y [0] = 1 ;
                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, -INF, INF, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    check_status (status, NAPHEAP_STATUS_UNBOUNDED) ;

                    y [0] = ysave ;
                    lo [0] = losave ;
                    hi [0] = hisave ;

                    /* try b unbounded, lo with INFs, but y zero */
                    printf ("\nd zero, set %d, unbounded, lo infs:\n", set) ;
                    lo [0] = -INF ;
                    y [0] = -1 ;
                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, -INF, INF, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    check_status (status, NAPHEAP_STATUS_UNBOUNDED) ;

                    y [0] = ysave ;
                    lo [0] = losave ;
                    hi [0] = hisave ;

                    /* try b unbounded, lo with INFs, but y zero */
                    printf ("\nd zero, set %d, unbounded, lo inf hi<0:\n",set) ;
                    lo [0] = -INF ;
                    y [0] = 0 ;
                    hi [0] = -1 ;
                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, -INF, INF, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    if (! (status == NAPHEAP_STATUS_UNBOUNDED ||
                           status == NAPHEAP_STATUS_OK) )
                    {
                        printf ("got status %d, should be %d or %d\n",
                            status, NAPHEAP_STATUS_UNBOUNDED,
                            NAPHEAP_STATUS_OK) ;
                        fprintf (stderr, "TEST FAILURE\n") ;
                        exit (0) ;
                    }

                    y [0] = ysave ;
                    lo [0] = losave ;
                    hi [0] = hisave ;

                    /* try b bounded, a zero, hi unbounded */
                    printf ("\nd zero, set %d, bounded, a zero, hi inf:\n",set);
                    a [0] = 0 ;
                    y [0] = 1 ;
                    hi [0] = INF ;
                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, l, u, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    check_status (status, NAPHEAP_STATUS_UNBOUNDED) ;

                    /* try b bounded, a zero, lo unbounded */
                    printf ("\nd zero, set %d, bounded, a zero, lo inf:\n",set);
                    a [0] = 0 ;
                    y [0] = -1 ;
                    lo [0] = -INF ;
                    status = napheap_wrapper (d_is_zero,
                        x, y, d, a, l, u, lo, hi, n, &lambda,
                        &Stats, &Parm, &cpu, 0, print) ;
                    napheap_print_stats (&Stats) ;
                    check_status (status, NAPHEAP_STATUS_UNBOUNDED) ;

                    y [0] = ysave ;
                    lo [0] = losave ;
                    hi [0] = hisave ;
                    a [0] = asave ;

                }

                printf ("\ndone Testing error handling and special cases ]\n") ;

                /* ---------------------------------------------------------- */
                /* free the problem */
                /* ---------------------------------------------------------- */

                free (x) ;
                free (d) ;
                free (a) ;
                free (y) ;
                free (lo) ;
                free (hi) ;
                free (perm) ;
            }
        }
        fprintf (stderr, " OK\n") ;
        if (f != NULL) fclose (f) ;
        f = NULL ;
    }

    printf ("\nnapheap default parameters:\n") ;
    napheap_print_parms (NULL) ;

    fprintf (stderr, "All tests passed\n") ;
    printf ("All tests passed\n") ;
    return (0) ;
}
