/* ========================================================================== */
/* === Source/napheap.c ===================================================== */
/* ========================================================================== */

/*
       ________________________________________________________________
      |Solve a separable convex quadratic knapsack problem of the form:|
      |                                                                |
      | min .5x'Dx - y'x  subject to  lo <= x <= hi, blo <= a'x <= bhi |
      |                                                                |
      |where lo and hi are vectors, blo and bhi are scalars, and D is a|
      |           a diagonal matrix with nonnegative diagonal.         |
      |                                                                |
      |                 Copyright by Timothy A. Davis                  |
      |                              William W. Hager                  |
      |                              James T. Hungerford               |
      |                                                                |
      |          http://www.math.ufl.edu/~hager/papers/Software        |
      |                                                                |
      |  Disclaimer: The views expressed are those of the authors and  |
      |              do not reflect the official policy or position of |
      |              the Department of Defense or the U.S. Government. |
      |                                                                |
      |      Approved for Public Release, Distribution Unlimited       |
      |________________________________________________________________|

    The constraints are rewritten as

               lo <= x <= hi,    blo <= b <= bhi,    a'x = b

    The code solves the dual problem obtained by introducing
    a multiplier lambda for the constraint a'x = b.  The dual function is

    L(lambda) = inf {.5x'Dx - y'x + lambda (a'x - b):
                                 lo <= x <= hi, blo <= b <= bhi}

    The dual function is concave, and the solution to the original (primal)
    problem can be constructed from the maximizer of the dual function.
    If blo < bhi, then there is a discontinuity at lambda = 0.  D is a
    diagonal matrix, and if d = diag(D) contains zeros, there could be
    additional points of discontinuity.  The algorithm iterates on lambda
    until 0 lies in the subdifferential of L(lambda).  The optimal x is
    constructed from the minimizers of the dual function.  Three algorithms
    are implemented: a break point searching algorithm, a variable fixing
    algorithm, and a Newton/secant algorithm.  A starting guess for the
    optimal multiplier can be provided.  If the user does not have a good
    starting guess, Parm->K (default 4) iterations of a Newton-type
    algorithm are used to generate a starting guess.  By default,
    the variable fixing algorithm is used for the startup. Set
    Parm->newton to true to choose Newton's method for startup. After
    the startup iterations, the breakpoint algorithm is used to
    compute the final solution. When Newton's method is used for the
    startup, all entries of d must be positive (otherwise an error is
    returned).  No value of d can be negative, for any method.

    The user can either provide work arrays or let the code malloc work arrays.
    The necessary size for the work arrays depends on the choice of the
    algorithm, as can be seen in the memory allocation routine.  The most
    memory that the code will require is an integer work array of size 3*(n+1)
    and a real work array of size 5*n where n is the problem dimension.

    The default parameter values used by the software are given in the
    napheap_defaults routine.  To alter the default parameter values, the
    user can first initialize the parameters to their default values using
    napheap_defaults, make any changes to the values in the parameter
    structure, and provide the updated parameter structure as an input to the
    napheap code.

    The code records the number of iterations for either the variable fixing or
    Newton method, and the number of break points that are passed through
    during the solution process.  These statistics are returned when the user
    provides a nonnull structure for the statistics argument of the code.  The
    statistics and parameters can be printed with napheap_print_stats and
    napheap_print_parms.

    If the user wishes to override the default definition of infinity
    given in Include/napheap.h, modify the OPTFLAGS as explained in
    Lib/napheap.mk.
       ________________________________________________________________
      |This program is free software; you can redistribute it and/or   |
      |modify it under the terms of the GNU General Public License as  |
      |published by the Free Software Foundation; either version 2 of  |
      |the License, or (at your option) any later version.             |
      |This program is distributed in the hope that it will be useful, |
      |but WITHOUT ANY WARRANTY; without even the implied warranty of  |
      |MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   |
      |GNU General Public License for more details.                    |
      |                                                                |
      |You should have received a copy of the GNU General Public       |
      |License along with this program; if not, write to the Free      |
      |Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, |
      |MA  02110-1301  USA                                             |
      |________________________________________________________________|

      Reference:
         T. A. Davis, W. W. Hager, and J. T. Hungerford, The separable
         convex quadratic knapsack problem, Doc/napheap.pdf
*/

#include "napheap.h"

#ifdef MATLAB_MEX_FILE
#define malloc mxMalloc
#define free mxFree
#endif

#define FALSE 0
#define TRUE 1
#ifndef NULL
#define NULL 0
#endif

/* true infinity (ANSI C99) or DBL_MAX (ANSI C90) */
#define INF NAPHEAP_INF

#define PRIVATE static
#define ZERO ((double) 0)
#define ONE ((double) 1)
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))

/* additional status value, not returned to the user */
#define NAPHEAP_STATUS_KEEP_LOOKING     (-1)

/* disable debuging, even if requested at the compile command */
#ifndef NDEBUG
#define NDEBUG
#endif

/* uncomment this line to enable debugging
#undef NDEBUG
*/

#include <assert.h>

/* -------------------------------------------------------------------------- */
/* prototypes of non-user callable functions */
/* -------------------------------------------------------------------------- */

PRIVATE void napsearch
(
    double        *x,    /* size n. solution (output) */
    double   *lambda,    /* optimal multiplier (input = guess, output = sol) */
    double   lambdaL,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double   lambdaR,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double         B,    /* a'x = B */
    double    *slope,    /* derivative of dual function */
    double const  *y,    /* size n. linear term in objective function */
    double const  *d,    /* size n, diagonal of objective Hessian */
    double const  *a,    /* size n, linear constraint vector */
    double const *lo,    /* size n, lower bounds on x */
    double const *hi,    /* size n, upper bounds on x */
    NAPHEAP_INT    n,    /* problem dimension */
    NAPHEAP_INT    K,    /* number of variable fixing iterations */
    double *breakpts,    /* size n, breakpts [j] is br_hi[j] or br_lo[j] */
    NAPHEAP_INT *next,   /* size n+1, next[-1..n-1] link list of unbound */
    NAPHEAP_INT *free_heap,  /* size n+1, heap of free variables */
    NAPHEAP_INT *bound_heap, /* size n+1, heap of bound variables */
    napheap_stats *stats,    /* solution statistics */
    double const *br_lo, /* size n, br_lo[j] lo break pt for x[j] */
    double const *br_hi, /* size n. br_hi[j] hi break pt for x[j] */
    napheap_parameter *Parm /* parameters */
) ;

PRIVATE void napsearch0
(
    double        *x,    /* size n. solution (output) */
    double   *lambda,    /* optimal multiplier (input = guess, output = sol) */
    double   lambdaL,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double   lambdaR,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double        *b,    /* a'x = b */
    double    *slope,    /* slope excluding term associated with break point*/
    double  *slopelo,    /* lower subdifferential associated with break point */
    double  *slopehi,    /* upper subdifferential associated with break point */
    NAPHEAP_INT    n,    /* problem dimension */
    double const  *a,    /* size n, linear constraint vector */
    double const *lo,    /* size n, lower bounds on x */
    double const *hi,    /* size n, upper bounds on x */
    double const *breakpts, /* size n, breakpts [j] is br_hi[j] or br_lo[j] */
    NAPHEAP_INT *next,   /* size n+1, next[-1..n-1] link list of unbound var. */
    NAPHEAP_INT *bound_heap,    /* size n+1, heap of bound variables */
    int      goright,    /* if true, move right, else move left */
    napheap_stats *stats,    /* solution statistics */
    int recompute_slope, /* if true, then recompute slope */
    napheap_parameter *Parm  /* parameters */
) ;

PRIVATE void napsearch1
(
    double        *x,    /* size n. solution (output) */
    double   *lambda,    /* optimal multiplier (input = guess, output = sol) */
    double   lambdaL,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double   lambdaR,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double        *b,    /* a'x = b */
    double    *slope,    /* slope excluding term associated with break point*/
    double  *slopelo,    /* lower subdifferential associated with break point */
    double  *slopehi,    /* upper subdifferential associated with break point */
    double const  *y,    /* size n. linear term in objective function */
    double const  *d,    /* size n, diagonal of objective Hessian */
    double const  *a,    /* size n, linear constraint vector */
    double const *lo,    /* size n, lower bounds on x */
    double const *hi,    /* size n, upper bounds on x */
    NAPHEAP_INT    n,    /* problem dimension */
    NAPHEAP_INT    K,    /* number iterations in variable fixing method */
    double *breakpts,    /* size n, breakpts [j] is br_hi[j] or br_lo[j] */
    NAPHEAP_INT *next,   /* size n+1, next[-1..n-1] link list of unbound var. */
    NAPHEAP_INT *free_heap,     /* size n+1, heap of free variables */
    NAPHEAP_INT *bound_heap,    /* size n+1, heap of bound variables */
    int      goright,    /* if true, move right, else move left */
    napheap_stats *stats,    /* solution statistics */
    double const *br_lo, /* size n, br_lo[j] lo break pt for x[j] */
    double const *br_hi, /* size n. br_hi[j] hi break pt for x[j] */
    int recompute_slope, /* if true, then recompute slope */
    napheap_parameter *Parm  /* parameters */
) ;

PRIVATE void napfix0
(
    double        *x, /* size n. solution (output) */
    double const  *a, /* size n, linear constraint vector */
    double const *lo, /* size n, lower bounds on x */
    double const *hi, /* size n, upper bounds on x */
    NAPHEAP_INT    n, /* problem dimension */
    double        *b, /* b - ai*xi for fixed variables */
    double    lambda, /* optimal multiplier */
    NAPHEAP_INT *next,/* size n+1, next[-1..n-1] list of unbound var. */
    double const *breakpts /* break points */
) ;

PRIVATE void napfix1
(
    double        *x, /* size n. solution (output) */
    double const  *a, /* size n, linear constraint vector */
    double const  *y, /* size n. linear term in objective function */
    double const  *d, /* size n, diagonal of objective Hessian */
    double const *lo, /* size n, lower bounds on x */
    double const *hi, /* size n, upper bounds on x */
    NAPHEAP_INT    n, /* problem dimension */
    double        *b, /* b - ai*xi for fixed variables */
    double    lambda, /* optimal multiplier */
    NAPHEAP_INT *next,/* size n+1, next[-1..n-1] list of unbound var. */
    double const *breakpts /* break points */
) ;

PRIVATE void napsolution
(
    double       *x,  /* size n. solution (output) */
    double const *y,  /* size n. linear term in objective function */
    double const *d,  /* size n, diagonal of objective Hessian */
    double const *a,  /* size n, linear constraint vector */
    double const *lo, /* size n, lower bounds on x */
    double const *hi, /* size n, upper bounds on x */
    NAPHEAP_INT    n, /* problem dimension */
    double    lambda, /* optimal multiplier */
    NAPHEAP_INT const *next  /* size n+1, next[-1..n-1] link list of unbound */
) ;

PRIVATE void napsolution0
(
    double        *x,    /* size n. solution (output) */
    double const  *a,    /* size n, linear constraint vector */
    double const *lo,    /* size n, lower bounds on x */
    double const *hi,    /* size n, upper bounds on x */
    NAPHEAP_INT    n,    /* problem dimension */
    double         B,    /* the constraint is A'X = B */
    double *breakpts,    /* size n, breakpts [j] is br_hi[j] or br_lo[j] */
    NAPHEAP_INT *next,   /* size n+1, next[-1..n-1] link list of unbound var */
    NAPHEAP_INT *free_heap,/* size n+1, heap of free variables */
    NAPHEAP_INT *bound_heap/* size n+1, heap of bound variables */
) ;

PRIVATE void napminheap_build
(
    NAPHEAP_INT *heap,  /* on input, unsorted set of indices.  size nheap+1 */
    double const *X,    /* numbers to sort */
    NAPHEAP_INT nheap   /* number of elements to build into the heap */
) ;

PRIVATE NAPHEAP_INT napminheap_delete  /* return new size of heap */
(
    NAPHEAP_INT *heap,  /* indices into X, 1..n on input.  size nheap+1 */
    double const *X,    /* not modified */
    NAPHEAP_INT nheap   /* number of items in heap */
) ;

PRIVATE NAPHEAP_INT napminheap_add
(
    NAPHEAP_INT *heap,  /* size n, containing indices into X.  size nheap+2 */
    double const *X,    /* not modified */
    NAPHEAP_INT  leaf,  /* the new leaf */
    NAPHEAP_INT nheap   /* number of elements in heap not counting new one */
) ;

PRIVATE void napminheapify
(
    NAPHEAP_INT *heap,  /* size nheap+1, containing indices into X */
    double const *X,    /* not modified */
    NAPHEAP_INT     p,  /* start at node p in the heap */
    NAPHEAP_INT nheap   /* heap [1 ... n] is in use */
) ;

PRIVATE void napmaxheap_build
(
    NAPHEAP_INT *heap,  /* on input, unsorted set of elements. size nheap+1 */
    double const *X,    /* not modified */
    NAPHEAP_INT nheap   /* number of elements to build into the heap */
) ;

PRIVATE NAPHEAP_INT napmaxheap_delete  /* return new size of heap */
(
    NAPHEAP_INT *heap,  /* containing indices into X, size nheap+1 */
    double const *X,    /* not modified */
    NAPHEAP_INT nheap   /* number of items in heap */
) ;

PRIVATE NAPHEAP_INT napmaxheap_add
(
    NAPHEAP_INT *heap,  /* size nheap+2, containing indices into X */
    double const *X,    /* not modified */
    NAPHEAP_INT  leaf,  /* the new leaf */
    NAPHEAP_INT nheap   /* number of elements in heap not counting new one */
) ;

PRIVATE void napmaxheapify
(
    NAPHEAP_INT *heap,  /* size nheap+1, containing indices into X */
    double const *X,    /* not modified */
    NAPHEAP_INT     p,  /* start at node p in the heap */
    NAPHEAP_INT nheap   /* heap [1 ... nheap] is in use */
) ;

PRIVATE void nap_infeasible
(
    double        *x,   /* size n. solution (output) */
    double const  *a,   /* size n. linear constraint vector */
    double const *lo,   /* size n. lower bounds for x */
    double const *hi,   /* size n. upper bounds for x */
    NAPHEAP_INT    n,   /* problem dimension */
    NAPHEAP_INT *next,  /* size n+1, next[-1..n-1]: link list of unbound var */
    double minax,
    double maxax,
    double blo,         /* lower bound for a'x */
    double bhi          /* upper bound for a'x */
) ;

PRIVATE int nap_wrapup  /* returns status from napheap to user */
(
    int status,
    napheap_stats *Stats,
    double *xw2,        /* workspace, freed if not NULL */
    NAPHEAP_INT *iw2,   /* workspace, freed if not NULL */
    double *Lambda,     /* returns lambda from napheap to user */
    double lambda
) ;

PRIVATE int nap_work
(
    NAPHEAP_INT n,
    double *User_Work,          /* user-provide workspace */
    double **xw2,               /* return malloc'ed space */
    double **ada,               /* the 5 work arrays */
    double **ady,
    double **br_hi,
    double **br_lo,
    double **breakpts
) ;

/* ========================================================================== */
/* === user-callable functions ============================================== */
/* ========================================================================== */

/* There are only four user-callable functions in this file:
    napheap_defaults    sets default parameters for napheap
    napheap             solves a separable convex quadratic knapsack problem
*/

/* ========================================================================== */
/* === napheap_defaults ===================================================== */
/* ========================================================================== */

/*  Set default parameter values for the napheap code. If the parameter
    argument of napheap is NULL, this routine is called automatically to
    set parameters to their default value.  If the user wishes to set
    parameter values, then the napheap_parameter structure should be declared
    in the main program. The user could call napheap_defaults to initialize
    the structure, and then individual elements in the structure
    could be changed, before passing the structure to napheap.
*/

void napheap_defaults
(
    napheap_parameter *Parm
)
{
    /* T => perform an iterative refinement step to improve solution accurary
       F => no refinement */
    Parm->refine = FALSE ;

    /* The solution is refined if |a'x - b| > err */
    Parm->err = ZERO ;

    /* T => use the Newton to update lambda when K > 1
       F => use variable fixing to update lambda when K > 1.
       If K = 0 or 1, this parameter is ignored (Newton not used) */
    Parm->newton = FALSE ;

    /* T => use input lambda as a starting guess for multiplier
       F => ignore input lambda (final lambda is returned in lambda argument) */
    Parm->use_lambda = FALSE ;

    /* T => diagonal of objective Hessian is positive
       F => there could be zeros on the diagonal of the Hessian
       Note: If d = 0, then the d argument of napheap should be NULL */
    Parm->d_is_pos = TRUE ;

    /* T => diagonal of objective Hessian is zero
       F => there could be nonzeros on the diagonal of the Hessian
            Note: If d = 0, then the d argument of napheap can be NULL */
    Parm->d_is_zero = FALSE ;

    /* Number of Newton or variable fixing iterations.  After performing these
       K iterations, the code switches to the break point searching algorithm.
       K = 4 seems to perform well for a wide range of problems. */
    Parm->K = 4 ;
}

/* ========================================================================== */
/* === napheap ============================================================== */
/* ========================================================================== */

int napheap /* return status of solution process:
                   0 (no errors detected)
                   1 (infeasible, return x closest to feasible set)
                   2 (primal is -infinity)
                   3 (out of memory)
                   4 (n < 0)
                   5 (lo [j] > hi [j] for some j)
                   6 (Newton's method chosen with K > 1, but d_is_pos FALSE )
                   7 (Either d [j] < 0 or d [j] = 0 and d_is_pos TRUE) */
(
    double        *x,   /* size n. solution (output) */
    double const  *y,   /* size n. linear term in objective function */
    double const  *d,   /* size n. diagonal of objective Hessian */
    double const  *a,   /* size n. linear constraint vector */
    double       Blo,   /* lower bound for a'x */
    double       Bhi,   /* upper bound for a'x */
    double const *lo,   /* size n. lower bounds for x */
    double const *hi,   /* size n. upper bounds for x */
    NAPHEAP_INT    n,   /* problem dimension */

    double   *Lambda,   /* input: starting multiplier guess if parameter
                                  use_lambda = TRUE, otherwise ignored on input.
                           output: final multiplier (scalar) */

    napheap_stats *UStats,  /* structure with statistics (can be NULL) */

    napheap_parameter *UParm, /*parameter structure, NULL => use defaults */

    double *Work,       /* NULL => let code allocate double memory
                       not NULL => use array Work for double memory.
                           The amount of memory needed depends on the
                           parameters in the Parm structure. See the
                           memory allocation section below for the precise
                           memory requirement. */

    NAPHEAP_INT *iWork    /* NULL => let code allocate NAPHEAP_INT memory
                         not NULL => use array iWork for NAPHEAP_INT memory. */
)
{
    NAPHEAP_INT it, j, k, nsecant, *iwork ;
    int newton ;     /* TRUE if newton iteration performed */
    int use_lambda ; /* TRUE if starting lambda is provided */
    int status ;     /* returned status of solution process */
    int varfix ;     /* TRUE if variable fixing performed */
    int secant ;     /* TRUE if the Newton iteration is a secant iteration */
    int goright ;    /* TRUE if optimal multiplier to right of current lambda */
    int found ;      /* TRUE if variable fixing was able to fix a variable */
    int Rout ;       /* FALSE if optimal lambda in interior of current interval
                        [lambdaL, lambdaR] */
    int Lout ;       /* FALSE if optimal lambda in interior of current interval
                        [lambdaL, lambdaR] */
    int d_is_zero ;  /* TRUE if d is all zero (input d is ignored) */
    int d_is_pos ;   /* TRUE if d is all positive */
    int K ;          /* number of iterations in Newton's or variable fixing
                        methods */
    int k_gt_1 ;     /* TRUE if K > 1 */
    int recompute_slope ; /* TRUE if slope at current lambda not known */
    double adj, aj, ajaj, ajy, dj, maxax, minax, hij, loj, xj, yj, r, s, t ;
    double *ada = NULL ;      /* size n. ada[j] == (a[j]^2)/d[j] */
    double *ady = NULL ;      /* size n. ady[j] == a[j]*y[j]/d[j] */
    double *breakpts = NULL ; /* size n */
    double *br_hi = NULL ;    /* size n. br_hi[j] hi break pt for x[j] */
    double *br_lo = NULL ;    /* size n, br_lo[j] lo break pt for x[j] */
    double blo ;              /* lower bound for a'x */
    double bhi ;              /* upper bound for a'x */
    double b ;                /* a'x = b */
    double domlo ;            /* lower bound for domain of dual function */
    double domhi ;            /* upper bound for domain of dual function */
    double slope ;            /* slope at lambda excluding subdiffertial term */
    double slopelo ;          /* lower subdifferential at lambda */
    double slopehi ;          /* upper subdifferential at lambda */
    double slope0 ;           /* slope at lambda = 0 excluding subdiff term */
    double slope0lo ;         /* lower subdifferential at lambda = 0 */
    double slope0hi ;         /* upper subdifferential at lambda = 0 */
    double lambda ;           /* multiplier estimate */
    double lambdaL ;          /* lower bound for optimal multiplier */
    double lambdaR ;          /* upper bound for optimal multiplier */
    double dL ;               /* slope at lambdaL */
    double dR ;               /* slope at lambdaR */
    double num ;              /* numerator of variable fixing ratio for lambda*/
    double den ;              /* denominator of variable fixing ratio */

    napheap_parameter *Parm, ParmStruc ;
    napheap_stats *Stats, Private_Stats_struct ;

    /* breakpts [j] is defined only when j is in the list of undetermined
       variables (the 'next' link list).  For the general case, breakpts [j] is
       equal to either br_lo [j] or br_hi [j].  When d is zero or when K = 0,
       breakpts [j] is computed as needed.  */

    NAPHEAP_INT *next = NULL ;          /* size n+1, next[-1..n-1] is a link
                                           list of unbound variables */
    NAPHEAP_INT *free_heap = NULL ;     /* size n+1, heap of free variables */
    NAPHEAP_INT *bound_heap = NULL ;    /* size n+1, heap of bound variables */
    NAPHEAP_INT *iw2 = NULL ;
    double *xw2 = NULL ;

    /* ====================================================================== */
    /* === initializations ================================================== */
    /* ====================================================================== */

    /* initialize the statistics structure */
    if ( UStats == NULL)
    {
        Stats = &Private_Stats_struct ;
    }
    else
    {
        Stats = UStats ;
    }

    status = NAPHEAP_STATUS_OK ;
    Stats->status = status ;
    Stats->jerror = 0 ;
    Stats->lobad = 0 ;
    Stats->hibad = 0 ;
    Stats->dbad = 0 ;
    Stats->nfree = 0 ;
    Stats->nbound = 0 ;
    Stats->nbrks = 0 ;
    Stats->nrefine = 0 ;
    Stats->nvarfix = 0 ;
    Stats->nnewton = 0 ;
    Stats->nsecant = 0 ;

    /* initialize the parameters */
    if ( UParm == NULL )
    {
        Parm = &ParmStruc ;
        napheap_defaults (Parm) ;
    }
    else
    {
        Parm = UParm ;
    }

    d_is_zero = FALSE ;
    if ( d == NULL )
    {
        d_is_zero = TRUE ;
    }
    else if ( Parm->d_is_zero == TRUE )
    {
        d_is_zero = TRUE ;
    }

    lambda = ZERO ; /* always return a value for lambda */

    /* check if the user flipped blo and bhi */
    blo = Blo ;
    bhi = Bhi ;
    if (blo > bhi)
    {
        /* interchange blo and bhi */
        t = blo ;
        blo = bhi ;
        bhi = t ;
    }

    /* check for basic errors */
    if (n < 0)
    {
        status = NAPHEAP_STATUS_INVALID_N ;
        Stats->jerror = n ;
        return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
    }
    else if (n == 0)
    {
        /* empty problem, do nothing */
        status = NAPHEAP_STATUS_OK ;
        return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
    }

    /* make sure x, y, a, lo, hi, Lambda are not NULL (user may have run
       out of memory but didn't check before calling napheap) */
    if ( x == NULL || y == NULL || a == NULL || lo == NULL || hi == NULL
        || Lambda == NULL )
    {
        status = NAPHEAP_STATUS_INVALID_ARG ;
        return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
    }

    use_lambda = Parm->use_lambda ;
    newton = Parm->newton ;

    if ( d_is_zero )
    {
        /* d is all zero, so d_is_pos = all (d > 0) is false */
        d_is_pos = FALSE ;
    }
    else
    {
        /* get the status of all (d > 0) from Parm, without checking d.
           This is just to save time. */
        d_is_pos = Parm->d_is_pos ;
    }

    K = Parm->K ;
    if ( K < 0 )
    {
        /* silently treat this as K = 0 */
        K = 0 ;
    }

    /* Newton reduces to variable fixing when K = 1 */
    if ( K <= 1 )
    {
        /* so disable Newton's method when K <= 1 */
        newton = FALSE ;
    }

    if ( !d_is_pos && newton )
    {
        /* Newton's method with K > 1 is requested, but it requires all(d>0).
           This is an error. */
        status = NAPHEAP_STATUS_INVALID_NEWTON ;
        return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
    }

    if ( use_lambda ) lambda = *Lambda ;

    recompute_slope = FALSE ;

    /* ---------------------------------------------------------------------- */
    /* check if the equality constraint can be ignored */
    /* ---------------------------------------------------------------------- */

    if ( (blo <= -INF) && (bhi >= INF) )
    {
        lambda = ZERO ;
        if ( d_is_zero )
        {
            for (j = 0; j < n; j++)
            {
                if ( lo [j] > hi [j] )
                {
                    /* invalid problem */
                    status = NAPHEAP_STATUS_INVALID_BOUNDS ;
                    Stats->jerror = j ;
                    Stats->lobad = lo [j] ;
                    Stats->hibad = hi [j] ;
                    return (nap_wrapup (status, Stats,
                        xw2, iw2, Lambda, lambda)) ;
                }
                if ( y [j] > ZERO )
                {
                    if ( hi [j] >= INF )
                    {
                        status = NAPHEAP_STATUS_UNBOUNDED ;
                    }
                    x [j] = hi [j] ;
                }
                else if ( y [j] < ZERO )
                {
                    if ( lo [j] <= -INF )
                    {
                        status = NAPHEAP_STATUS_UNBOUNDED ;
                    }
                    x [j] = lo [j] ;
                }
                else /* choose any value between loj and hij */
                {
                    if ( lo [j] <= -INF )
                    {
                        if ( hi [j] >= ZERO ) x [j] = ZERO ;
                        else                  x [j] = hi [j] ;
                    }
                    else x [j] = lo [j] ;
                }
            }
        }
        else
        {
            for (j = 0; j < n; j++)
            {
                if ( lo [j] > hi [j] )
                {
                    /* invalid problem */
                    status = NAPHEAP_STATUS_INVALID_BOUNDS ;
                    Stats->jerror = j ;
                    Stats->lobad = lo [j] ;
                    Stats->hibad = hi [j] ;
                    return (nap_wrapup (status, Stats,
                        xw2, iw2, Lambda, lambda)) ;
                }
                if ( d [j] < ZERO )
                {
                    status = NAPHEAP_STATUS_INVALID_D ;
                    Stats->jerror = j ;
                    Stats->dbad = d [j] ;
                    return (nap_wrapup (status, Stats,
                        xw2, iw2, Lambda, lambda)) ;
                }
                else if ( d [j] > ZERO )
                {
                    t = y [j]/d [j] ;
                    if      ( t < lo [j] ) t = lo [j] ;
                    else if ( t > hi [j] ) t = hi [j] ;
                    x [j] = t ;
                }
                else    /* d [j] is zero */
                {
                    if ( y [j] > ZERO )
                    {
                        if ( hi [j] >= INF )
                        {
                            status = NAPHEAP_STATUS_UNBOUNDED ;
                        }
                        x [j] = hi [j] ;
                    }
                    else if ( y [j] < ZERO )
                    {
                        if ( lo [j] <= -INF )
                        {
                            status = NAPHEAP_STATUS_UNBOUNDED;
                        }
                        x [j] = lo [j] ;
                    }
                    else /* choose any value between loj and hij */
                    {
                        if ( lo [j] <= -INF )
                        {
                            if ( hi [j] >= ZERO ) x [j] = ZERO ;
                            else x [j] = hi [j] ;
                        }
                        else x [j] = lo [j] ;
                    }
                }
            }
        }
        /* status is NAPHEAP_STATUS_OK or NAPHEAP_STATUS_UNBOUNDED */
        assert (status == NAPHEAP_STATUS_OK
             || status == NAPHEAP_STATUS_UNBOUNDED ) ;
        return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
    }

    /* ---------------------------------------------------------------------- */
    /* allocate integer workspace */
    /* ---------------------------------------------------------------------- */

    if ( iWork == NULL )
    {
        iw2 = (NAPHEAP_INT *) malloc ((3*(n+1))*sizeof (NAPHEAP_INT)) ;
        /* Need space for next, free_heap, and bound_heap.
           The arrays have size n+1: next [-1], next [0], ..., next [n-1],
           and heap [0], heap [1], ... , heap [n] */
        if ( iw2 == NULL )
        {
            status = NAPHEAP_STATUS_OUT_OF_MEMORY ;
            return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
        }
        iwork = iw2 ;
    }
    else
    {
        /* use the user-provided workspace, iWork */
        iwork = iWork ;
    }

    /* assign pointers to integer work arrays */
    next = iwork+1 ;
    bound_heap = next+n ;
    free_heap = bound_heap+(n+1) ;

    /* ---------------------------------------------------------------------- */
    /* allocate double workspace */
    /* ---------------------------------------------------------------------- */

    if ( K == 0 ) /* breakpoint search */
    {
        status = nap_work (n, Work, &xw2, NULL, NULL, NULL,  NULL,  &breakpts) ;
    }
    else if ( newton ) /* Newton's method */
    {
        status = nap_work (n, Work, &xw2, &ada, NULL, &br_hi,&br_lo,&breakpts) ;
    }
    else if ( (K == 1) && d_is_pos ) /* variable fixing and K = 1 */
    {
        status = nap_work (n, Work, &xw2, NULL, NULL, &br_hi,&br_lo,&breakpts) ;
    }
    else if ( !d_is_zero ) /* d != 0 */
    {
        status = nap_work (n, Work, &xw2, &ada, &ady, &br_hi,&br_lo,&breakpts) ;
    }
    else if ( K > 1 ) /* d_is_zero and K > 1 */
    {
        status = nap_work (n, Work, &xw2, &ada, &ady, NULL,  NULL,  &breakpts) ;
    }
    else    /* d_is_zero and K = 1 */
    {
        status = nap_work (n, Work, &xw2, NULL, NULL, NULL,  NULL,  &breakpts) ;
    }
    if (status != NAPHEAP_STATUS_OK)
    {
        return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
    }

    /* ---------------------------------------------------------------------- */
    /* remaining initializations */
    /* ---------------------------------------------------------------------- */

    /* initialize variables for feasibility testing */
    maxax = ZERO ; /* stores hi range of possible values for a'x */
    minax = ZERO ; /* stores lo range of possible values for a'x */

    domhi = INF ;  /* upper bound on domain where dual max achieved */
    domlo = -INF ; /* lower bound on domain where dual max achieved */

    /* initialize other variables */
    slope = ZERO ; /* slope of dual function ignoring break points at lambda */
    slope0 = ZERO ;/* slope at lambda = 0 */
    b = ZERO ;     /* adjustments to right side due to fixed variables */
    k = -1 ;       /* next [-1] is pointer to first nonfixed variable */
    lambdaL = +INF ;   /* most negative break point */
    lambdaR = -INF ;   /* most positive break point */
    dL = +INF ;        /* value of slope at lambdaL */
    dR = -INF ;        /* value of slope at lambdaR */
    num = ZERO ;       /* numerator of lambda estimate in variable fixing */
    den = ZERO ;       /* denominator of lambda estimate in variable fixing */

    /* ====================================================================== */
    /* === solve the knapsack problem (3 cases) ============================= */
    /* ====================================================================== */

    if ( d_is_pos )
    {

        /* ================================================================== */
        /* === case when all(d>0) holds ===================================== */
        /* ================================================================== */

        for (j = 0; j < n; j++)
        {
            dj = d [j] ;
            if ( dj <= 0 )
            {
                status = NAPHEAP_STATUS_INVALID_D ;
                Stats->jerror = j ;
                Stats->dbad = d [j] ;
                return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
            }
            aj = a [j] ;
            yj = y [j] ;
            loj = lo [j] ;
            hij = hi [j] ;

            if ( loj >= hij )
            {
                if ( loj == hij )
                {
                    t = aj*loj ;
                    b -= t ;
                    x [j] = loj ;
                }
                else /* lo [j] > hi [j], bogus problem */
                {
                    status = NAPHEAP_STATUS_INVALID_BOUNDS ;
                    Stats->jerror = j ;
                    Stats->lobad = lo [j] ;
                    Stats->hibad = hi [j] ;
                    return (nap_wrapup (status, Stats, xw2, iw2,
                        Lambda, lambda)) ;
                }
            }
            else if ( aj == ZERO )
            {
                xj = y [j]/dj ;
                if      ( xj < lo [j] ) xj = loj ;
                else if ( xj > hi [j] ) xj = hij ;
                x [j] = xj ;
            }
            else
            {
                if ( aj > ZERO )
                {
                    maxax += aj*hij ;
                    minax += aj*loj ;

                    /* also store max and min break points */
                    t = (yj - loj*dj)/aj ;
                    if ( t > lambdaR )
                    {
                        if ( loj > -INF ) lambdaR = t ;
                        else              lambdaR = INF ;
                    }
                    t = (yj - hij*dj)/aj ;
                    if ( t < lambdaL )
                    {
                        if ( hij < INF ) lambdaL = t ;
                        else             lambdaL = -INF ;
                    }
                }
                else
                {
                    maxax += aj*loj ;
                    minax += aj*hij ;
                    /* also store max and min break points */
                    t = (yj - hij*dj)/aj ;
                    if ( t > lambdaR )
                    {
                        if ( hij < INF ) lambdaR = t ;
                        else             lambdaR = INF ;
                    }
                    t = (yj - loj*dj)/aj ;
                    if ( t < lambdaL )
                    {
                        if ( loj > -INF ) lambdaL = t ;
                        else              lambdaL = -INF ;
                    }
                }

                /* slope at lambda */
                if ( use_lambda )
                {
                    t = (yj - lambda*aj)/dj ;
                    if      ( t < loj ) t = loj ;
                    else if ( t > hij ) t = hij ;
                    slope += aj*t ;
                }

                /* slope at lambda = 0 */
                yj /= dj ; /* yj now stores yj/dj */
                if      ( yj < loj ) yj = loj ;
                else if ( yj > hij ) yj = hij ;
                slope0 += aj*yj ;

                next [k] = j ;
                k = j ;
            }
        }
        next [k] = n ;

        if ( (minax - b > bhi) || (maxax - b < blo) ) /* problem infeasible */
        {
            nap_infeasible (x, a, lo, hi, n, next, minax, maxax, blo, bhi) ;
            status = NAPHEAP_STATUS_INFEASIBLE ;
            return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
        }

        /* set rhs b */
        if ( slope0 < blo + b ) /* optimal multiplier < 0 */
        {
            b += blo ;
            if ( lambdaL > -INF ) dL = maxax - b ;
            else                  dL = INF ;
            if ( lambdaR > ZERO )
            {
                lambdaR = ZERO ;
                dR = slope0 - b ;
            }
            else dR = minax - b ;
        }
        else if ( slope0 > bhi + b ) /* optimal multiplier > 0 */
        {
            b += bhi ;
            if ( lambdaR < INF ) dR = minax - b ;
            else                 dR = -INF ;
            if ( lambdaL < ZERO )
            {
                lambdaL = ZERO ;
                dL = slope0 - b ;
            }
            else dL = maxax - b ;
        }
        else   /* lambda = 0 is optimal, blo < a'x < bhi */
        {
            lambda = ZERO ;
            j = -1 ;
            while ( (j = next [j]) < n )
            {
                t = y [j]/d [j] ;
                if      ( t < lo [j] ) t = lo [j] ;
                else if ( t > hi [j] ) t = hi [j] ;
                x [j] = t ;
            }
            status = NAPHEAP_STATUS_OK ;
            return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
        }

        /* b has been set, update slope */
        slope0 -= b ;

        /* if lambda is given, check to see whether lambda = 0 is a
           better starting point */
        if ( use_lambda )
        {
            slope -= b ;
            if ( slope > ZERO )
            {
                if ( dL < slope )
                {
                    lambda = lambdaL ;
                    slope = dL ;
                }
                else
                {
                    lambdaL = lambda ;
                    dL = slope ;
                }
            }
            else if ( slope < ZERO )
            {
                if ( dR > slope )
                {
                    lambda = lambdaR ;
                    slope = dR ;
                }
                else
                {
                    lambdaR = lambda ;
                    dR = slope ;
                }
            }
            else                          /* given lambda is optimal */
            {
                napsolution (x, y, d, a, lo, hi, n, lambda, next) ;
                assert (status == NAPHEAP_STATUS_OK) ;
                return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
            }
        }
        else if ( K == 0 ) /* if no lambda given and K = 0, use smaller slope */
        {
            if ( dL < -dR )
            {
                lambda = lambdaL ;
                slope = dL ;
            }
            else
            {
                if ( lambdaR > -INF )
                {
                    lambda = lambdaR ;
                    slope = dR ;
                }
                else K = 1 ; /* starting guess lambda = -INF not allowed */
            }
        }

        /* search the break points if K = 0 */
        if ( K == 0 )
        {

            #ifdef NAPSACK_BRUTAL
            /* !!!!!!!!! mess up the slope for debugging !!!!!!!!!! */
            /* This is only for extensive debugging, not for production use. */
            if ( Parm->refine && n > 3 )
            {
                if ( slope > 0 )
                {
                    if ( a [3] > 0 ) slope += 1 ;
                    else             slope /= 2 ;
                }
                else
                {
                    if ( a [3] > 0 ) slope -= 1 ;
                    else             slope /= 2 ;
                }
            }
            #endif

            /* br_lo and br_hi not used when K == 0 */
            napsearch (x, &lambda, lambdaL, lambdaR, b, &slope,
                                   y, d, a, lo, hi, n, K, breakpts, next,
                                   free_heap, bound_heap, Stats, br_lo,
                                   br_hi, Parm) ;
            napsolution (x, y, d, a, lo, hi, n, lambda, next) ;
            assert (status == NAPHEAP_STATUS_OK) ;
            return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
        }

        k = -1 ;
        j =  k ;
        if ( newton )
        {
            if ( use_lambda )
            {
                /* lambda is given, evaluate br_hi, br_lo, and ada */
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    dj = d [j] ;
                    yj = y [j] ;
                    if ( aj > ZERO )
                    {
                        loj = lo [j] ;
                        r = (yj - loj*dj)/aj ; /* high break point */
                        if ( r <= lambdaL )    /* j is fixed */
                        {
                            b -= aj*loj ;
                            x [j] = loj ;
                            continue ;
                        }

                        hij = hi [j] ;
                        s = (yj - hij*dj)/aj ; /* low break point */
                        if ( s >= lambdaR )    /* j is fixed */
                        {
                            b -= aj*hij ;
                            x [j] = hij ;
                            continue ;
                        }
                        br_hi [j] = r ;
                        br_lo [j] = s ;
                    }
                    else
                    {
                        loj = lo [j] ;
                        r = (yj - loj*dj)/aj ; /* low break point */
                        if ( r >= lambdaR )    /* j is fixed */
                        {
                            b -= aj*loj ;
                            x [j] = loj ;
                            continue ;
                        }

                        hij = hi [j] ;
                        s = (yj - hij*dj)/aj ; /* high break point */
                        if ( s <= lambdaL )    /* j is fixed */
                        {
                            b -= aj*hij ;
                            x [j] = hij ;
                            continue ;
                        }
                        br_hi [j] = s ;
                        br_lo [j] = r ;
                    }

                    /* j not fixed */
                    t = aj*aj/dj ;    /* aj*aj/dj */
                    ada [j] = t ;
                    den += t ;
                    next [k] = j ;
                    k = j ;
                }
            }
            else
            {
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    dj = d [j] ;
                    yj = y [j] ;
                    if ( aj > ZERO )
                    {
                        t = lo [j] ;
                        r = (yj - t*dj)/aj ; /* high break point */
                        if ( r <= lambdaL )   /* j is fixed */
                        {
                            b -= aj*t ;
                            x [j] = t ;
                            continue ;
                        }

                        t = hi [j] ;
                        s = (yj - t*dj)/aj ; /* low break point */
                        if ( s >= lambdaR )   /* j is fixed */
                        {
                            b -= aj*t ;
                            x [j] = t ;
                            continue ;
                        }
                        br_hi [j] = r ;
                        br_lo [j] = s ;
                    }
                    else
                    {
                        t = lo [j] ;
                        r = (yj - t*dj)/aj ; /* low break point */
                        if ( r >= lambdaR )  /* j is fixed */
                        {
                            b -= aj*t ;
                            x [j] = t ;
                            continue ;
                        }

                        t = hi [j] ;
                        s = (yj - t*dj)/aj ; /* high break point */
                        if ( s <= lambdaL )  /* j is fixed */
                        {
                            b -= aj*t ;
                            x [j] = t ;
                            continue ;
                        }
                        br_hi [j] = s ;
                        br_lo [j] = r ;
                    }

                    /* j not fixed */
                    r = aj/dj ;
                    t = r*yj ;    /* aj*yj/dj */
                    num += t ;
                    t = aj*r ;    /* aj*aj/dj */
                    ada [j] = t ;
                    den += t ;
                    next [k] = j ;
                    k = j ;
                }
                num -= b ;
            }
            next [k] = j ;

            /* perform up to K iterations of Newton's method */
            status = NAPHEAP_STATUS_KEEP_LOOKING ; /* default status */
            nsecant = 0 ;               /* count number of secant iterations */
            for (it = 1; it <= K; it++)
            {
                secant = FALSE ;
                varfix = FALSE ;
                if ( den == ZERO ) /* switch to secant method */
                {
                    /* express the secant update relative to the side with slope
                       of smallest magnitude */
                    if ( dL < -dR )
                    {
                        lambda = lambdaL + dL*(lambdaL - lambdaR)/(dR-dL) ;
                    }
                    else
                    {
                        lambda = lambdaR + dR*(lambdaL - lambdaR)/(dR-dL) ;
                    }
                    secant = TRUE ;
                    nsecant++ ;
                }
                else if ( it == 1 && !use_lambda )
                {
                    lambda = num/den ;
                    varfix = TRUE ;
                }
                else lambda += slope/den ;

                /* use secant step if lambda outside [lambdaL, lambdaR] */
                if ( (lambda > lambdaR) || (lambda < lambdaL) )
                {
                    secant = TRUE ;
                    nsecant++ ;
                    /* express the secant update relative to the side with slope
                       of smallest magnitude */
                    if ( dL < -dR )
                    {
                        lambda = lambdaL + dL*(lambdaL - lambdaR)/(dR-dL) ;
                    }
                    else
                    {
                        lambda = lambdaR + dR*(lambdaL - lambdaR)/(dR-dL) ;
                    }
                }

                /* evaluate slope at lambda */
                slope = ZERO ;
                j = -1 ;
                den = ZERO ;
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    t = (y [j] - lambda*aj)/d [j] ;
                    if      ( t < lo [j] ) t = lo [j] ;
                    else if ( t > hi [j] ) t = hi [j] ;
                    else  den += ada [j] ;
                    slope += aj*t ;
                }
                slope -= b ;
                k = -1 ;
                j =  k ;
                found = FALSE ;
                if ( slope > ZERO )     /* optimal multiplier is to the right */
                {
                    if ( lambdaL < lambda )
                    {
                        lambdaL = lambda ;
                        dL = slope ;
                    }
                    else
                    {
                        lambda = lambdaL ;
                        slope = dL ;
                    }
                    if ( it == K ) break ;
                    while ( (j = next [j]) < n )
                    {
                        if ( br_hi [j] <= lambda ) /* fix xj */
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            b -= aj*t ;
                            x [j] = t ;
                            found = TRUE ;
                        }
                        else
                        {
                            next [k] = j ;
                            k = j ;
                        }
                    }
                }
                else if ( slope < ZERO ) /* optimal multiplier is to the left */
                {
                    if ( lambdaR > lambda )
                    {
                        lambdaR = lambda ;
                        dR = slope ;
                    }
                    else
                    {
                        lambda = lambdaR ;
                        slope = dR ;
                    }
                    if ( it == K ) break ;
                    while ( (j = next [j]) < n )
                    {
                        if ( br_lo [j] >= lambda ) /* fix xj */
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            b -= aj*t ;
                            x [j] = t ;
                            found = TRUE ;
                        }
                        else
                        {
                            next [k] = j ;
                            k = j ;
                        }
                    }
                }
                else     /* lambda is optimal */
                {
                    status = NAPHEAP_STATUS_OK ;
                    break ;
                }
                next [k] = j ;
                if ( (found == FALSE) || (k == -1) )
                {
                    if ( secant )
                    {
                        /* breakpoint algorithm */
                        status = NAPHEAP_STATUS_KEEP_LOOKING ;
                    }
                    else if ( varfix || (k == -1) )
                    {
                        /* optimal solution found */
                        status = NAPHEAP_STATUS_OK ;
                    }
                    else
                    {
                        /* breakpoint algorithm */
                        status = NAPHEAP_STATUS_KEEP_LOOKING ;
                    }
                    break ;
                }
            }
            Stats->nnewton = it - nsecant ;
            Stats->nsecant = nsecant ;
            if ( status == NAPHEAP_STATUS_OK )
            {
                napsolution (x, y, d, a, lo, hi, n, lambda, next) ;
            }
        }
        else /* variable fixing algorithm */
        {
            /* If K = 1, then perform 1 iteration of variable fixing
               to get a starting guess for break point searching algorithm.
               No need to store ada or ady */
            if ( K == 1 )
            {
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    dj = d [j] ;
                    yj = y [j] ;
                    if ( aj > ZERO )
                    {
                        loj = lo [j] ;
                        r = (yj - loj*dj)/aj ; /* high break point */
                        if ( r < lambdaL )     /* j is fixed */
                        {
                            b -= aj*loj ;
                            x [j] = loj ;
                            continue ;
                        }

                        hij = hi [j] ;
                        s = (yj - hij*dj)/aj ; /* low break point */
                        if ( s >= lambdaR )    /* j is fixed */
                        {
                            b -= aj*hij ;
                            x [j] = hij ;
                            continue ;
                        }
                        br_hi [j] = r ;
                        br_lo [j] = s ;
                    }
                    else
                    {
                        loj = lo [j] ;
                        r = (yj - loj*dj)/aj ; /* low break point */
                        if ( r > lambdaR )     /* j is fixed */
                        {
                            b -= aj*loj ;
                            x [j] = loj ;
                            continue ;
                        }

                        hij = hi [j] ;
                        s = (yj - hij*dj)/aj ; /* high break point */
                        if ( s < lambdaL )     /* j is fixed */
                        {
                            b -= aj*hij ;
                            x [j] = hij ;
                            continue ;
                        }
                        br_hi [j] = s ;
                        br_lo [j] = r ;
                    }

                    /* j not fixed */
                    num += aj*yj/dj ;
                    den += aj*aj/dj ;
                    next [k] = j ;
                    k = j ;
                }
            }
            else /* K > 1, save break points, ada, and ady */
            {
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    dj = d [j] ;
                    yj = y [j] ;
                    if ( aj > ZERO )
                    {
                        t = lo [j] ;
                        r = (yj - t*dj)/aj ; /* high break point */
                        if ( r < lambdaL )     /* j is fixed */
                        {
                            b -= aj*t ;
                            x [j] = t ;
                            continue ;
                        }

                        t = hi [j] ;
                        s = (yj - t*dj)/aj ; /* low break point */
                        if ( s > lambdaR )     /* j is fixed */
                        {
                            b -= aj*t ;
                            x [j] = t ;
                            continue ;
                        }
                        br_hi [j] = r ;
                        br_lo [j] = s ;
                    }
                    else
                    {
                        t = lo [j] ;
                        r = (yj - t*dj)/aj ; /* low break point */
                        if ( r > lambdaR )     /* j is fixed */
                        {
                            b -= aj*t ;
                            x [j] = t ;
                            continue ;
                        }

                        t = hi [j] ;
                        s = (yj - t*dj)/aj ; /* high break point */
                        if ( s < lambdaL )    /* j is fixed */
                        {
                            b -= aj*t ;
                            x [j] = t ;
                            continue ;
                        }
                        br_hi [j] = s ;
                        br_lo [j] = r ;
                    }

                    /* j not fixed */
                    r = aj/dj ;
                    t = r*yj ;    /* aj*yj/dj */
                    ady [j] = t ;
                    num += t ;
                    t = aj*r ;    /* aj*aj/dj */
                    ada [j] = t ;
                    den += t ;
                    next [k] = j ;
                    k = j ;
                }
            }
            next [k] = j ;

            /* Perform up to K iterations of the variable fixing algorithm */
            status = NAPHEAP_STATUS_KEEP_LOOKING ; /* default status */
            for (it = 1; it <= K; it++)
            {
                /* evaluate new lambda and the slope */
                lambda  = (num-b)/den ;
                slope = ZERO ;
                j = -1 ;
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    t = (y [j] - lambda*aj)/d [j] ;
                    if      ( t < lo [j] ) t = lo [j] ;
                    else if ( t > hi [j] ) t = hi [j] ;
                    slope += aj*t ;
                }
                slope -= b ;

                k = -1 ;
                j =  k ;
                found = FALSE ;

                if ( slope > ZERO )     /* optimal multiplier is to the right */
                {
                    if ( lambdaL < lambda )
                    {
                        lambdaL = lambda ;
                        dL = slope ;
                    }
                    else
                    {
                        lambda = lambdaL ;
                        slope = dL ;
                    }
                    if ( it == K ) break ;
                    /* fix variables and compute next step */
                    while ( (j = next [j]) < n )
                    {
                        if ( br_hi [j] <= lambda ) /* fix xj */
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            b -= aj*t ;
                            x [j] = t ;
                            num -= ady [j] ;
                            den -= ada [j] ;
                            found = TRUE ;
                        }
                        else
                        {
                            next [k] = j ;
                            k = j ;
                        }
                    }
                }
                else if ( slope < ZERO ) /* optimal multiplier is to the left */
                {
                    if ( lambdaR > lambda )
                    {
                        lambdaR = lambda ;
                        dR = slope ;
                    }
                    else
                    {
                        lambda = lambdaR ;
                        slope = dR ;
                    }
                    if ( it == K ) break ;
                    /* fix variables and compute next step */
                    while ( (j = next [j]) < n )
                    {
                        if ( br_lo [j] >= lambda ) /* fix xj */
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            b -= aj*t ;
                            x [j] = t ;
                            num -= ady [j] ;
                            den -= ada [j] ;
                            found = TRUE ;
                        }
                        else
                        {
                            next [k] = j ;
                            k = j ;
                        }
                    }
                }
                else /* optimal solution found (slope is zero) */
                {
                    status = NAPHEAP_STATUS_OK ;
                    break ;
                }

                next [k] = j ;
                if ( (found == FALSE) || (k == -1) )
                {
                    /* otherwise variable fixing, optimal solution found */
                    status = NAPHEAP_STATUS_OK ;
                    break ;
                }
            }

            Stats->nvarfix = it ;
            if ( status == NAPHEAP_STATUS_OK )
            {
                napsolution (x, y, d, a, lo, hi, n, lambda, next) ;
            }
        }

        if ( (status == NAPHEAP_STATUS_KEEP_LOOKING) || Parm->refine )
        {
            /* optimal solution not found */
            if ( dL < -dR ) /* go right */
            {
                lambda = lambdaL ;
                slope = dL ;
            }
            else
            {
                lambda = lambdaR ;
                slope = dR ;
            }

            #ifdef NAPSACK_BRUTAL
            /* !!!!!!!!! mess up the slope for debugging !!!!!!!!!! */
            /* This is only for extensive debugging, not for production use. */
            if ( Parm->refine && n > 3 )
            {
                if ( slope > 0 )
                {
                    if ( a [3] > 0 ) slope += 1 ;
                    else             slope /= 2 ;
                }
                else
                {
                    if ( a [3] > 0 ) slope -= 1 ;
                    else             slope /= 2 ;
                }
            }
            #endif

            napsearch (x, &lambda, lambdaL, lambdaR, b, &slope,
                                   y, d, a, lo, hi, n, K, breakpts, next,
                                   free_heap, bound_heap, Stats, br_lo,
                                   br_hi, Parm) ;
            napsolution (x, y, d, a, lo, hi, n, lambda, next) ;
            status = NAPHEAP_STATUS_OK ;
        }

    }
    else if ( d_is_zero )
    {

        /* ================================================================== */
        /* === case when all(d==0) holds ==================================== */
        /* ================================================================== */

        /* initialization of the slope variables */
        slopehi  = ZERO ; /* upper subdifferential at given lambda */
        slopelo  = ZERO ; /* lower subdifferential at given lambda */
        slope0 = ZERO ;   /* terms with no break point at lambda=0 */
        slope0hi = ZERO ; /* upper subdifferential terms at lambda=0 */
        slope0lo = ZERO ; /* lower subdifferential terms at lambda=0 */
        for (j = 0; j < n; j++)
        {
            aj = a [j] ;
            yj = y [j] ;
            loj = lo [j] ;
            hij = hi [j] ;

            if ( loj >= hij )
            {
                if ( loj == hij )
                {
                    t = aj*loj ;
                    b -= t ;
                    x [j] = loj ;
                }
                else
                {
                    status = NAPHEAP_STATUS_INVALID_BOUNDS ;
                    Stats->jerror = j ;
                    Stats->lobad = lo [j] ;
                    Stats->hibad = hi [j] ;
                    return (nap_wrapup (status, Stats, xw2, iw2,
                        Lambda, lambda)) ;
                }
            }
            else if ( aj == ZERO )
            {
                if ( yj > ZERO )
                {
                    if ( hij >= INF )
                    {
                        status = NAPHEAP_STATUS_UNBOUNDED ;
                        return (nap_wrapup (status, Stats, xw2, iw2,
                            Lambda, lambda)) ;
                    }
                    x [j] = hij ;
                }
                else if ( yj < ZERO )
                {
                    if ( loj <= -INF )
                    {
                        status = NAPHEAP_STATUS_UNBOUNDED ;
                        return (nap_wrapup (status, Stats, xw2, iw2,
                            Lambda, lambda)) ;
                    }
                    x [j] = loj ;
                }
            }
            else
            {
                s = yj/aj ;
                breakpts [j] = s ;
                if ( s > lambdaR ) lambdaR = s ;
                if ( s < lambdaL ) lambdaL = s ;

                /* update the domain of dual function */
                if ( aj > ZERO )
                {
                    if ( (hij < INF) && (maxax < INF) )
                    {
                        maxax += aj*hij ;
                    }
                    else
                    {
                        maxax  = INF ;
                    }
                    if ( (loj > -INF) && (minax > -INF) )
                    {
                        minax += aj*loj ;
                    }
                    else
                    {
                        minax = -INF ;
                    }
                    if ( loj <= -INF ) domhi = MIN (domhi, s) ;
                    if ( hij >=  INF ) domlo = MAX (domlo, s) ;
                }
                else /* aj < 0 */
                {

                    if ( (loj > -INF) && (maxax < INF) )
                    {
                        maxax += aj*loj ;
                    }
                    else
                    {
                        maxax  = INF ;
                    }
                    if ( (hij < INF) && (minax > -INF) )
                    {
                        minax += aj*hij ;
                    }
                    else
                    {
                        minax = -INF ;
                    }
                    if ( loj <= -INF ) domlo = MAX (domlo, s) ;
                    if ( hij >=  INF ) domhi = MIN (domhi, s) ;
                }

                /* update the slope of dual function at lambda */
                if ( use_lambda )
                {
                    if ( s < lambda )
                    {
                        if ( aj > ZERO ) xj = lo [j] ;
                        else             xj = hi [j] ;
                        slope += aj*xj ;
                    }
                    else if ( s > lambda )
                    {
                        if ( aj > ZERO ) xj = hi [j] ;
                        else             xj = lo [j] ;
                        slope += aj*xj ;
                    }
                    else /* lambda is a break point */
                    {
                        if ( aj > ZERO )
                        {
                            slopehi += aj*hij ;
                            slopelo += aj*loj ;
                        }
                        else
                        {
                            slopelo += aj*hij ;
                            slopehi += aj*loj ;
                        }
                    }
                }

                /* update the slope of the dual function at lambda = 0 */
                if      ( yj > ZERO ) slope0 += aj*hij ;
                else if ( yj < ZERO ) slope0 += aj*loj ;
                else
                {
                    if ( aj > ZERO )
                    {
                        slope0hi += aj*hij ;
                        slope0lo += aj*loj ;
                    }
                    else
                    {
                        slope0hi += aj*loj ;
                        slope0lo += aj*hij ;
                    }
                }
                next [k] = j ;
                k = j ;
            }
        }
        next [k] = n ;

        if ( (minax - b > bhi) || ( maxax - b < blo) ) /* problem infeasible */
        {
            nap_infeasible (x, a, lo, hi, n, next, minax, maxax, blo, bhi) ;
            status = NAPHEAP_STATUS_INFEASIBLE ;
            return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
        }

        if ( domlo > domhi ) /* dual function is -inf */
        {
            status = NAPHEAP_STATUS_UNBOUNDED ;
            return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
        }
        if ( use_lambda )
        {
            if ( (lambda < domlo) || (lambda > domhi) )
            {
                use_lambda = FALSE ; /* cannot use lambda if outside domain */
            }
        }

        /* if 0 in [domlo, domhi], then use slope at 0 to choose b */
        if ( (domlo <= ZERO) && (domhi >= ZERO) )
        {
            if ( slope0 + slope0hi < blo + b ) /* optimal lambda < 0 */
            {
                b += blo ;
                if ( lambdaR >= ZERO )
                {
                    lambdaR = ZERO ; /* lambda = 0 is closer than lambdaR */
                    dR = slope0 + slope0hi - b ;
                    Rout = FALSE ; /* optimal lambda left of lambdaR */
                    dL = maxax - b ;
                    Lout = TRUE ;
                }
                else /* lambdaR is closer to optimal lambda */
                {
                    dR = minax - b ;
                    /* lambdaR and lambdaL outside interval containing lambda */
                    Rout = TRUE ;
                    Lout = TRUE ;
                    if ( lambdaL < domlo )
                    {
                        lambdaL = domlo ; /* domlo closer to optimal lambda */
                        dL = INF ;
                    }
                    else dL = maxax - b ;
                }
            }
            else if ( slope0 + slope0lo > b + bhi ) /* optimal lambda > 0 */
            {
                b += bhi ;
                if ( lambdaL <= ZERO )
                {
                    lambdaL = ZERO ; /* lambda = 0 is closer than lambdaL */
                    dL = slope0 + slope0lo - b ;
                    Lout = FALSE ;
                    dR = minax - b ;
                    Rout = TRUE ;
                }
                else /* lambdaL is closer to optimal lambda */
                {
                    dL = maxax - b ;
                    /* lambdaR and lambdaL outside interval containing lambda */
                    Lout = TRUE ;
                    Rout = TRUE ;
                    if ( lambdaR > domhi )
                    {
                        lambdaR = domhi ;
                        dR = -INF ;
                    }
                    else dR = minax - b ;
                }
            }
            else /* lambda = 0 optimal, blo <= a'x <= bhi */
            {
                lambda = ZERO ;
                b -= slope0 ;
                slopehi = slope0hi ;
                slopelo = slope0lo ;
                k = -1 ;
                j =  k ;
                while ( (j = next [j]) < n )
                {
                    if      ( y [j] < ZERO ) x [j] = lo [j] ;
                    else if ( y [j] > ZERO ) x [j] = hi [j] ;
                    else /* x [j] not fixed */
                    {
                        next [k] = j ;
                        k = j ;
                    }
                }
                next [k] = j ;

                /* Determine the target b = a'x to have the smallest magnitude
                   while satisfying the constraint that blo <= b <= bhi.
                   Below, b initial stores -sum a_i x_i with x_i already at
                   a bound*/
                if ( (slopelo >= b+blo) && (slopehi <= b+bhi) )
                {
                    if ( fabs (slopelo) <= fabs (slopehi) )
                    {
                        b = slopelo ;
                    }
                    else
                    {
                        b = slopehi ;
                    }
                }
                else if ( slopehi > b + bhi )
                {
                    if ( slopelo >= b + blo )
                    {
                        if ( fabs (slopelo) <= fabs (b + bhi) )
                        {
                            b = slopelo ;
                        }
                        else
                        {
                            b += bhi ;
                        }
                    }
                    else if ( fabs (b+blo) <= fabs (b+bhi) )
                    {
                        b += blo ;
                    }
                    else
                    {
                        b += bhi ;
                    }
                }
                else
                {
                    if ( fabs (slopehi) <= fabs (b+blo) )
                    {
                        b = slopehi ;
                    }
                    else
                    {
                        b += blo ;
                    }
                }
                napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                              free_heap, bound_heap) ;
                assert (status == NAPHEAP_STATUS_OK) ;
                return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
            }
        }
        else /* lambda = 0 outside domain [domlo, domhi] of dual function */
        {
            if ( domlo > ZERO )
            {
                b += bhi ;
                lambdaL = domlo ;
                dL = INF ;
                Lout = TRUE ;
                Rout = TRUE ;
                if ( domhi < INF )
                {
                    lambdaR = domhi ;
                    dR = -INF ;
                }
                else
                {
                    if ( minax > -INF ) dR = minax - b ;
                    else                dR = -INF ;
                }
            }
            else /* domhi < ZERO */
            {
                b += blo ;
                lambdaR = domhi ;
                dR = -INF ;
                Rout = TRUE ;
                Lout = TRUE ;
                if ( domlo > -INF )
                {
                    lambdaL = domlo ;
                    dL = INF ;
                }
                else
                {
                    if ( maxax < INF ) dL = maxax - b ;
                    else               dL = INF ;
                }
            }
            if ( domlo == domhi )
            {
                lambda = domlo ;
                k = -1 ;
                j =  k ;
                while ( (j = next [j]) < n )
                {
                    s = breakpts [j] ;
                    aj = a [j] ;
                    if ( s < lambda )
                    {
                        if ( aj > ZERO ) t = lo [j] ;
                        else             t = hi [j] ;
                        x [j] = t ;
                        b -= aj*t ;
                    }
                    else if ( s > lambda )
                    {
                        if ( aj > ZERO ) t = hi [j] ;
                        else             t = lo [j] ;
                        x [j] = t ;
                        b -= aj*t ;
                    }
                    else /* lambda is the break point */
                    {
                        next [k] = j ;
                        k = j ;
                    }
                }
                next [k] = j ;
                napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                              free_heap, bound_heap) ;
                assert (status == NAPHEAP_STATUS_OK) ;
                return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
            }
        }

        /* b has been set, adjust slope */
        slope0 -= b ;

        /* if lambda has been given, check that it lies in the domain of the
           dual function, and exploit its associated slope */
        recompute_slope = FALSE ;
        if ( use_lambda )
        {
            if ( (lambda >= lambdaL) && (lambda <= lambdaR) )
            {
                slope -= b ;
                if ( slope + slopelo > ZERO )  /* lambda* is to the right */
                {
                    goright = TRUE ;
                    slope += slopelo ;
                    dL = slope ;
                    lambdaL = lambda ;
                    Lout = FALSE ;
                }
                else if ( slope + slopehi < ZERO ) /* lambda* is to the left */
                {
                    goright = FALSE ;
                    slope += slopehi ;
                    dR = slope ;
                    lambdaR = lambda ;
                    Rout = FALSE ;
                }
                else
                {
                    /* lambda is optimal */
                    napfix0 (x, a, lo, hi, n, &b, lambda, next, breakpts) ;
                    napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                                  free_heap, bound_heap) ;
                    assert (status == NAPHEAP_STATUS_OK) ;
                    return (nap_wrapup (status, Stats, xw2, iw2,
                        Lambda, lambda)) ;
                }
            }
            else if ( lambda < lambdaL )
            {
                goright = TRUE ; /* optimal multiplier to the right */
                lambda = lambdaL ;
                if ( Lout ) recompute_slope = TRUE ; /* lambda outside domain */
                else        slope = dL ;
            }
            else /* ( lambda > lambdaR ) */
            {
                goright = FALSE ; /* optimal multiplier to the left */
                lambda = lambdaR ;
                if ( Rout ) recompute_slope = TRUE ;
                else        slope = dR ;
            }
        }
        else if ( K == 0 ) /* if no lambda given and K = 0, use smaller slope */
        {
            if ( dL < -dR )
            {
                lambda = lambdaL ;
                goright = TRUE ;
                if ( Lout ) recompute_slope = TRUE ;
                else        slope = dL ;
            }
            else
            {
                if ( lambdaR < INF )
                {
                    lambda = lambdaR ;
                    goright = FALSE ;
                    if ( Rout ) recompute_slope = TRUE ;
                    else        slope = dR ;
                }
                else
                {
                    lambda = lambdaL ;
                    goright = TRUE ;
                    if ( Lout ) recompute_slope = TRUE ;
                    else        slope = dL ;
                }
            }
        }

        if ( K == 0 ) /* use break point search algorithm */
        {

            #ifdef NAPSACK_BRUTAL
            /* !!!!!!!!! mess up the slope for debugging !!!!!!!!!! */
            /* This is only for extensive debugging, not for production use. */
            if ( Parm->refine && n > 3 )
            {
                if ( slope > 0 )
                {
                    if ( a [3] > 0 ) slope += 1 ;
                    else             slope /= 2 ;
                }
                else
                {
                    if ( a [3] > 0 ) slope -= 1 ;
                    else             slope /= 2 ;
                }
            }
            #endif

            napsearch0 (x, &lambda, lambdaL, lambdaR, &b,
                                    &slope, &slopelo, &slopehi, n, a, lo, hi,
                                    breakpts, next, bound_heap, goright,
                                    Stats, recompute_slope, Parm) ;
            napfix0 (x, a, lo, hi, n, &b, lambda, next, breakpts) ;
            napsolution0 (x, a, lo, hi, n, b, breakpts, next, free_heap,
                          bound_heap) ;
            assert (status == NAPHEAP_STATUS_OK) ;
            return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
        }

        /* K > 0, evaluate num and den and prepare for variable fixing over the
           domain of the dual function */
        k = -1 ;
        j =  k ;
        if ( K == 1 ) /* do not save aj*yj and aj*aj */
        {
            while ( (j = next [j]) < n )
            {
                /* if j is in the list of undetermined variables,
                   then breakpts [j] has been computed above */
                aj = a [j] ;
                t = breakpts [j] ;
                if ( t < lambdaL ) /* xj is fixed */
                {
                    if ( aj > ZERO ) xj = lo [j] ;
                    else             xj = hi [j] ;
                    x [j] = xj ;
                    b -= aj*xj ;
                }
                else if ( t > lambdaR ) /* xj is fixed */
                {
                    if ( aj > ZERO ) xj = hi [j] ;
                    else             xj = lo [j] ;
                    x [j] = xj ;
                    b -= aj*xj ;
                }
                else
                {
                    num += aj*y [j] ;
                    den += aj*aj ;
                    next [k] = j ;
                    k = j ;
                }
            }
        }
        else /* K > 1, also save aj*yj and aj*aj */
        {
            while ( (j = next [j]) < n )
            {
                aj = a [j] ;
                t = breakpts [j] ;
                if ( t < lambdaL ) /* xj is fixed */
                {
                    if ( aj > ZERO ) xj = lo [j] ;
                    else             xj = hi [j] ;
                    x [j] = xj ;
                    b -= aj*xj ;
                }
                else if ( t > lambdaR ) /* xj is fixed */
                {
                    if ( aj > ZERO ) xj = hi [j] ;
                    else             xj = lo [j] ;
                    x [j] = xj ;
                    b -= aj*xj ;
                }
                else
                {
                    t = aj*y [j] ;
                    ady [j] = t ;
                    num += t ;
                    t = aj*aj ;
                    ada [j] = t ;
                    den += t ;
                    next [k] = j ;
                    k = j ;
                }
            }
        }
        next [k] = j ;

        /* perform K iterations of variable fixing algorithm */
        for (it = 1; it <= K; it++)
        {
            lambda = num/den ;
            if ( lambda < lambdaL )
            {
                lambda = lambdaL ;
            }
            else if ( lambda > lambdaR )
            {
                lambda = lambdaR ;
            }
            j = -1 ;
            /* slope corresponds to terms with no break point at lambda */
            slope = ZERO ;
            /* upper subdifferential terms associated with lambda */
            slopehi = ZERO ;
            /* lower subdifferential terms associated with lambda */
            slopelo = ZERO ;
            while ( (j = next [j]) < n )
            {
                aj = a [j] ;
                t = breakpts [j] ;
                if ( aj > ZERO )
                {

                    if      ( lambda > t ) slope += aj*lo [j] ;
                    else if ( lambda < t ) slope += aj*hi [j] ;
                    else
                    {
                        slopehi += aj*hi [j] ;
                        slopelo += aj*lo [j] ;
                    }
                }
                else
                {

                    if      ( lambda > t ) slope += aj*hi [j] ;
                    else if ( lambda < t ) slope += aj*lo [j] ;
                    else
                    {
                        slopehi += aj*lo [j] ;
                        slopelo += aj*hi [j] ;
                    }
                }
            }
            slope -= b ;

            /* check for optimal solution */
            if ( (slope+slopelo <= ZERO) && (slope + slopehi >= ZERO) )
            {
                napfix0 (x, a, lo, hi, n, &b, lambda, next, breakpts) ;
                napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                              free_heap, bound_heap) ;
                Stats->nvarfix = it ;
                assert (status == NAPHEAP_STATUS_OK) ;
                return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
            }

            k = -1 ;
            j =  k ;
            found = FALSE ;
            if ( slope + slopelo > ZERO )  /* optimal multiplier is to right */
            {
                if ( lambdaL < lambda )
                {
                    lambdaL = lambda ;
                    dL = slope + slopelo ;
                    Lout = FALSE ;
                }
                else lambda = lambdaL ;
                if ( it == K ) break ;
                while ( (j = next [j]) < n )
                {
                    if ( breakpts [j] < lambda ) /* xj is fixed */
                    {
                        aj = a [j] ;
                        if ( aj > ZERO ) xj = lo [j] ;
                        else             xj = hi [j] ;
                        b -= aj*xj ;
                        x [j] = xj ;
                        num -= ady [j] ;
                        den -= ada [j] ;
                        found = TRUE ;
                    }
                    else
                    {
                        next [k] = j ;
                        k = j ;
                    }
                }
                next [k] = j ;
            }
            else if ( slope + slopehi < ZERO ) /* optimal multiplier to left */
            {
                if ( lambdaR > lambda )
                {
                    lambdaR = lambda ;
                    dR = slope + slopehi ;
                    Rout = FALSE ;
                }
                else lambda = lambdaR ;
                if ( it == K ) break ;
                while ( (j = next [j]) < n )
                {
                    if ( breakpts [j] > lambda ) /* xj is fixed */
                    {
                        aj = a [j] ;
                        if ( aj > ZERO ) xj = hi [j] ;
                        else             xj = lo [j] ;
                        x [j] = xj ;
                        b -= aj*xj ;
                        num -= ady [j] ;
                        den -= ada [j] ;
                        found = TRUE ;
                    }
                    else
                    {
                        next [k] = j ;
                        k = j ;
                    }
                }
                next [k] = j ;
            }
            else    /* optimal solution found */
            {
                napfix0 (x, a, lo, hi, n, &b, lambda, next, breakpts) ;
                napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                              free_heap, bound_heap) ;
                Stats->nvarfix = it ;
                assert (status == NAPHEAP_STATUS_OK) ;
                return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
            }
            if ( !found ) break ;
        }
        Stats->nvarfix = K ;
        if ( dL < -dR ) /* go right */
        {
            lambda = lambdaL ;
            slope = dL ;
            goright = TRUE ;
            if ( Lout ) recompute_slope = TRUE ;
        }
        else /* go left unless lambdaR = INF */
        {
            if ( lambdaR < INF )
            {
                lambda = lambdaR ;
                slope = dR ;
                goright = FALSE ;
                if ( Rout ) recompute_slope = TRUE ;
            }
            else
            {
                lambda = lambdaL ;
                slope = dL ;
                goright = TRUE ;
                if ( Lout ) recompute_slope = TRUE ;
            }
        }

        #ifdef NAPSACK_BRUTAL
        /* !!!!!!!!! mess up the slope for debugging !!!!!!!!!! */
        /* This is only for extensive debugging, not for production use. */
        if ( Parm->refine && n > 3 )
        {
            if ( slope > 0 )
            {
                if ( a [3] > 0 ) slope += 1 ;
                else             slope /= 2 ;
            }
            else
            {
                if ( a [3] > 0 ) slope -= 1 ;
                else             slope /= 2 ;
            }
        }
        #endif

        napsearch0 (x, &lambda, lambdaL, lambdaR, &b,
                    &slope, &slopelo, &slopehi, n, a, lo, hi,
                    breakpts, next, bound_heap, goright,
                    Stats, recompute_slope, Parm) ;
        napfix0 (x, a, lo, hi, n, &b, lambda, next, breakpts) ;
        napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                      free_heap, bound_heap) ;

    }
    else /* !d_is_zero && !d_is_pos (diagonal not fully 0 or fully positive) */
    {

        /* ================================================================== */
        /* === case for general d =========================================== */
        /* ================================================================== */

        /* initialization of the slope variables */
        slopehi  = ZERO ; /* upper subdifferential at given lambda */
        slopelo  = ZERO ; /* lower subdifferential at given lambda */
        slope0 = ZERO ;   /* terms with no break point at lambda=0 */
        slope0hi = ZERO ; /* upper subdifferential terms at lambda=0 */
        slope0lo = ZERO ; /* lower subdifferential terms at lambda=0 */
        for (j = 0; j < n; j++)
        {
            dj = d [j] ;
            if ( dj < 0 )
            {
                status = NAPHEAP_STATUS_INVALID_D ;
                Stats->jerror = j ;
                Stats->dbad = d [j] ;
                return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
            }
            aj = a [j] ;
            yj = y [j] ;
            loj = lo [j] ;
            hij = hi [j] ;

            if ( loj >= hij )
            {
                if ( loj == hij )
                {
                    t = aj*loj ;
                    b -= t ;
                    x [j] = loj ;
                    continue ;
                }
                else
                {
                    status = NAPHEAP_STATUS_INVALID_BOUNDS ;
                    Stats->jerror = j ;
                    Stats->lobad = lo [j] ;
                    Stats->hibad = hi [j] ;
                    return (nap_wrapup (status, Stats, xw2, iw2,
                        Lambda, lambda)) ;
                }
            }
            if ( dj ) /* dj > 0 */
            {
                if ( aj == ZERO )
                {
                    xj = yj/dj ;
                    if      ( xj < loj ) xj = loj ;
                    else if ( xj > hij ) xj = hij ;
                    x [j] = xj ;
                }
                else
                {
                    if ( aj > ZERO )
                    {
                        if ( (hij < INF) && (maxax < INF) )
                        {
                            maxax += aj*hij ;
                        }
                        else
                        {
                            maxax  = INF ;
                        }
                        if ( (loj > -INF) && (minax > -INF) )
                        {
                            minax += aj*loj ;
                        }
                        else
                        {
                            minax = -INF ;
                        }
                        /* also store max and min break points */
                        t = (yj - loj*dj)/aj ;
                        if ( t > lambdaR )
                        {
                            if ( loj > -INF ) lambdaR = t ;
                            else              lambdaR = INF ;
                        }
                        t = (yj - hij*dj)/aj ;
                        if ( t < lambdaL )
                        {
                            if ( hij < INF ) lambdaL = t ;
                            else             lambdaL = -INF ;
                        }
                    }
                    else
                    {
                        if ( (loj > -INF) && (maxax < INF) )
                        {
                            maxax += aj*loj ;
                        }
                        else
                        {
                            maxax  = INF ;
                        }
                        if ( (hij < INF) && (minax > -INF) )
                        {
                            minax += aj*hij ;
                        }
                        else
                        {
                            minax = -INF ;
                        }
                        /* also store max and min break points */
                        t = (yj - hij*dj)/aj ;
                        if ( t > lambdaR )
                        {
                            if ( hij < INF ) lambdaR = t ;
                            else             lambdaR = INF ;
                        }
                        t = (yj - loj*dj)/aj ;
                        if ( t < lambdaL )
                        {
                            if ( loj > -INF ) lambdaL = t ;
                            else              lambdaL = -INF ;
                        }
                    }

                    /* slope at lambda */
                    if ( use_lambda )
                    {
                        t = (yj - lambda*aj)/dj ;
                        if      ( t < loj ) t = loj ;
                        else if ( t > hij ) t = hij ;
                        slope += aj*t ;
                    }

                    /* slope at lambda = 0 */
                    yj /= dj ; /* yj now stores yj/dj */
                    if      ( yj < loj ) yj = loj ;
                    else if ( yj > hij ) yj = hij ;
                    slope0 += aj*yj ;
                    next [k] = j ;
                    k = j ;
                }
            }
            else /* dj = 0 */
            {
                if ( aj == ZERO )
                {
                    if ( yj > ZERO )
                    {
                        if ( hij >= INF )
                        {
                            status = NAPHEAP_STATUS_UNBOUNDED ;
                            return (nap_wrapup (status, Stats, xw2, iw2,
                                Lambda, lambda)) ;
                        }
                        x [j] = hij ;
                    }
                    else if ( yj < ZERO )
                    {
                        if ( loj <= -INF )
                        {
                            status = NAPHEAP_STATUS_UNBOUNDED ;
                            return (nap_wrapup (status, Stats, xw2, iw2,
                                Lambda, lambda)) ;
                        }
                        x [j] = loj ;
                    }
                }
                else
                {
                    s = yj/aj ;
                    breakpts [j] = s ;
                    if ( s > lambdaR ) lambdaR = s ;
                    if ( s < lambdaL ) lambdaL = s ;

                    /* update the domain of dual function */
                    if ( aj > ZERO )
                    {
                        if ( (hij < INF) && (maxax < INF) )
                        {
                            maxax += aj*hij ;
                        }
                        else
                        {
                            maxax  = INF ;
                        }
                        if ( (loj > -INF) && (minax > -INF) )
                        {
                            minax += aj*loj ;
                        }
                        else
                        {
                            minax = -INF ;
                        }
                        if ( loj <= -INF ) domhi = MIN (domhi, s) ;
                        if ( hij >=  INF ) domlo = MAX (domlo, s) ;
                    }
                    else /* aj < 0 */
                    {
                        if ( (loj > -INF) && (maxax < INF) )
                        {
                            maxax += aj*loj ;
                        }
                        else
                        {
                            maxax  = INF ;
                        }
                        if ( (hij < INF) && (minax > -INF) )
                        {
                            minax += aj*hij ;
                        }
                        else
                        {
                            minax = -INF ;
                        }
                        if ( loj <= -INF ) domlo = MAX (domlo, s) ;
                        if ( hij >=  INF ) domhi = MIN (domhi, s) ;
                    }

                    /* compute slope at lambda */
                    if ( use_lambda )
                    {
                        if ( s < lambda )
                        {
                            if ( aj > ZERO ) xj = lo [j] ;
                            else             xj = hi [j] ;
                            slope += aj*xj ;
                        }
                        else if ( s > lambda )
                        {
                            if ( aj > ZERO ) xj = hi [j] ;
                            else             xj = lo [j] ;
                            slope += aj*xj ;
                        }
                        else /* lambda is a break point */
                        {
                            if ( aj > ZERO )
                            {
                                slopehi += aj*hij ;
                                slopelo += aj*loj ;
                            }
                            else
                            {
                                slopelo += aj*hij ;
                                slopehi += aj*loj ;
                            }
                        }
                    }

                    /* compute slope at lambda = 0 */
                    if      ( yj > ZERO ) slope0 += aj*hij ;
                    else if ( yj < ZERO ) slope0 += aj*loj ;
                    else
                    {
                        if ( aj > ZERO )
                        {
                            slope0hi += aj*hij ;
                            slope0lo += aj*loj ;
                        }
                        else
                        {
                            slope0lo += aj*hij ;
                            slope0hi += aj*loj ;
                        }
                    }
                    next [k] = j ;
                    k = j ;
                }
            }
        }
        next [k] = j ;

        /* check if problem feasible */
        if ( (minax - b > bhi) || (maxax - b < blo) )
        {
            nap_infeasible (x, a, lo, hi, n, next, minax, maxax, blo, bhi) ;
            status = NAPHEAP_STATUS_INFEASIBLE ;
            return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
        }

        if ( domlo > domhi )
        {
            status = NAPHEAP_STATUS_UNBOUNDED ;
            return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
        }
       if ( use_lambda )
       {
           if ( (lambda < domlo) || (lambda > domhi) )
           {
               use_lambda = FALSE ; /* cannot use lambda if outside domain */
           }
       }

        /* if 0 in [domlo, domhi], then use slope at 0 to choose b */
        if ( (domlo <= ZERO) && (domhi >= ZERO) )
        {
            if ( slope0 + slope0hi < blo + b ) /* optimal lambda < 0 */
            {
                b += blo ;
                if ( lambdaR >= ZERO )
                {
                    lambdaR = ZERO ; /* lambda = 0 is closer than lambdaR */
                    dR = slope0 + slope0hi - b ;
                    Rout = FALSE ; /* optimal lambda left of lambdaR */
                    if ( lambdaL > -INF ) dL = maxax - b ;
                    else                  dL = INF ;
                    Lout = TRUE ;
                }
                else /* lambdaR is closer to optimal lambda */
                {
                    dR = minax - b ;
                    /* lambdaR and lambdaL outside interval containing lambda */
                    Rout = TRUE ;
                    Lout = TRUE ;
                    if ( lambdaL < domlo )
                    {
                        lambdaL = domlo ; /* domlo closer to optimal lambda */
                        dL = INF ;
                    }
                    else dL = maxax - b ;
                }
            }
            else if ( slope0 + slope0lo > b + bhi ) /* optimal lambda > 0 */
            {
                b += bhi ;
                if ( lambdaL <= ZERO )
                {
                    lambdaL = ZERO ; /* lambda = 0 is closer than lambdaL */
                    dL = slope0 + slope0lo - b ;
                    Lout = FALSE ;
                    if ( lambdaR < INF ) dR = minax - b ;
                    else                 dR = -INF ;
                    Rout = TRUE ;
                }
                else /* lambdaL is closer to optimal lambda */
                {
                    dL = maxax - b ;
                    /* lambdaR and lambdaL outside interval containing lambda */
                    Lout = TRUE ;
                    Rout = TRUE ;
                    if ( lambdaR > domhi )
                    {
                        lambdaR = domhi ;
                        dR = -INF ;
                    }
                    else dR = minax - b ;
                }
            }
            else /* lambda = 0 optimal, blo <= a'x <= bhi */
            {
                lambda = ZERO ;
                b -= slope0 ;
                slopehi = slope0hi ;
                slopelo = slope0lo ;
                k = -1 ;
                j =  k ;
                while ( (j = next [j]) < n )
                {
                    dj = d [j] ;
                    if ( dj )
                    {
                        xj = y [j]/dj ;
                        if      ( xj < lo [j] ) x [j] = lo [j] ;
                        else if ( xj > hi [j] ) x [j] = hi [j] ;
                        else                    x [j] = xj ;
                    }
                    else
                    {
                        if      ( y [j] < ZERO ) x [j] = lo [j] ;
                        else if ( y [j] > ZERO ) x [j] = hi [j] ;
                        else
                        {
                            next [k] = j ;
                            k = j ;
                        }
                    }
                }
                next [k] = j ;

                /* Determine the target b = a'x to have the smallest magnitude
                   while satisfying the constraint that blo <= b <= bhi.
                   Below, b initial stores -sum a_i x_i with x_i already at
                   a bound*/
                if ( (slopelo >= b+blo) && (slopehi <= b+bhi) )
                {
                    if ( fabs (slopelo) <= fabs (slopehi) )
                    {
                        b = slopelo ;
                    }
                    else
                    {
                        b = slopehi ;
                    }
                }
                else if ( slopehi > b + bhi )
                {
                    if ( slopelo >= b + blo )
                    {
                        if ( fabs (slopelo) <= fabs (b + bhi) )
                        {
                            b = slopelo ;
                        }
                        else
                        {
                            b += bhi ;
                        }
                    }
                    else if ( fabs (b+blo) <= fabs (b+bhi) )
                    {
                        b += blo ;
                    }
                    else
                    {
                        b += bhi ;
                    }
                }
                else
                {
                    if ( fabs (slopehi) <= fabs (b+blo) )
                    {
                        b = slopehi ;
                    }
                    else
                    {
                        b += blo ;
                    }
                }

                /* fix all variables for which the break point != lambda */
                napfix1 (x, a, y, d, lo, hi, n, &b, lambda, next, breakpts) ;
                napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                              free_heap, bound_heap) ;
                assert (status == NAPHEAP_STATUS_OK) ;
                return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
            }
        }
        else /* lambda = 0 outside domain [domlo, domhi] of dual function */
        {
            if ( domlo > ZERO )
            {
                b += bhi ;
                lambdaL = domlo ;
                dL = INF ;
                Lout = TRUE ;
                Rout = TRUE ;
                if ( domhi < INF )
                {
                    lambdaR = domhi ;
                    dR = -INF ;
                }
                else
                {
                    if ( minax > -INF ) dR = minax - b ;
                    else                dR = -INF ;
                }
            }
            else /* domhi < ZERO */
            {
                b += blo ;
                lambdaR = domhi ;
                dR = -INF ;
                Rout = TRUE ;
                Lout = TRUE ;
                if ( domlo > -INF )
                {
                    lambdaL = domlo ;
                    dL = INF ;
                }
                else
                {
                    if ( maxax < INF ) dL = maxax - b ;
                    else               dL = INF ;
                }
            }
            if ( domlo == domhi )
            {
                lambda = domlo ;
                napfix1 (x, a, y, d, lo, hi, n, &b, lambda, next, breakpts) ;
                napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                              free_heap, bound_heap) ;
                assert (status == NAPHEAP_STATUS_OK) ;
                return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
            }
        }

        /* b has been set, adjust slope */
        slope0 -= b ;

        /* if lambda has been given, check that it lies in the domain of the
           dual function, and exploit its associated slope */
        recompute_slope = FALSE ;
        if ( use_lambda )
        {
            if ( (lambda >= lambdaL) && (lambda <= lambdaR) )
            {
                slope -= b ;
                if ( slope + slopelo > ZERO )  /* opt lambda is to right */
                {
                    goright = TRUE ;
                    slope += slopelo ;
                    dL = slope ;
                    lambdaL = lambda ;
                    Lout = FALSE ;
                }
                else if ( slope + slopehi < ZERO ) /* opt lambda is to left */
                {
                    goright = FALSE ;
                    slope += slopehi ;
                    dR = slope ;
                    lambdaR = lambda ;
                    Rout = FALSE ;
                }
                else
                {
                    /* lambda is optimal */
                    napfix1 (x, a, y, d, lo, hi, n, &b, lambda, next, breakpts);
                    napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                                  free_heap, bound_heap) ;
                    assert (status == NAPHEAP_STATUS_OK) ;
                    return (nap_wrapup (status, Stats, xw2, iw2,
                        Lambda, lambda)) ;
                }
            }
            else if ( lambda < lambdaL ) /* optimal multiplier to the right */
            {
                goright = TRUE ;
                lambda = lambdaL ;
                if ( Lout ) recompute_slope = TRUE ; /* lambda outside domain */
                else        slope = dL ;
            }
            else /* ( lambda > lambdaR )  optimal multiplier to the left */
            {
                goright = FALSE ;
                lambda = lambdaR ;
                if ( Rout ) recompute_slope = TRUE ;
                else        slope = dR ;
            }
        }
        else if ( K == 0 ) /* if no lambda given and K = 0, use smaller slope */
        {
            if ( dL < -dR )
            {
                lambda = lambdaL ;
                goright = TRUE ;
                if ( Lout ) recompute_slope = TRUE ;
                else        slope = dL ;
            }
            else
            {
                if ( lambdaR < INF )
                {
                    lambda = lambdaR ;
                    goright = FALSE ;
                    if ( Rout ) recompute_slope = TRUE ;
                    else        slope = dR ;
                }
                else
                {
                    lambda = lambdaL ;
                    goright = TRUE ;
                    if ( Lout ) recompute_slope = TRUE ;
                    else        slope = dL ;
                }
            }
        }

        if ( K == 0 ) /* use break point search algorithm */
        {

            #ifdef NAPSACK_BRUTAL
            /* !!!!!!!!! mess up the slope for debugging !!!!!!!!!! */
            /* This is only for extensive debugging, not for production use. */
            if ( Parm->refine && n > 3 )
            {
                if ( slope > 0 )
                {
                    if ( a [3] > 0 ) slope += 5 ;
                    else             slope /= 4 ;
                }
                else
                {
                    if ( a [3] > 0 ) slope -= 5 ;
                    else             slope /= 4 ;
                }
            }
            #endif

            napsearch1 (x, &lambda, lambdaL, lambdaR, &b, &slope, &slopelo,
                        &slopehi, y, d, a, lo, hi, n, K, breakpts, next,
                        free_heap, bound_heap, goright, Stats, br_lo, br_hi,
                        recompute_slope, Parm) ;
            /* fix all variables for which the break point != lambda */
            napfix1 (x, a, y, d, lo, hi, n, &b, lambda, next, breakpts) ;

            /* determine the variables for which di = 0 and the
               associated break point = lambda */
            napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                          free_heap, bound_heap) ;
            assert (status == NAPHEAP_STATUS_OK) ;
            return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
        }

        /* K > 0, evaluate num and den and prepare for variable fixing over the
           domain of the dual function */
        if ( K > 1 ) k_gt_1 = TRUE ;
        else         k_gt_1 = FALSE ;
        k = -1 ;
        j =  k ;
        while ( (j = next [j]) < n )
        {
            aj = a [j] ;
            dj = d [j] ;
            if ( dj > ZERO )
            {
                yj = y [j] ;
                if ( aj > ZERO )
                {
                    t = lo [j] ;
                    r = (yj - t*dj)/aj ; /* high break point */
                    if ( r < lambdaL )   /* j is fixed */
                    {
                        b -= aj*t ;
                        x [j] = t ;
                        continue ;
                    }

                    t = hi [j] ;
                    s = (yj - t*dj)/aj ; /* low break point */
                    if ( s > lambdaR )   /* j is fixed */
                    {
                        b -= aj*t ;
                        x [j] = t ;
                        continue ;
                    }
                    br_hi [j] = r ;
                    br_lo [j] = s ;
                }
                else
                {
                    t = lo [j] ;
                    r = (yj - t*dj)/aj ; /* low break point */
                    if ( r > lambdaR )   /* j is fixed */
                    {
                        b -= aj*t ;
                        x [j] = t ;
                        continue ;
                    }

                    t = hi [j] ;
                    s = (yj - t*dj)/aj ; /* high break point */
                    if ( s < lambdaL )   /* j is fixed */
                    {
                        b -= aj*t ;
                        x [j] = t ;
                        continue ;
                    }
                    br_hi [j] = s ;
                    br_lo [j] = r ;
                }

                /* j not fixed */
                next [k] = j ;
                k = j ;
            }
            else
            {
                if ( breakpts [j] < lambdaL ) /* xj is fixed */
                {
                    if ( aj > ZERO ) xj = lo [j] ;
                    else             xj = hi [j] ;
                    x [j] = xj ;
                    b -= aj*xj ;
                }
                else if ( breakpts [j] > lambdaR ) /* xj is fixed */
                {
                    if ( aj > ZERO ) xj = hi [j] ;
                    else             xj = lo [j] ;
                    x [j] = xj ;
                    b -= aj*xj ;
                }
                else
                {
                    ajy = aj*y [j] ;
                    num += ajy ;
                    ajaj = aj*aj ;
                    den += ajaj ;
                    next [k] = j ;
                    k = j ;
                    if ( k_gt_1 )
                    {
                        ada [j] = ajaj ;
                        ady [j] = ajy ;
                    }
                }
            }
        }
        next [k] = j ;

        /* Perform up to K iterations of the variable fixing algorithm
           treating the zero diagonal elements as vanishingly small numbers. */
        if ( den ) /* there exist some zero diagonal elements no at bounds */
        {
            status = NAPHEAP_STATUS_KEEP_LOOKING ; /* default status */
            for (it = 1; it <= K; it++)
            {
                lambda = num/den ;
                if ( lambda < lambdaL )
                {
                    lambda = lambdaL ;
                }
                else if ( lambda > lambdaR )
                {
                    lambda = lambdaR ;
                }
                /* slope corresponds to terms with no break point at lambda */
                slope = ZERO ;
                /* upper subdifferential terms associated with lambda */
                slopehi = ZERO ;
                /* lower subdifferential terms associated with lambda */
                slopelo = ZERO ;
                j = -1 ;
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    if ( d [j] )
                    {
                        t = (y [j] - lambda*aj)/d [j] ;
                        if      ( t < lo [j] ) t = lo [j] ;
                        else if ( t > hi [j] ) t = hi [j] ;
                        slope += aj*t ;
                    }
                    else
                    {
                        t = breakpts [j] ;
                        /* update the slope of dual function */
                        if ( lambda < t )
                        {
                            if ( aj > ZERO ) slope += aj*hi [j] ;
                            else             slope += aj*lo [j] ;
                        }
                        else if ( lambda > t )
                        {
                            if ( aj > ZERO ) slope += aj*lo [j] ;
                            else             slope += aj*hi [j] ;
                        }
                        else /* lambda is a break point */
                        {
                            if ( aj > ZERO )
                            {
                                slopehi += aj*hi [j] ;
                                slopelo += aj*lo [j] ;
                            }
                            else
                            {
                                slopelo += aj*hi [j] ;
                                slopehi += aj*lo [j] ;
                            }
                        }
                    }
                }
                slope -= b ;
                k = -1 ;
                j =  k ;
                found = FALSE ;
                if ( slope + slopelo > ZERO )  /* optimal multiplier is right */
                {
                    if ( lambdaL < lambda )
                    {
                        lambdaL = lambda ;
                        dL = slope + slopelo ;
                        Lout = FALSE ;
                    }
                    else lambda = lambdaL ;
                    if ( it == K ) break ;
                    while ( (j = next [j]) < n )
                    {
                        if ( d [j] )
                        {
                            if ( br_hi [j] <= lambda ) /* fix xj */
                            {
                                aj = a [j] ;
                                if ( aj > ZERO ) xj = lo [j] ;
                                else             xj = hi [j] ;
                                b -= aj*xj ;
                                x [j] = xj ;
                                found = TRUE ;
                            }
                            else
                            {
                                next [k] = j ;
                                k = j ;
                            }
                        }
                        else
                        {
                            if ( breakpts [j] < lambda ) /* xj is fixed */
                            {
                                aj = a [j] ;
                                if ( aj > ZERO ) xj = lo [j] ;
                                else             xj = hi [j] ;
                                b -= aj*xj ;
                                x [j] = xj ;
                                num -= ady [j] ;
                                den -= ada [j] ;
                                found = TRUE ;
                            }
                            else
                            {
                                next [k] = j ;
                                k = j ;
                            }
                        }
                    }
                }
                else if ( slope + slopehi < ZERO ) /* opt multiplier is left */
                {
                    if ( lambdaR > lambda )
                    {
                        lambdaR = lambda ;
                        dR = slope + slopehi ;
                        Rout = FALSE ;
                    }
                    else lambda = lambdaR ;
                    if ( it == K ) break ;
                    while ( (j = next [j]) < n )
                    {
                        if ( d [j] )
                        {
                            if ( br_lo [j] >= lambda ) /* fix xj */
                            {
                                aj = a [j] ;
                                if ( aj > ZERO ) xj = hi [j] ;
                                else             xj = lo [j] ;
                                b -= aj*xj ;
                                x [j] = xj ;
                                found = TRUE ;
                            }
                            else
                            {
                                next [k] = j ;
                                k = j ;
                            }
                        }
                        else
                        {
                            if ( breakpts [j] > lambda ) /* xj is fixed */
                            {
                                aj = a [j] ;
                                if ( aj > ZERO ) xj = hi [j] ;
                                else             xj = lo [j] ;
                                b -= aj*xj ;
                                x [j] = xj ;
                                num -= ady [j] ;
                                den -= ada [j] ;
                                found = TRUE ;
                            }
                            else
                            {
                                next [k] = j ;
                                k = j ;
                            }
                        }
                    }
                }
                else /* optimal solution found */
                {
                    status = NAPHEAP_STATUS_OK ;
                    break ;
                }

                num -= b ;
                next [k] = j ;
                if ( (found == FALSE) || (k == -1) )
                {
                    /* solution not found since variable fixing only applied to
                       indices where dj = 0 */
                    break ;
                }
            }
            Stats->nvarfix = it ;
            if ( status == NAPHEAP_STATUS_OK )
            {
                napfix1 (x, a, y, d, lo, hi, n, &b, lambda, next, breakpts) ;
                napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                              free_heap, bound_heap) ;
            }
        }
        else /* d is positive, variables for zero diagonal elements at bounds */
        {
            /* d is positive, no linear part */
            d_is_pos = TRUE ;
            j = -1 ;
            num = ZERO ;
            if ( K == 1 )
            {
                while ( (j = next [j]) < n )
                {
                    dj = d [j] ;
                    aj = a [j] ;
                    yj = y [j] ;
                    adj = aj/dj ;
                    num += adj*yj ;
                    t = adj*aj ;
                    den += t ;
                }
            }
            else
            {
                while ( (j = next [j]) < n )
                {
                    dj = d [j] ;
                    aj = a [j] ;
                    yj = y [j] ;
                    adj = aj/dj ;
                    t = adj*yj ;
                    num += t ;
                    ady [j] = t ;
                    t = adj*aj ;
                    den += t ;
                    ada [j] = t ;
                }
            }

            /* Perform up to K iterations of the variable fixing algorithm */
            status = NAPHEAP_STATUS_KEEP_LOOKING ; /* default status */
            for (it = 1; it <= K; it++)
            {
                /* evaluate new lambda and the slope */
                lambda  = (num-b)/den ;
                slope = ZERO ;
                j = -1 ;
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    t = (y [j] - lambda*aj)/d [j] ;
                    if      ( t < lo [j] ) t = lo [j] ;
                    else if ( t > hi [j] ) t = hi [j] ;
                    slope += aj*t ;
                }
                slope -= b ;

                k = -1 ;
                j =  k ;
                found = FALSE ;

                if ( slope > ZERO )     /* optimal multiplier is to the right */
                {
                    if ( lambdaL < lambda )
                    {
                        lambdaL = lambda ;
                        dL = slope ;
                    }
                    else
                    {
                        lambda = lambdaL ;
                        slope = dL ;
                    }
                    if ( it == K ) break ;
                    /* fix variables and compute next step */
                    while ( (j = next [j]) < n )
                    {
                        if ( br_hi [j] <= lambda ) /* fix xj */
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            b -= aj*t ;
                            x [j] = t ;
                            num -= ady [j] ;
                            den -= ada [j] ;
                            found = TRUE ;
                        }
                        else
                        {
                            next [k] = j ;
                            k = j ;
                        }
                    }
                }
                else if ( slope < ZERO ) /* optimal multiplier is to the left */
                {
                    if ( lambdaR > lambda )
                    {
                        lambdaR = lambda ;
                        dR = slope ;
                    }
                    else
                    {
                        lambda = lambdaR ;
                        slope = dR ;
                    }
                    if ( it == K ) break ;
                    /* fix variables and compute next step */
                    while ( (j = next [j]) < n )
                    {
                        if ( br_lo [j] >= lambda ) /* fix xj */
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            b -= aj*t ;
                            x [j] = t ;
                            num -= ady [j] ;
                            den -= ada [j] ;
                            found = TRUE ;
                        }
                        else
                        {
                            next [k] = j ;
                            k = j ;
                        }
                    }
                }
                else /* optimal solution found (slope is zero) */
                {
                    status = NAPHEAP_STATUS_OK ;
                    break ;
                }

                next [k] = j ;
                if ( (found == FALSE) || (k == -1) )
                {
                    /* otherwise variable fixing, optimal solution found */
                    status = NAPHEAP_STATUS_OK ;
                    break ;
                }
            }

            Stats->nvarfix = it ;
            if ( status == NAPHEAP_STATUS_OK )
            {
                napsolution (x, y, d, a, lo, hi, n, lambda, next) ;
            }
        }

        if ( status == NAPHEAP_STATUS_KEEP_LOOKING ) /* solution not found */
        {
            /* slope was computed in variable fixing algorithm */
            recompute_slope = FALSE ;
            status = NAPHEAP_STATUS_OK ;
            if ( dL < -dR ) /* go right */
            {
                lambda = lambdaL ;
                slope = dL ;
                goright = TRUE ;
                if ( Lout ) recompute_slope = TRUE ;
            }
            else /* go left unless lambdaR = INF */
            {
                if ( lambdaR < INF )
                {
                    lambda = lambdaR ;
                    slope = dR ;
                    goright = FALSE ;
                    if ( Rout ) recompute_slope = TRUE ;
                }
                else
                {
                    lambda = lambdaL ;
                    slope = dL ;
                    goright = TRUE ;
                    if ( Lout ) recompute_slope = TRUE ;
                }
            }

            #ifdef NAPSACK_BRUTAL
            /* !!!!!!!!! mess up the slope for debugging !!!!!!!!!! */
            /* This is only for extensive debugging, not for production use. */
            if ( Parm->refine && n > 3 )
            {
                if ( slope > 0 )
                {
                    if ( a [3] > 0 ) slope += 5 ;
                    else             slope /= 4 ;
                }
                else
                {
                    if ( a [3] > 0 ) slope -= 5 ;
                    else             slope /= 4 ;
                }
            }
            #endif

            if ( d_is_pos )
            {
                napsearch (x, &lambda, lambdaL, lambdaR, b, &slope,
                                       y, d, a, lo, hi, n, K, breakpts, next,
                                       free_heap, bound_heap, Stats, br_lo,
                                       br_hi, Parm) ;
                napsolution (x, y, d, a, lo, hi, n, lambda, next) ;
            }
            else
            {
                napsearch1 (x, &lambda, lambdaL, lambdaR, &b, &slope, &slopelo,
                            &slopehi, y, d, a, lo, hi, n, K, breakpts, next,
                            free_heap, bound_heap, goright, Stats,
                            br_lo, br_hi, recompute_slope, Parm) ;
                /* fix all variables for which the break point != lambda */
                napfix1 (x, a, y, d, lo, hi, n, &b, lambda, next, breakpts) ;

                /* determine the variables for which di = 0 and the
                   associated break point = lambda */
                napsolution0 (x, a, lo, hi, n, b, breakpts, next,
                              free_heap, bound_heap) ;
            }
        }
    }

    /* wrapup and return solution */
    return (nap_wrapup (status, Stats, xw2, iw2, Lambda, lambda)) ;
}



/* ========================================================================== */
/* === non-user-callable functions ========================================== */
/* ========================================================================== */

/* The remainder of the functions below are not user-callable. */

/* ========================================================================== */
/* === nap_wrapup =========================================================== */
/* ========================================================================== */

/* finalize the solution, free workspace, and return to the user */

int nap_wrapup
(
    int status,
    napheap_stats *Stats,
    double *xw2,
    NAPHEAP_INT *iw2,
    double *Lambda,
    double lambda
)
{
    if ( Lambda != NULL) *Lambda = lambda ;
    Stats->status = status ;
    if ( xw2 != NULL ) free (xw2) ;
    if ( iw2 != NULL ) free (iw2) ;
    return (status) ;
}

/*  ==========================================================================
    === napsearch ============================================================
    ==========================================================================
    If the slope of the dual function at lambda is positive, then search
    over break points to the right. If the slope is negative, then search
    over break points to the left. Recompute the slope at the current lambda
    and possibly refine x.
    ========================================================================= */

PRIVATE void napsearch
(
    double        *x,    /* size n. solution (output) */
    double   *lambda,    /* optimal multiplier (input = guess, output = sol) */
    double   lambdaL,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double   lambdaR,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double         B,    /* a'x = B */
    double    *slope,    /* derivative of dual function */
    double const  *y,    /* size n. linear term in objective function */
    double const  *d,    /* size n, diagonal of objective Hessian */
    double const  *a,    /* size n, linear constraint vector */
    double const *lo,    /* size n, lower bounds on x */
    double const *hi,    /* size n, upper bounds on x */
    NAPHEAP_INT    n,    /* problem dimension */
    NAPHEAP_INT    K,    /* number of variable fixing iterations */
    double *breakpts,    /* size n, breakpts [j] is br_hi[j] or br_lo[j] */
    NAPHEAP_INT *next,   /* size n+1, next[-1..n-1] link list of unbound var */
    NAPHEAP_INT *free_heap,     /* size n+1, heap of free variables */
    NAPHEAP_INT *bound_heap,    /* size n+1, heap of bound variables */
    napheap_stats *Stats,    /* solution statistics */
    double const *br_lo, /* size n, br_lo[j] lo break pt for x[j] */
    double const *br_hi, /* size n. br_hi[j] hi break pt for x[j] */
    napheap_parameter *Parm /* parameters */
)
{
    NAPHEAP_INT j, jb, jfree, k, Nb, Nf, nb, nf, nbound, nfree ;
    int goright, refine ;
    double a2sum, aj, b, dj, hij, loj, yj, s, t, err,
           maxbound, maxfree, minbound, minfree, new_break ;

    /* initialization */
    b = B ;
    err = Parm->err ; /* error threshhold that triggers refinement */
    refine = Parm->refine ; /* TRUE means to refine the solution */
    nbound = 0 ;   /* counts number of bound variables */
    nfree = 0 ;    /* counts number of free variables */
    nb = n ;       /* pointer to start of previously bound variables */
    nf = n ;       /* pointer to start of previously free variables */
    a2sum = ZERO ; /* sum aj*aj/dj for free variables */
    k = -1 ;
    j =  k ;

    if ( *slope > ZERO ) /* move right */
    {
        /* d is positive and lambda is to the left of the true multiplier.
           Get ready to search over the break points to the right by
           constructing the heaps for currently bound and free variables,
           and by fixing variables whose break points are to the left. */
        goright = TRUE ;
        if ( K > 0 )
        {
            /* either variable fixing or Newton done, br_lo and br_hi exist */
            while ( (j = next [j]) < n )
            {
                s = br_hi [j] ;
                if ( s <= *lambda ) /* j is fixed */
                {
                    aj = a [j] ;
                    if ( aj > ZERO ) t = lo [j] ;
                    else             t = hi [j] ;
                    x [j] = t ;
                    b -= aj*t ;
                }
                else
                {
                    t = br_lo [j] ;
                    if ( t >= lambdaR ) /* j is fixed */
                    {
                        aj = a [j] ;
                        if ( aj > ZERO ) t = hi [j] ;
                        else             t = lo [j] ;
                        x [j] = t ;
                        b -= aj*t ;
                    }
                    else                     /* keep this variable */
                    {
                        next [k] = j ;
                        k = j ;
                        if ( t > *lambda )   /* j is bound */
                        {
                            nbound++ ;
                            bound_heap [nbound] = j ;
                            breakpts [j] = t ;
                        }
                        else                 /* j is free */
                        {
                            aj = a [j] ;
                            a2sum += aj*aj/d [j] ;
                            nfree++ ;
                            free_heap [nfree] = j ;
                            breakpts [j] = s ;
                        }
                    }
                }
            }
        }
        else /* K = 0 */
        {
            while ( (j = next [j]) < n )
            {
                aj = a [j] ;
                yj = y [j] ;
                dj = d [j] ;
                if ( aj > ZERO )
                {
                    loj = lo [j] ;
                    s = (yj - loj*dj)/aj ;        /* hi break point */
                    if ( s <= *lambda )           /* j is fixed */
                    {
                        x [j] = loj ;
                        b -= aj*loj ;
                    }
                    else
                    {
                        hij = hi [j] ;
                        t = (yj - hij*dj)/aj ;    /* lo break point */
                        if ( t >= lambdaR )       /* j is fixed */
                        {
                            x [j] = hij ;
                            b -= aj*hij ;
                        }
                        else                      /* keep this variable */
                        {
                            next [k] = j ;
                            k = j ;
                            if ( t > *lambda )    /* j is bound */
                            {
                                nbound++ ;
                                bound_heap [nbound] = j ;
                                breakpts [j] = t ;
                            }
                            else                  /* j is free */
                            {
                                a2sum += aj*aj/dj ;
                                nfree++ ;
                                free_heap [nfree] = j ;
                                breakpts [j] = s ;
                            }
                        }
                    }
                }
                else /* aj < 0 */
                {
                    hij = hi [j] ;
                    s = (yj - hij*dj)/aj ;        /* hi break point */
                    if ( s <= *lambda )           /* j is fixed */
                    {
                        x [j] = hij ;
                        b -= aj*hij ;
                    }
                    else
                    {
                        loj = lo [j] ;
                        t = (yj - loj*dj)/aj ;    /* lo break point */
                        if ( t >= lambdaR )       /* j is fixed */
                        {
                            x [j] = loj ;
                            b -= aj*loj ;
                        }
                        else                      /* keep this variable */
                        {
                            next [k] = j ;
                            k = j ;
                            if ( t > *lambda )    /* j is bound */
                            {
                                nbound++ ;
                                bound_heap [nbound] = j ;
                                breakpts [j] = t ;
                            }
                            else                  /* j is free */
                            {
                                a2sum += aj*aj/dj ;
                                nfree++ ;
                                free_heap [nfree] = j ;
                                breakpts [j] = s ;
                            }
                        }
                    }
                }
            }
        }
        next [k] = j ;

        /* build heap for free variables */
        Stats->nfree = nfree ;
        napminheap_build (free_heap, breakpts, nfree) ;

        /* build heap for bound variables */
        Stats->nbound = nbound ;
        napminheap_build (bound_heap, breakpts, nbound) ;

        /* d is positive, lambda is to the left of the true multiplier, and
           the break points are stored in a heap.  Search over the break points
           to the right until finding 0 slope */
        if ( nfree > 0 )
        {
            jfree = free_heap [1] ;
            minfree = breakpts [jfree] ;
        }
        else minfree = INF ;

        if ( nbound > 0 )
        {
            jb = bound_heap [1] ;
            minbound = breakpts [jb] ;
        }
        else minbound = INF ;

        while ( 1 )
        {
            new_break = MIN (minfree, minbound) ;
            t = *slope - a2sum*(new_break - *lambda) ;
            if ( (t <= ZERO) || (new_break >= INF) ) /* done */
            {
                Stats->nbrks = (n-nf) + (n-nb) ;
                /* When refine is TRUE, better to do last update in
                   refinement phase */
                if ( nfree && !refine ) *lambda += *slope/a2sum ;
                break ;
            }
            *lambda = new_break ;
            *slope = t ;

            /* update the heaps */
            while ( *lambda == minfree )
            {
                /* delete break point from heap */
                nfree = napminheap_delete (free_heap, breakpts, nfree) ;

                /* add break point to list of previously free variables */
                free_heap [nf] = jfree ;
                nf-- ;

                /* update a2sum */
                a2sum -= a [jfree]*a [jfree]/d [jfree] ;

                /* get the smallest entry in heap */
                if ( nfree == 0 )
                {
                    minfree = INF ;
                    if ( a2sum < ZERO ) a2sum = ZERO ;
                    break ;
                }
                else
                {
                    jfree = free_heap [1] ;
                    minfree = breakpts [jfree] ;
                }
            }

            while ( *lambda == minbound )
            {
                /* delete break point from heap */
                nbound  = napminheap_delete(bound_heap, breakpts, nbound);

                /* add break point to list of previously bound  variables */
                bound_heap [nb] = jb ;
                nb-- ;

                /* compute new hi break point */
                aj = a [jb] ;
                dj = d [jb] ;
                if ( K > 0 )
                {
                    s = br_hi [jb] ;
                }
                else
                {
                    if ( aj > ZERO ) s = (y [jb] - lo [jb]*dj)/aj ;
                    else             s = (y [jb] - hi [jb]*dj)/aj ;
                }
                breakpts [jb] = s ;

                /* add new break point to free heap */
                nfree = napminheap_add (free_heap, breakpts, jb, nfree) ;

                /* get the smallest entry in free heap */
                jfree = free_heap [1] ;
                minfree = breakpts [jfree] ;

                /* update a2sum */
                a2sum += aj*aj/dj ;

                /* get the smallest entry in bound heap */
                if ( nbound == 0 ) minbound = INF ;
                else
                {
                    jb = bound_heap [1] ;
                    minbound = breakpts [jb] ;
                }
            }
        }
    }
    else if ( *slope < ZERO ) /* move left */
    {
        /* lambda is to the right of the true multiplier. Get ready to
           search over the break points to the left by constructing the heaps
           for currently bound and free variables, and by fixing variables
           whose break points are to the right. */
        goright = FALSE ;
        if ( K > 0 )
        {
            /* either variable fixing or Newton done, br_lo and br_hi exist */
            while ( (j = next [j]) < n )
            {
                s = br_lo [j] ;
                if ( s >= *lambda ) /* j is fixed */
                {
                    aj = a [j] ;
                    if ( aj > ZERO ) t = hi [j] ;
                    else             t = lo [j] ;
                    x [j] = t ;
                    b -= aj*t ;
                }
                else
                {
                    t = br_hi [j] ;
                    if ( t <= lambdaL ) /* j is fixed */
                    {
                        aj = a [j] ;
                        if ( aj > ZERO ) t = lo [j] ;
                        else             t = hi [j] ;
                        x [j] = t ;
                        b -= aj*t ;
                    }
                    else                        /* keep this variable */
                    {
                        next [k] = j ;
                        k = j ;
                        if ( t < *lambda )      /* j is bound */
                        {
                            nbound++ ;
                            bound_heap [nbound] = j ;
                            breakpts [j] = t ;
                        }
                        else                    /* j is free */
                        {
                            aj = a [j] ;
                            a2sum += aj*aj/d [j] ;
                            nfree++ ;
                            free_heap [nfree] = j ;
                            breakpts [j] = s ;
                        }
                    }
                }
            }
        }
        else
        {
            while ( (j = next [j]) < n )
            {
                aj = a [j] ;
                yj = y [j] ;
                dj = d [j] ;
                if ( aj > ZERO )
                {
                    hij = hi [j] ;
                    s = (yj - hij*dj)/aj ;      /* lo break point */
                    if ( s >= *lambda )         /* j is fixed */
                    {
                        x [j] = hij ;
                        b -= aj*hij ;
                    }
                    else
                    {
                        loj = lo [j] ;
                        t = (yj - loj*dj)/aj ;  /* hi break point */
                        if ( t <= lambdaL )     /* j is fixed */
                        {
                            x [j] = loj ;
                            b -= aj*loj ;
                        }
                        else                    /* keep this variable */
                        {
                            next [k] = j ;
                            k = j ;
                            if ( t < *lambda )  /* j is bound */
                            {
                                nbound++ ;
                                bound_heap [nbound] = j ;
                                breakpts [j] = t ;
                            }
                            else                /* j is free */
                            {
                                a2sum += aj*aj/dj ;
                                nfree++ ;
                                free_heap [nfree] = j ;
                                breakpts [j] = s ;
                            }
                        }
                    }
                }
                else
                {
                    loj = lo [j] ;
                    s = (yj - loj*dj)/aj ;      /* lo break point */
                    if ( s >= *lambda )         /* j is fixed */
                    {
                        x [j] = loj ;
                        b -= aj*loj ;
                    }
                    else
                    {
                        hij = hi [j] ;
                        t = (yj - hij*dj)/aj ;  /* hi break point */
                        if ( t <= lambdaL )     /* j is fixed */
                        {
                            x [j] = hij ;
                            b -= aj*hij ;
                        }
                        else                    /* keep this variable */
                        {
                            next [k] = j ;
                            k = j ;
                            if ( t < *lambda )  /* j is bound */
                            {
                                nbound++ ;
                                bound_heap [nbound] = j ;
                                breakpts [j] = t ;
                            }
                            else                /* j is free */
                            {
                                a2sum += aj*aj/dj ;
                                nfree++ ;
                                free_heap [nfree] = j ;
                                breakpts [j] = s ;
                            }
                        }
                    }
                }
            }
        }
        next [k] = j ;

        /* build the free heap */
        Stats->nfree = nfree ;
        napmaxheap_build (free_heap, breakpts, nfree) ;

        /* build the bound heap */
        Stats->nbound = nbound ;
        napmaxheap_build (bound_heap, breakpts, nbound) ;

        /* d is positive, lambda is to the right of the true multiplier, and
           the break points are stored in a heap.  Search over the break points
           to the left until finding 0 slope */
        if ( nfree > 0 )
        {
            jfree = free_heap [1] ;
            maxfree = breakpts [jfree] ;
        }
        else maxfree = -INF ;

        if ( nbound > 0 )
        {
            jb = bound_heap [1] ;
            maxbound = breakpts [jb] ;
        }
        else maxbound = -INF ;

        while ( 1 )
        {
            new_break = MAX (maxfree, maxbound) ;
            t = *slope - a2sum*(new_break - *lambda) ;
            if ( (t >= ZERO) || (new_break <= -INF) ) /* done */
            {
                Stats->nbrks = (n-nb) + (n-nf) ;
                /* When refine is TRUE, better to do last update in
                   refinement phase */
                if ( nfree && !refine ) *lambda += *slope/a2sum ;
                break ;
            }
            *lambda = new_break ;
            *slope = t ;

            /* update the heaps */
            while ( *lambda == maxfree )
            {
                /* delete break point from heap */
                nfree = napmaxheap_delete (free_heap, breakpts, nfree) ;

                /* add break point to list of previously free variables */
                free_heap [nf] = jfree ;
                nf-- ;

                /* update a2sum */
                a2sum -= a [jfree]*a [jfree]/d [jfree] ;

                /* get the smallest entry in heap */
                if ( nfree == 0 )
                {
                    maxfree = -INF ;
                    if ( a2sum < ZERO ) a2sum = ZERO ;
                    break ;
                }
                else
                {
                    jfree = free_heap [1] ;
                    maxfree = breakpts [jfree] ;
                }
            }
            while ( *lambda == maxbound )
            {
                /* delete break point from heap */
                nbound  = napmaxheap_delete(bound_heap, breakpts, nbound);

                /* add break point to list of previously bound variables */
                bound_heap [nb] = jb ;
                nb-- ;

                /* compute new lo break point */
                aj = a [jb] ;
                dj = d [jb] ;
                if ( K > 0 )
                {
                    s = br_lo [jb] ;
                }
                else
                {
                    if ( aj > ZERO ) s = (y [jb] - hi [jb]*dj)/aj ;
                    else             s = (y [jb] - lo [jb]*dj)/aj ;
                }
                breakpts [jb] = s ;

                /* add new break point to free heap */
                nfree = napmaxheap_add (free_heap, breakpts, jb, nfree) ;

                /* get the smallest entry in free heap */
                jfree = free_heap [1] ;
                maxfree = breakpts [jfree] ;

                /* update a2sum and evaluate new break point */
                a2sum += aj*aj/d [jb] ;

                /* get the largest entry in bound heap */
                if ( nbound == 0 ) maxbound = -INF ;
                else
                {
                    jb = bound_heap [1] ;
                    maxbound = breakpts [jb] ;
                }
            }
        }
    }
    else /* slope = 0, optimization solution found */
    {
        return ;
    }

    if ( refine )
    {
        /* evaluate slope at lambda */
        *slope = ZERO ;
        j = -1 ;
        while ( (j = next [j]) < n )
        {
            t = (y [j] - *lambda*a [j])/d [j] ;
            if      ( t < lo [j] ) t = lo [j] ;
            else if ( t > hi [j] ) t = hi [j] ;
            *slope += a [j]*t ;
        }
        *slope -= b ;

        Nb = nb ;
        Nf = nf ;
        if ( goright )
        {
            if ( *slope >  err )
            {
                /* d is positive, lambda is to the left of the true multiplier,
                   and the break points are stored in a heap.  Search over
                   the break points to the right until finding 0 slope */
                if ( nfree > 0 )
                {
                    jfree = free_heap [1] ;
                    minfree = breakpts [jfree] ;
                }
                else minfree = INF ;

                if ( nbound > 0 )
                {
                    jb = bound_heap [1] ;
                    minbound = breakpts [jb] ;
                }
                else minbound = INF ;

                while ( 1 )
                {
                    new_break = MIN (minfree, minbound) ;
                    t = *slope - a2sum*(new_break - *lambda) ;
                    if ( (t <= ZERO) || (new_break >= INF) ) /* done */
                    {
                        Stats->nbrks = (n-nf) + (n-nb) ;
                        /* When refine is TRUE, better to do last update in
                           refinement phase */
                        if ( nfree ) *lambda += *slope/a2sum ;
                        break ;
                    }
                    *lambda = new_break ;
                    *slope = t ;

                    /* update the heaps */
                    while ( *lambda == minfree )
                    {
                        /* delete break point from heap */
                        nfree = napminheap_delete (free_heap, breakpts, nfree) ;

                        /* add break point previously free list */
                        free_heap [nf] = jfree ;
                        nf-- ;

                        /* update a2sum */
                        a2sum -= a [jfree]*a [jfree]/d [jfree] ;

                        /* get the smallest entry in heap */
                        if ( nfree == 0 )
                        {
                            minfree = INF ;
                            if ( a2sum < ZERO ) a2sum = ZERO ;
                            break ;
                        }
                        else
                        {
                            jfree = free_heap [1] ;
                            minfree = breakpts [jfree] ;
                        }
                    }

                    while ( *lambda == minbound )
                    {
                        /* delete break point from heap */
                        nbound  = napminheap_delete(bound_heap, breakpts,
                                                    nbound);

                        /* add break point to previously bound list */
                        bound_heap [nb] = jb ;
                        nb-- ;

                        /* compute new hi break point */
                        aj = a [jb] ;
                        dj = d [jb] ;
                        if ( K > 0 )
                        {
                            s = br_hi [jb] ;
                        }
                        else
                        {
                            if ( aj > ZERO ) s = (y [jb] - lo [jb]*dj)/aj ;
                            else             s = (y [jb] - hi [jb]*dj)/aj ;
                        }
                        breakpts [jb] = s ;

                        /* add new break point to free heap */
                        nfree = napminheap_add (free_heap, breakpts, jb, nfree);

                        /* get the smallest entry in free heap */
                        jfree = free_heap [1] ;
                        minfree = breakpts [jfree] ;

                        /* update a2sum */
                        a2sum += aj*aj/dj ;

                        /* get the smallest entry in bound heap */
                        if ( nbound == 0 ) minbound = INF ;
                        else
                        {
                            jb = bound_heap [1] ;
                            minbound = breakpts [jb] ;
                        }
                    }
                }
                Stats->nbrks = (Nb - nb) + (Nf - nf) ;
            }
            else if ( *slope < -err )
            {
                /* Search over break points that we passed over already.
                   Apparently, due to rounding errors, we did not stop
                   as soon as we should have stopped. */
                if ( nb < n )
                {
                    nb++ ;
                    jb = bound_heap [nb] ;
                    if ( K > 0 )
                    {
                        maxbound = br_lo [jb] ;
                    }
                    else
                    {
                        if ( a [jb] > ZERO )
                        {
                            maxbound = (y [jb]-hi [jb]*d [jb])/a [jb] ;
                        }
                        else
                        {
                            maxbound = (y [jb]-lo [jb]*d [jb])/a [jb] ;
                        }
                    }
                }
                else maxbound = -INF ;

                if ( nf < n )
                {
                    nf++ ;
                    jfree = free_heap [nf] ;
                    maxfree = breakpts [jfree] ;
                }
                else maxfree = -INF ;

                Nb = nb ;
                Nf = nf ;

                while ( (maxfree > -INF) || (maxbound > -INF) )
                {
                    if ( maxfree >= maxbound ) new_break = maxfree ;
                    else                       new_break = maxbound ;
                    t = *slope - a2sum*(new_break - *lambda) ;
                    if ( t >= ZERO ) /* done */
                    {
                        break ;
                    }
                    *lambda = new_break ;
                    *slope = t ;
                    if ( *lambda == maxfree )
                    {
                        /* adjust a2sum since variable became free */
                        a2sum += a [jfree]*a [jfree]/d [jfree] ;
                        nf++ ;
                        if ( nf <= n )
                        {
                            jfree = free_heap [nf] ;
                            maxfree = breakpts [jfree] ;
                        }
                        else maxfree = -INF ;
                    }
                    else
                    {
                        /* adjust a2sum since variable became bound */
                        a2sum -= a [jb]*a [jb]/d [jb] ; /* becomes bound */
                        nb++ ;
                        if ( nb <= n )
                        {
                            jb = bound_heap [nb] ;
                            if ( K > 0 )
                            {
                                maxbound = br_lo [jb] ;
                            }
                            else
                            {
                                if ( a [jb] > ZERO )
                                {
                                    maxbound = (y [jb] - hi [jb]*d [jb])/a [jb];
                                }
                                else
                                {
                                    maxbound = (y [jb] - lo [jb]*d [jb])/a [jb];
                                }
                            }
                        }
                        else maxbound = -INF ;
                    }
                }
                if ( nfree - (nb-Nb) + (nf-Nf) ) *lambda += *slope/a2sum ;
                Stats->nrefine = (nb-Nb) + (nf-Nf) ;
            }
        }
        else
        {
            if ( *slope < -err )
            {
                /* d is positive, lambda is to the right of the true multiplier,
                   and the break points are stored in a heap.  Search over
                   the break points to the left until finding 0 slope */
                if ( nfree > 0 )
                {
                    jfree = free_heap [1] ;
                    maxfree = breakpts [jfree] ;
                }
                else maxfree = -INF ;

                if ( nbound > 0 )
                {
                    jb = bound_heap [1] ;
                    maxbound = breakpts [jb] ;
                }
                else maxbound = -INF ;

                while ( 1 )
                {
                    new_break = MAX (maxfree, maxbound) ;
                    t = *slope - a2sum*(new_break - *lambda) ;
                    if ( (t >= ZERO) || (new_break <= -INF) ) /* done */
                    {
                        Stats->nbrks = (n-nb) + (n-nf) ;
                        /* When refine is TRUE, better to do last update in
                           refinement phase */
                        if ( nfree ) *lambda += *slope/a2sum ;
                        break ;
                    }
                    *lambda = new_break ;
                    *slope = t ;

                    /* update the heaps */
                    while ( *lambda == maxfree )
                    {
                        /* delete break point from heap */
                        nfree = napmaxheap_delete (free_heap, breakpts, nfree) ;

                        /* add break point to previously free list */
                        free_heap [nf] = jfree ;
                        nf-- ;

                        /* update a2sum */
                        a2sum -= a [jfree]*a [jfree]/d [jfree] ;

                        /* get the smallest entry in heap */
                        if ( nfree == 0 )
                        {
                            maxfree = -INF ;
                            if ( a2sum < ZERO ) a2sum = ZERO ;
                            break ;
                        }
                        else
                        {
                            jfree = free_heap [1] ;
                            maxfree = breakpts [jfree] ;
                        }
                    }
                    while ( *lambda == maxbound )
                    {
                        /* delete break point from heap */
                        nbound  = napmaxheap_delete(bound_heap, breakpts,
                                                    nbound);

                        /* add break point to previously bound list */
                        bound_heap [nb] = jb ;
                        nb-- ;

                        /* compute new lo break point */
                        aj = a [jb] ;
                        dj = d [jb] ;
                        if ( K > 0 )
                        {
                            s = br_lo [jb] ;
                        }
                        else
                        {
                            if ( aj > ZERO ) s = (y [jb] - hi [jb]*dj)/aj ;
                            else             s = (y [jb] - lo [jb]*dj)/aj ;
                        }
                        breakpts [jb] = s ;

                        /* add new break point to free heap */
                        nfree = napmaxheap_add (free_heap, breakpts, jb, nfree);

                        /* get the smallest entry in free heap */
                        jfree = free_heap [1] ;
                        maxfree = breakpts [jfree] ;

                        /* update a2sum and evaluate new break point */
                        aj = a [jb] ;
                        a2sum += aj*aj/d [jb] ;

                        /* get the largest entry in bound heap */
                        if ( nbound == 0 ) maxbound = -INF ;
                        else
                        {
                            jb = bound_heap [1] ;
                            maxbound = breakpts [jb] ;
                        }
                    }
                }
                Stats->nbrks = (Nb - nb) + (Nf - nf) ;
            }
            else if ( *slope > err )
            {
                /* Search over break points that we passed over already.
                   Apparently, due to rounding errors, we did not stop
                   as soon as we should have stopped. */
                if ( nb < n )
                {
                    nb++ ;
                    jb = bound_heap [nb] ;
                    if ( K > 0 )
                    {
                        minbound = br_hi [jb] ;
                    }
                    else
                    {
                        if ( a [jb] > ZERO )
                        {
                            minbound = (y [jb]-lo [jb]*d [jb])/a [jb] ;
                        }
                        else
                        {
                            minbound = (y [jb]-hi [jb]*d [jb])/a [jb] ;
                        }
                    }
                }
                else minbound = INF ;

                if ( nf < n )
                {
                    nf++ ;
                    jfree = free_heap [nf] ;
                    minfree = breakpts [jfree] ;
                }
                else minfree = INF ;

                Nb = nb ;
                Nf = nf ;

                while ( (minfree < INF) || (minbound < INF) )
                {
                    if ( minfree <= minbound ) new_break = minfree ;
                    else                       new_break = minbound ;
                    t = *slope - a2sum*(new_break - *lambda) ;
                    if ( t <= ZERO ) /* done */
                    {
                        break ;
                    }
                    *lambda = new_break ;
                    *slope = t ;
                    if ( *lambda == minfree )
                    {
                        /* adjust a2sum since variable became free */
                        a2sum += a [jfree]*a [jfree]/d [jfree] ;
                        nf++ ;
                        if ( nf <= n )
                        {
                            jfree = free_heap [nf] ;
                            minfree = breakpts [jfree] ;
                        }
                        else minfree = INF ;
                    }
                    else
                    {
                        /* adjust a2sum since variable became bound */
                        a2sum -= a [jb]*a [jb]/d [jb] ;
                        nb++ ;
                        if ( nb <= n )
                        {
                            jb = bound_heap [nb] ;
                            if ( K > 0 ) minbound = br_hi [jb] ;
                            else
                            {
                                if ( a [jb] > ZERO )
                                {
                                    minbound = (y [jb] - lo [jb]*d [jb])/a [jb];
                                }
                                else
                                {
                                    minbound = (y [jb] - hi [jb]*d [jb])/a [jb];
                                }
                            }
                        }
                        else minbound = INF ;
                    }
                }
                if ( nfree - (nb-Nb) + (nf-Nf) ) *lambda += *slope/a2sum ;
                Stats->nrefine = (nb-Nb) + (nf-Nf) ;
            }
        }
    }
}

/*  =========================================================================
    === napsearch0 ==========================================================
    =========================================================================
    If the slope of the dual function at lambda is positive,
    then search over break points to the right. If the slope is negative,
    then search over break points to the left. Recompute the slope at the
    current lambda and possibly refine x. It is assumed that d is zero.
    =========================================================================*/

PRIVATE void napsearch0
(
    double        *x,    /* size n. solution (output) */
    double   *lambda,    /* optimal multiplier (input = guess, output = sol) */
    double   lambdaL,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double   lambdaR,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double        *b,    /* a'x = b */
    double    *slope,    /* slope excluding term associated with break point*/
    double  *slopelo,    /* lower subdifferential associated with break point */
    double  *slopehi,    /* upper subdifferential associated with break point */
    NAPHEAP_INT    n,    /* problem dimension */
    double const  *a,    /* size n, linear constraint vector */
    double const *lo,    /* size n, lower bounds on x */
    double const *hi,    /* size n, upper bounds on x */
    double const *breakpts, /* size n, breakpts [j] is br_hi[j] or br_lo[j] */
    NAPHEAP_INT *next,   /* size n+1, next[-1..n-1] link list of unbound var. */
    NAPHEAP_INT *bound_heap,    /* size n+1, heap of bound variables */
    int      goright,    /* if true, move right, else move left */
    napheap_stats *Stats,    /* solution statistics */
    int recompute_slope, /* if true, then recompute slope */
    napheap_parameter *Parm  /* parameters */
)
{
    int refine ;
    NAPHEAP_INT j, k, Nb, nb, nbound ;
    double aj, t, err ;

    /* initialization */
    err = Parm->err ; /* error threshhold that triggers refinement */
    refine = Parm->refine ; /* TRUE means to refine the solution */
    k = -1 ;
    j =  k ;
    nb = n ;       /* pointer to start of previously bound variables */
    nbound = 0 ;   /* number of break points */

    if ( goright ) /* move right */
    {
        /* d is zero and lambda is to the left of the true multiplier.
           Get ready to search over the break points to the right by
           constructing the heaps for currently bound,
           and by fixing variables whose break points are to the left.
           Moreover, if the slope at lambda is not known, then compute it. */
        if ( recompute_slope )
        {
            *slope = ZERO ;
            *slopehi = ZERO ;
            *slopelo = ZERO ;
            while ( (j = next [j]) < n )
            {
                aj = a [j] ;
                t = breakpts [j] ;
                if ( t < *lambda )
                {
                    if ( aj > ZERO ) t = lo [j] ;
                    else             t = hi [j] ;
                    *b -= aj*t ;
                    x [j] = t ;
                }
                else if ( t > lambdaR )
                {
                    if ( aj > ZERO ) t = hi [j] ;
                    else             t = lo [j] ;
                    *b -= aj*t ;
                    x [j] = t ;
                }
                else
                {
                    next [k] = j ;
                    k = j ;
                    if ( t > *lambda )
                    {
                        nbound++ ;
                        bound_heap [nbound] = j ;
                        if ( aj > ZERO ) *slope += aj*hi [j] ;
                        else             *slope += aj*lo [j] ;
                    }
                    else /* lambda = break point */
                    {
                        if ( aj > ZERO )
                        {
                            *slopehi += aj*hi [j] ;
                            *slopelo += aj*lo [j] ;
                        }
                        else
                        {
                            *slopelo += aj*hi [j] ;
                            *slopehi += aj*lo [j] ;
                        }
                    }
                }
            }
            *slope -= *b ;
            if ( *slope + *slopelo <= ZERO ) /* lambda is optimal */
            {
                next [k] = j ;
                return ;
            }
            *slope += *slopelo ;
        }
        else /* no need to compute slope, focus on heaps */
        {
            while ( (j = next [j]) < n )
            {
                t = breakpts [j] ;
                if ( t <= *lambda )
                {
                    aj = a [j] ;
                    if ( aj > ZERO ) t = lo [j] ;
                    else             t = hi [j] ;
                    *b -= aj*t ;
                    x [j] = t ;
                }
                else if ( t > lambdaR )
                {
                    aj = a [j] ;
                    if ( aj > ZERO ) t = hi [j] ;
                    else             t = lo [j] ;
                    *b -= aj*t ;
                    x [j] = t ;
                }
                else
                {
                    nbound++ ;
                    bound_heap [nbound] = j ;
                    next [k] = j ;
                    k = j ;
                }
            }
        }
        next [k] = j ;

        /* build the bound heap */
        Stats->nbound = nbound ;
        napminheap_build (bound_heap, breakpts, nbound) ;

        if ( nbound == 0 ) /* we cannot make the slope any smaller */
        {
            Stats->nbrks = n - nb ;
            return ;
        }
        /* d is zero, lambda is to the left of the true multiplier, and
           the break points are stored in a heap.  Search over the break
           points to the right until finding 0 slope */
        while ( *lambda == breakpts [bound_heap [1]] ) /* skip over lambda */
        {
            nbound = napminheap_delete (bound_heap, breakpts, nbound) ;
            if ( nbound == 0 ) break ;
        }
        while ( nbound > 0 )
        {

            j = bound_heap [1] ;
            *lambda = breakpts [j] ;
            aj = a [j] ;
            t = *slope - fabs (aj)*(hi [j] - lo [j]) ;
            if ( t <= ZERO )
            {
                break ; /* done */
            }
            *slope = t ;

            /* delete break point from heap */
            nbound = napminheap_delete (bound_heap, breakpts, nbound) ;

            /* save break point for potential later refinement */
            bound_heap [nb] = j ;
            nb-- ;
        }
    }
    else
    {
        /* d is zero and lambda is to the left of the true multiplier.
           Get ready to search over the break points to the right by
           constructing the heaps for currently bound,
           and by fixing variables whose break points are to the left.
           Moreover, if the slope at lambda is not known, then compute it. */
        if ( recompute_slope )
        {
            *slope = ZERO ;
            *slopehi = ZERO ;
            *slopelo = ZERO ;
            while ( (j = next [j]) < n )
            {
                aj = a [j] ;
                t = breakpts [j] ;
                if ( t > *lambda )
                {
                    if ( aj > ZERO ) t = hi [j] ;
                    else             t = lo [j] ;
                    *b -= aj*t ;
                    x [j] = t ;
                }
                else if ( t < lambdaL )
                {
                    if ( aj > ZERO ) t = lo [j] ;
                    else             t = hi [j] ;
                    *b -= aj*t ;
                    x [j] = t ;
                }
                else
                {
                    next [k] = j ;
                    k = j ;
                    if ( t < *lambda )
                    {
                        nbound++ ;
                        bound_heap [nbound] = j ;
                        if ( aj > ZERO ) *slope += aj*lo [j] ;
                        else             *slope += aj*hi [j] ;
                    }
                    else /* lambda = break point */
                    {
                        if ( aj > ZERO )
                        {
                            *slopehi += aj*hi [j] ;
                            *slopelo += aj*lo [j] ;
                        }
                        else
                        {
                            *slopelo += aj*hi [j] ;
                            *slopehi += aj*lo [j] ;
                        }
                    }
                }
            }
            *slope -= *b ;
            if ( *slope + *slopehi >= ZERO ) /* lambda is optimal */
            {
                next [k] = j ;
                return ;
            }
            *slope += *slopehi ;
        }
        else /* no need to compute slope, focus on heaps */
        {
            while ( (j = next [j]) < n )
            {
                t = breakpts [j] ;
                if ( t >= *lambda )
                {
                    aj = a [j] ;
                    if ( aj > ZERO ) t = hi [j] ;
                    else             t = lo [j] ;
                    *b -= aj*t ;
                    x [j] = t ;
                }
                else if ( t < lambdaL )
                {
                    aj = a [j] ;
                    if ( aj > ZERO ) t = lo [j] ;
                    else             t = hi [j] ;
                    *b -= aj*t ;
                    x [j] = t ;
                }
                else
                {
                    nbound++ ;
                    bound_heap [nbound] = j ;
                    next [k] = j ;
                    k = j ;
                }
            }
        }
        next [k] = j ;

        /* build the bound heap */
        Stats->nbound = nbound ;
        napmaxheap_build (bound_heap, breakpts, nbound) ;

        /* d is zero, lambda is to the right of the true multiplier, and
           the break points are stored in a heap.  Search over the break points
           to the left until finding 0 slope */
        if ( nbound == 0 ) /* we cannot make the slope any smaller */
        {
            Stats->nbrks = n - nb ;
            return ;
        }
        /* skip over breakpoints = lambda */
        while ( *lambda == breakpts [bound_heap [1]] )
        {
            nbound = napmaxheap_delete (bound_heap, breakpts, nbound) ;
            if ( nbound == 0 ) break ;
        }
        while ( nbound )
        {
            j = bound_heap [1] ;
            *lambda = breakpts [j] ;
            aj = a [j] ;
            *slope += fabs (aj)*(hi [j] - lo [j]) ;
            if ( *slope >= ZERO )
            {
                break ;
            }

            /* delete break point from heap */
            nbound = napmaxheap_delete (bound_heap, breakpts, nbound) ;

            /* save break point for potential later refinement */
            bound_heap [nb] = j ;
            nb-- ;
        }
    }
    Stats->nbrks = n - nb ;

    if ( refine )
    {
        /* evaluate slope at lambda */
        *slopehi  = ZERO ; /* upper bound on slope at given lambda */
        *slopelo  = ZERO ; /* lower bound on slope at given lambda */
        *slope = ZERO ;    /* slope ignoring any break point at lambda */
        j = -1 ;

        while ( (j = next [j]) < n )
        {
            aj = a [j] ;
            t = breakpts [j] ;
            /* update the slope of dual function */
            if ( *lambda < t )
            {
                if ( aj > ZERO ) *slope += aj*hi [j] ;
                else             *slope += aj*lo [j] ;
            }
            else if ( *lambda > t )
            {
                if ( aj > ZERO ) *slope += aj*lo [j] ;
                else             *slope += aj*hi [j] ;
            }
            else /* lambda is a break point */
            {
                if ( aj > ZERO )
                {
                    *slopehi += aj*hi [j] ;
                    *slopelo += aj*lo [j] ;
                }
                else
                {
                    *slopelo += aj*hi [j] ;
                    *slopehi += aj*lo [j] ;
                }
            }
        }
        *slope -= *b ;

        /* check if lambda is optimal */
        if ( (*slope+*slopelo <= ZERO) && (*slope+*slopehi >= ZERO) )
        {
            return ;
        }

        /* refine lambda */
        Nb = nb ;

        if ( goright )
        {
            if ( *slope + *slopelo >  err )
            {
                *slope += *slopelo ;
                if ( nbound == 0 ) /* we cannot make the slope any smaller */
                {
                    Stats->nbrks = n - nb ;
                    return ;
                }
                /* d is zero, lambda is to the left of the true multiplier, and
                   the break points are stored in a heap.  Search over the break
                   points to the right until finding 0 slope. Skip over the
                   break points equal to lambda */
                while ( *lambda == breakpts [bound_heap [1]] )
                {
                    nbound = napminheap_delete (bound_heap, breakpts, nbound) ;
                    if ( nbound == 0 ) break ;
                }
                while ( nbound > 0 )
                {

                    j = bound_heap [1] ;
                    *lambda = breakpts [j] ;
                    aj = a [j] ;
                    t = *slope - fabs (aj)*(hi [j] - lo [j]) ;
                    if ( t <= ZERO )
                    {
                        break ; /* done */
                    }
                    *slope = t ;

                    /* delete break point from heap */
                    nbound = napminheap_delete (bound_heap, breakpts, nbound) ;

                    /* save break point for potential later refinement */
                    bound_heap [nb] = j ;
                    nb-- ;
                }
                Stats->nrefine = Nb - nb ;
            }
            else if ( *slope + *slopehi < -err )
            {
                for (nb++; nb <= n; nb++)
                {
                    j = bound_heap [nb] ;
                    if ( *lambda > breakpts [j] )
                    {
                        break ;
                    }
                }
                *slope += *slopehi ;
                for (; nb <= n; nb++)
                {
                    j = bound_heap [nb] ;
                    t = *slope + fabs (a [j])*(hi [j] - lo [j]) ;
                    if ( t >= ZERO )
                    {
                        *lambda = breakpts [j] ;
                        break ;
                    }
                    *slope = t ;
                }
                /* with exact arithmetic, terminate inside the loop above */
                Stats->nrefine = nb - Nb ;
            }
        }
        else /* go left */
        {
            if ( *slope + *slopehi < -err ) /* continue going left */
            {
                *slope += *slopehi ;
                /* d is zero, lambda is to the right of the true multiplier, and
                   the break points are stored in a heap.  Search over the break
                   points to the left until finding 0 slope */
                if ( nbound == 0 ) /* we cannot make the slope any smaller */
                {
                    Stats->nbrks = n - nb ;
                    return ;
                }
                /* skip over breakpoints = lambda */
                while ( *lambda == breakpts [bound_heap [1]] )
                {
                    nbound = napmaxheap_delete (bound_heap, breakpts, nbound) ;
                    if ( nbound == 0 ) break ;
                }
                while ( nbound )
                {
                    j = bound_heap [1] ;
                    *lambda = breakpts [j] ;
                    aj = a [j] ;
                    *slope += fabs (aj)*(hi [j] - lo [j]) ;
                    if ( *slope >= ZERO )
                    {
                        break ;
                    }

                    /* delete break point from heap */
                    nbound = napmaxheap_delete (bound_heap, breakpts, nbound) ;

                    /* save break point for potential later refinement */
                    bound_heap [nb] = j ;
                    nb-- ;
                }
                Stats->nrefine = Nb - nb ;
            }
            else if ( *slope + *slopelo >  err ) /* switch to right */
            {
                *slope += *slopelo ;
                for (nb++; nb <= n; nb++)
                {
                    j = bound_heap [nb] ;
                    if ( *lambda < breakpts [j] )
                    {
                        break ;
                    }
                }
                for (; nb <= n; nb++)
                {
                    j = bound_heap [nb] ;
                    t = *slope - fabs (a [j])*(hi [j] - lo [j]) ;
                    if ( t <= ZERO )
                    {
                        *lambda = breakpts [j] ;
                        break ;
                    }
                    *slope = t ;
                }
                /* with exact arithmetic, terminate inside the loop above */
                Stats->nrefine = nb - Nb ;
            }
        }
    }
}

/*  =========================================================================
    === napsearch1 ==========================================================
    =========================================================================
    If the slope of the dual function at lambda is positive,
    then search over break points to the right. If the slope is negative,
    then search over break points to the left. Recompute the slope at the
    current lambda and possibly refine x. It is assumed that d is neither
    identically 0 nor identically positive.
    =========================================================================*/

PRIVATE void napsearch1
(
    double        *x,    /* size n. solution (output) */
    double   *lambda,    /* optimal multiplier (input = guess, output = sol) */
    double   lambdaL,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double   lambdaR,    /* optimal multiplier lies on [lambdaL, lambdaR] */
    double        *b,    /* a'x = b */
    double    *slope,    /*  slope excluding term associated with break point*/
    double  *slopelo,    /* lower subdifferential associated with break point */
    double  *slopehi,    /* upper subdifferential associated with break point */
    double const  *y,    /* size n. linear term in objective function */
    double const  *d,    /* size n, diagonal of objective Hessian */
    double const  *a,    /* size n, linear constraint vector */
    double const *lo,    /* size n, lower bounds on x */
    double const *hi,    /* size n, upper bounds on x */
    NAPHEAP_INT    n,    /* problem dimension */
    NAPHEAP_INT    K,    /* number iterations in variable fixing method */
    double *breakpts,    /* size n, breakpts [j] is br_hi[j] or br_lo[j] */
    NAPHEAP_INT *next,   /* size n+1, next[-1..n-1] link list of unbound var. */
    NAPHEAP_INT *free_heap,     /* size n+1, heap of free variables */
    NAPHEAP_INT *bound_heap,    /* size n+1, heap of bound variables */
    int      goright,    /* if true, move right, else move left */
    napheap_stats *Stats,    /* solution statistics */
    double const *br_lo, /* size n, br_lo[j] lo break pt for x[j] */
    double const *br_hi, /* size n. br_hi[j] hi break pt for x[j] */
    int recompute_slope, /* if true, then recompute slope */
    napheap_parameter *Parm  /* parameters */
)
{
    int refine, status ;
    NAPHEAP_INT j, jb, jfree, k, Dfree, nb, nf, Nb, Nf, nbound, nfree ;
    double a2sum, aj, dj, yj, hij, loj, s, t, err,
           minbound, minfree, maxbound, maxfree, new_break ;

    /* initialization */
    err = Parm->err ; /* error threshhold that triggers refinement */
    refine = Parm->refine ; /* TRUE means to refine the solution */
    nbound = 0 ;   /* number of variables that start out bound */
    nfree = 0 ;    /* number of variables that start out free */
    nb = n ;       /* pointer to start of previously bound variables */
    nf = n ;       /* pointer to start of previously free variables */
    a2sum = ZERO ; /* sum aj*aj/dj for free variables */
    k = -1 ;
    j =  k ;
    status = NAPHEAP_STATUS_KEEP_LOOKING ; /* default status */

    if ( goright ) /* move right */
    {
        /* d contains both positive and zero entries and lambda is to the
           left of the true multiplier.  Get ready to search over the
           break points to the right by constructing the heaps for currently
           bound and free variables, and by fixing variables whose break
           points are to the left.  Moreover, if the slope at lambda is
           not known, then compute it. */
        if ( recompute_slope )
        {
            *slope = ZERO ;
            *slopehi = ZERO ;
            *slopelo = ZERO ;
            if ( K > 0 ) /* variable fixing done, br_lo and br_hi exist */
            {
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    dj = d [j] ;
                    if ( dj > ZERO )
                    {
                        t = br_lo [j] ;
                        if ( t >= lambdaR ) /* j is fixed */
                        {
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            x [j] = t ;
                            *b -= aj*t ;
                        }
                        else                     /* keep this variable */
                        {
                            next [k] = j ;
                            k = j ;
                            if ( t > *lambda )   /* j is bound */
                            {
                                nbound++ ;
                                bound_heap [nbound] = j ;
                                breakpts [j] = t ;
                                if ( aj > ZERO ) *slope += aj*hi [j] ;
                                else             *slope += aj*lo [j] ;
                            }
                            else                 /* j is free */
                            {
                                a2sum += aj*aj/dj ;
                                *slope += aj*(y [j] - aj*(*lambda))/dj ;
                                nfree++ ;
                                free_heap [nfree] = j ;
                                breakpts [j] = br_hi [j] ;
                            }
                        }
                    }
                    else /* d [j] = 0 */
                    {
                        t = breakpts [j] ;
                        if ( t > lambdaR )
                        {
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else
                        {
                            next [k] = j ;
                            k = j ;
                            if ( t > *lambda )
                            {
                                nbound++ ;
                                bound_heap [nbound] = j ;
                                if ( aj > ZERO ) t = hi [j] ;
                                else             t = lo [j] ;
                                *slope += aj*t ;
                            }
                            else /* lambda = break point */
                            {
                                if ( aj > ZERO )
                                {
                                    *slopehi += aj*hi [j] ;
                                    *slopelo += aj*lo [j] ;
                                }
                                else
                                {
                                    *slopelo += aj*hi [j] ;
                                    *slopehi += aj*lo [j] ;
                                }
                            }
                        }
                    }
                }
            }
            else /* K = 0 */
            {
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    dj = d [j] ;
                    if ( dj > ZERO )
                    {
                        yj = y [j] ;
                        if ( aj > ZERO )
                        {
                            loj = lo [j] ;
                            s = (yj - loj*dj)/aj ;    /* hi break point */
                            if ( s <= *lambda )       /* j is fixed */
                            {
                                x [j] = loj ;
                                *b -= aj*loj ;
                            }
                            else
                            {
                                hij = hi [j] ;
                                t = (yj - hij*dj)/aj ;/* lo break point */
                                if ( t >= lambdaR )   /* j is fixed */
                                {
                                    x [j] = hij ;
                                    *b -= aj*hij ;
                                }
                                else                  /* keep this variable*/
                                {
                                    next [k] = j ;
                                    k = j ;
                                    if ( t > *lambda )/* j is bound */
                                    {
                                        *slope += aj*hij ;
                                        nbound++ ;
                                        bound_heap [nbound] = j ;
                                        breakpts [j] = t ;
                                    }
                                    else              /* j is free */
                                    {
                                        a2sum += aj*aj/dj ;
                                        *slope += aj*(yj - aj*(*lambda))/dj ;
                                        breakpts [j] = s ;
                                        nfree++ ;
                                        free_heap [nfree] = j ;
                                    }
                                }
                            }
                        }
                        else /* aj < 0 */
                        {
                            hij = hi [j] ;
                            s = (yj - hij*dj)/aj ;    /* hi break point */
                            if ( s <= *lambda )       /* j is fixed */
                            {
                                x [j] = hij ;
                                *b -= aj*hij ;
                            }
                            else
                            {
                                loj = lo [j] ;
                                t = (yj - loj*dj)/aj ;/* lo break point */
                                if ( t >= lambdaR )   /* j is fixed */
                                {
                                    x [j] = loj ;
                                    *b -= aj*loj ;
                                }
                                else                  /* keep this variable*/
                                {
                                    next [k] = j ;
                                    k = j ;
                                    if ( t > *lambda )/* j is bound */
                                    {
                                        *slope += aj*loj ;
                                        nbound++ ;
                                        bound_heap [nbound] = j ;
                                        breakpts [j] = t ;
                                    }
                                    else              /* j is free */
                                    {
                                        a2sum += aj*aj/dj ;
                                        *slope += aj*(yj - aj*(*lambda))/dj ;
                                        nfree++ ;
                                        free_heap [nfree] = j ;
                                        breakpts [j] = s ;
                                    }
                                }
                            }
                        }
                    }
                    else /* d [j] = 0 */
                    {
                        t = breakpts [j] ;
                        if ( t < *lambda ) /* fix xj */
                        {
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else if ( t > lambdaR )
                        {
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else
                        {
                            next [k] = j ;
                            k = j ;
                            if ( t > *lambda )
                            {
                                nbound++ ;
                                bound_heap [nbound] = j ;
                                if ( aj > ZERO ) *slope += aj*hi [j] ;
                                else             *slope += aj*lo [j] ;
                            }
                            else /* lambda = break point */
                            {
                                if ( aj > ZERO )
                                {
                                    *slopehi += aj*hi [j] ;
                                    *slopelo += aj*lo [j] ;
                                }
                                else
                                {
                                    *slopelo += aj*hi [j] ;
                                    *slopehi += aj*lo [j] ;
                                }
                            }
                        }
                    }
                }
            }
            *slope -= *b ;
            if ( *slope + *slopelo <= ZERO ) /* lambda is optimal */
            {
                status = NAPHEAP_STATUS_OK ;
            }
            else
            {
                *slope += *slopelo ;
            }
        }
        else /* do not recompute slope */
        {
            if ( K > 0 ) /* variable fixing done, br_lo and br_hi exist */
            {
                while ( (j = next [j]) < n )
                {
                    dj = d [j] ;
                    if ( dj > ZERO )
                    {
                        s = br_hi [j] ;
                        if ( s <= *lambda )           /* j is fixed */
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            x [j] = t ;
                            *b -= aj*t ;
                        }
                        else
                        {
                            t = br_lo [j] ;
                            if ( t  >= lambdaR )      /* j is fixed */
                            {
                                aj = a [j] ;
                                if ( aj > ZERO ) t = hi [j] ;
                                else             t = lo [j] ;
                                x [j] = t ;
                                *b -= aj*t ;
                            }
                            else                      /* keep this variable */
                            {
                                next [k] = j ;
                                k = j ;
                                if ( t > *lambda )    /* j is bound */
                                {
                                    nbound++ ;
                                    bound_heap [nbound] = j ;
                                    breakpts [j] = t ;
                                }
                                else                  /* j is free */
                                {
                                    aj = a [j] ;
                                    a2sum += aj*aj/dj ;
                                    nfree++ ;
                                    free_heap [nfree] = j ;
                                    breakpts [j] = s ;
                                }
                            }
                        }
                    }
                    else /* d [j] = 0 */
                    {
                        t = breakpts [j] ;
                        if ( t <= *lambda )
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else if ( t > lambdaR )
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else
                        {
                            nbound++ ;
                            bound_heap [nbound] = j ;
                            next [k] = j ;
                            k = j ;
                        }
                    }
                }
            }
            else
            {
                while ( (j = next [j]) < n )
                {
                    dj = d [j] ;
                    if ( dj > ZERO )
                    {
                        yj = y [j] ;
                        aj = a [j] ;
                        if ( aj > ZERO )
                        {
                            loj = lo [j] ;
                            s = (yj - loj*dj)/aj ;    /* hi break point */
                            if ( s <= *lambda )       /* j is fixed */
                            {
                                x [j] = loj ;
                                *b -= aj*loj ;
                            }
                            else
                            {
                                hij = hi [j] ;
                                t = (yj - hij*dj)/aj ;/* lo break point */
                                if ( t >= lambdaR )   /* j is fixed */
                                {
                                    x [j] = hij ;
                                    *b -= aj*hij ;
                                }
                                else                  /* keep this variable*/
                                {
                                    next [k] = j ;
                                    k = j ;
                                    if ( t > *lambda )/* j is bound */
                                    {
                                        nbound++ ;
                                        bound_heap [nbound] = j ;
                                        breakpts [j] = t ;
                                    }
                                    else              /* j is free */
                                    {
                                        a2sum += aj*aj/dj ;
                                        nfree++ ;
                                        free_heap [nfree] = j ;
                                        breakpts [j] = s ;
                                    }
                                }
                            }
                        }
                        else
                        {
                            hij = hi [j] ;
                            s = (yj - hij*dj)/aj ;    /* hi break point */
                            if ( s <= *lambda )       /* j is fixed */
                            {
                                x [j] = hij ;
                                *b -= aj*hij ;
                            }
                            else
                            {
                                loj = lo [j] ;
                                t = (yj - loj*dj)/aj ;/* lo break point */
                                if ( t >= lambdaR )   /* j is fixed */
                                {
                                    x [j] = loj ;
                                    *b -= aj*loj ;
                                }
                                else                  /* keep this variable*/
                                {
                                    next [k] = j ;
                                    k = j ;
                                    if ( t > *lambda )/* j is bound */
                                    {
                                        nbound++ ;
                                        bound_heap [nbound] = j ;
                                        breakpts [j] = t ;
                                    }
                                    else              /* j is free */
                                    {
                                        a2sum += aj*aj/dj ;
                                        nfree++ ;
                                        free_heap [nfree] = j ;
                                        breakpts [j] = s ;
                                    }
                                }
                            }
                        }
                    }
                    else /* d [j] = 0 */
                    {
                        t = breakpts [j] ;
                        if ( t <= *lambda ) /* fix xj */
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else if ( t > lambdaR )
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else
                        {
                            nbound++ ;
                            bound_heap [nbound] = j ;
                            next [k] = j ;
                            k = j ;
                        }
                    }
                }
            }
        }
        next [k] = j ;
        Stats->nfree = nfree ;
        Stats->nbound = nbound ;
        if ( status == NAPHEAP_STATUS_OK ) /* lambda was optimal */
        {
            return ;
        }

        /* build minheap of free variables */
        napminheap_build (free_heap, breakpts, nfree) ;

        /* build minheap of bound variables */
        napminheap_build (bound_heap, breakpts, nbound) ;

        /* d contains both zero and nonzero entries, lambda is to the left
           of the true multiplier, and the break points are stored in a heap.
           Search over the break points to the right until finding 0 slope */
        if ( nfree > 0 )
        {
            jfree = free_heap [1] ;
            minfree = breakpts [jfree] ;
        }
        else minfree = INF ;
        if ( nbound > 0 )
        {
            jb = bound_heap [1] ;
            minbound = breakpts [jb] ;
            while ( (minbound == *lambda) && nbound )
            {
                if ( d [jb] > ZERO )
                {
                    a2sum -= a [jb]*a [jb]/d [jb] ;
                }
                nbound = napminheap_delete(bound_heap, breakpts, nbound);
                if ( nbound )
                {
                    jb = bound_heap [1] ;
                    minbound = breakpts [jb] ;
                }
                else minbound = INF ;
            }
        }
        else minbound = INF ;

        while ( 1 )
        {
            if ( minfree < minbound )
            {
                new_break = minfree ;
                j = jfree ;
            }
            else
            {
                new_break = minbound ;
                j = jb ;
            }

            t = *slope - a2sum*(new_break - *lambda) ;
            if ( (t <= ZERO) || (new_break >= INF) ) /* done */
            {
                if ( nfree && !refine ) *lambda += *slope/a2sum ;
                break ;
            }
            *lambda = new_break ;

            dj = d [j] ;
            if ( dj ) *slope = t ;
            else
            {
                *slope = t - fabs (a [j])*(hi [j] - lo [j]) ;
                if ( *slope <= ZERO )
                {
                    break ;
                }
            }

            /* update the heaps */
            if ( *lambda == minfree ) /* only happens when dj > 0 */
            {
                /* delete break point from heap */
                nfree = napminheap_delete (free_heap, breakpts, nfree) ;

                /* add break point to list of previously free variables */
                free_heap [nf] = jfree ;
                nf-- ;

                /* update a2sum */
                a2sum -= a [jfree]*a [jfree]/dj ;

                /* get the smallest entry in heap */
                if ( nfree == 0 )
                {
                    minfree = INF ;
                    if ( a2sum < ZERO ) a2sum = ZERO ;
                }
                else
                {
                    jfree = free_heap [1] ;
                    minfree = breakpts [jfree] ;
                }
            }
            else
            {
                /* delete break point from heap */
                nbound  = napminheap_delete(bound_heap, breakpts, nbound);

                /* save break point in list of previously bound variables */
                bound_heap [nb] = jb ;
                nb-- ;

                if ( dj ) /* only do the following when dj > 0 */
                {
                    aj = a [jb] ;
                    if ( K > 0 )
                    {
                        s = br_hi [jb] ;      /* new break point in br_hi */
                    }
                    else                      /* compute break point */
                    {
                        if ( aj > ZERO ) s = (y [j] - lo [j]*dj)/aj ;
                        else             s = (y [j] - hi [j]*dj)/aj ;
                    }

                    breakpts [jb] = s ;
                    nfree = napminheap_add (free_heap, breakpts, jb, nfree);

                    /* get the smallest entry in free heap */
                    jfree = free_heap [1] ;
                    minfree = breakpts [jfree] ;

                    /* update a2sum */
                    a2sum += aj*aj/dj ;
                }

                /* get the smallest entry in bound heap */
                if ( nbound == 0 ) minbound = INF ;
                else
                {
                    jb = bound_heap [1] ;
                    minbound = breakpts [jb] ;
                }
            }
        }
        Stats->nbrks = (n-nb) + (n-nf) ;
    }
    else
    {
        /* d contains both positive and zero entries and lambda is to the
           right of the true multiplier.  Get ready to search over the
           break points to the left by constructing the heaps for currently
           bound and free variables , and by fixing variables whose break
           points are to the right.  Moreover, if the slope at lambda is
           not known, then compute it. (*/
        if ( recompute_slope ) /* lambda was outside domain of dual function */
        {
            *slope = ZERO ;
            *slopehi = ZERO ;
            *slopelo = ZERO ;
            if ( K > 0 ) /* variable fixing done, br_lo and br_hi exist */
            {
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    dj = d [j] ;
                    if ( dj > ZERO )
                    {
                        t = br_hi [j] ;
                        if ( t <= lambdaL ) /* j is fixed */
                        {
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            x [j] = t ;
                            *b -= aj*t ;
                        }
                        else                          /* keep this variable */
                        {
                            next [k] = j ;
                            k = j ;
                            if ( t < *lambda )        /* j is bound */
                            {
                                nbound++ ;
                                bound_heap [nbound] = j ;
                                breakpts [j] = t ;
                                if ( aj > ZERO ) *slope += aj*lo [j] ;
                                else             *slope += aj*hi [j] ;
                            }
                            else                      /* j is free */
                            {
                                a2sum += aj*aj/dj ;
                                *slope += aj*(y [j] - aj*(*lambda))/dj ;
                                nfree++ ;
                                free_heap [nfree] = j ;
                                breakpts [j] = br_lo [j] ;
                            }
                        }
                    }
                    else /* d [j] = 0 */
                    {
                        t = breakpts [j] ;
                        if ( t < lambdaL )
                        {
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else
                        {
                            next [k] = j ;
                            k = j ;
                            if ( t < *lambda )
                            {
                                nbound++ ;
                                bound_heap [nbound] = j ;
                                if ( aj > ZERO ) *slope += aj*lo [j] ;
                                else             *slope += aj*hi [j] ;
                            }
                            else /* lambda = break point */
                            {
                                if ( aj > ZERO )
                                {
                                    *slopehi += aj*hi [j] ;
                                    *slopelo += aj*lo [j] ;
                                }
                                else
                                {
                                    *slopelo += aj*hi [j] ;
                                    *slopehi += aj*lo [j] ;
                                }
                            }
                        }
                    }
                }
            }
            else /* K = 0 */
            {
                while ( (j = next [j]) < n )
                {
                    aj = a [j] ;
                    dj = d [j] ;
                    if ( dj > ZERO )
                    {
                        yj = y [j] ;
                        if ( aj > ZERO )
                        {
                            hij = hi [j] ;
                            s = (yj - hij*dj)/aj ;    /* lo break point */
                            if ( s >= *lambda )       /* j is fixed */
                            {
                                x [j] = hij ;
                                *b -= aj*hij ;
                            }
                            else
                            {
                                loj = lo [j] ;
                                t = (yj - loj*dj)/aj ;/* hi break point */
                                if ( t < lambdaL )    /* j is fixed */
                                {
                                    x [j] = loj ;
                                    *b -= aj*loj ;
                                }
                                else                  /* keep this variable*/
                                {
                                    next [k] = j ;
                                    k = j ;
                                    if ( t < *lambda )/* j is bound */
                                    {
                                        *slope += aj*loj ;
                                        nbound++ ;
                                        bound_heap [nbound] = j ;
                                        breakpts [j] = t ;
                                    }
                                    else              /* j is free */
                                    {
                                        a2sum += aj*aj/dj ;
                                        *slope += aj*(yj - aj*(*lambda))/dj ;
                                        nfree++ ;
                                        free_heap [nfree] = j ;
                                        breakpts [j] = s ;
                                    }
                                }
                            }
                        }
                        else /* aj < 0 */
                        {
                            loj = lo [j] ;
                            s = (yj - loj*dj)/aj ;    /* lo break point */
                            if ( s >= *lambda )       /* j is fixed */
                            {
                                x [j] = loj ;
                                *b -= aj*loj ;
                            }
                            else
                            {
                                hij = hi [j] ;
                                t = (yj - hij*dj)/aj ;/* hi break point */
                                if ( t <= lambdaL )   /* j is fixed */
                                {
                                    x [j] = hij ;
                                    *b -= aj*hij ;
                                }
                                else                  /* keep this variable*/
                                {
                                    next [k] = j ;
                                    k = j ;
                                    if ( t < *lambda )/* j is bound */
                                    {
                                        *slope += aj*hij ;
                                        nbound++ ;
                                        bound_heap [nbound] = j ;
                                        breakpts [j] = t ;
                                    }
                                    else              /* j is free */
                                    {
                                        a2sum += aj*aj/dj ;
                                        *slope += aj*(yj - aj*(*lambda))/dj ;
                                        nfree++ ;
                                        free_heap [nfree] = j ;
                                        breakpts [j] = s ;
                                    }
                                }
                            }
                        }
                    }
                    else /* d [j] = 0 */
                    {
                        t = breakpts [j] ;
                        if ( t > *lambda ) /* fix xj */
                        {
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else if ( t < lambdaL )
                        {
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else
                        {
                            next [k] = j ;
                            k = j ;
                            if ( t < *lambda )
                            {
                                nbound++ ;
                                bound_heap [nbound] = j ;
                                if ( aj > ZERO ) *slope += aj*lo [j] ;
                                else             *slope += aj*hi [j] ;
                            }
                            else
                            {
                                if ( aj > ZERO )
                                {
                                    *slopehi += aj*hi [j] ;
                                    *slopelo += aj*lo [j] ;
                                }
                                else
                                {
                                    *slopelo += aj*hi [j] ;
                                    *slopehi += aj*lo [j] ;
                                }
                            }
                        }
                    }
                }
            }

            *slope -= *b ;
            if ( *slope + *slopehi >= ZERO ) /* lambda is optimal */
            {
                status = NAPHEAP_STATUS_OK ;
            }
            else
            {
                *slope += *slopehi ;
            }
        }
        else /* do not recompute slope */
        {
            if ( K > 0 ) /* variable fixing done, br_lo and br_hi exist */
            {
                while ( (j = next [j]) < n )
                {
                    dj = d [j] ;
                    if ( dj > ZERO )
                    {
                        s = br_lo [j] ;
                        if ( s >= *lambda )           /* j is fixed */
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            x [j] = t ;
                            *b -= aj*t ;
                        }
                        else
                        {
                            t = br_hi [j] ;
                            if ( t <= lambdaL )       /* j is fixed */
                            {
                                aj = a [j] ;
                                if ( aj > ZERO ) t = lo [j] ;
                                else             t = hi [j] ;
                                x [j] = t ;
                                *b -= aj*t ;
                            }
                            else                      /* keep this variable */
                            {
                                next [k] = j ;
                                k = j ;
                                if ( t < *lambda )    /* j is bound */
                                {
                                    nbound++ ;
                                    bound_heap [nbound] = j ;
                                    breakpts [j] = t ;
                                }
                                else                  /* j is free */
                                {
                                    aj = a [j] ;
                                    a2sum += aj*aj/dj ;
                                    nfree++ ;
                                    free_heap [nfree] = j ;
                                    breakpts [j] = s ;
                                }
                            }
                        }
                    }
                    else /* d [j] = 0 */
                    {
                        t = breakpts [j] ;
                        if ( t >= *lambda ) /* fix xj */
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else if ( t < lambdaL )
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else
                        {
                            nbound++ ;
                            bound_heap [nbound] = j ;
                            next [k] = j ;
                            k = j ;
                        }
                    }
                }
            }
            else /* K = 0 */
            {
                while ( (j = next [j]) < n )
                {
                    dj = d [j] ;
                    if ( dj > ZERO )
                    {
                        yj = y [j] ;
                        aj = a [j] ;
                        if ( aj > ZERO )
                        {
                            hij = hi [j] ;
                            s = (yj - hij*dj)/aj ;    /* lo break point */
                            if ( s >= *lambda )       /* j is fixed */
                            {
                                x [j] = hij ;
                                *b -= aj*hij ;
                            }
                            else
                            {
                                loj = lo [j] ;
                                t = (yj - loj*dj)/aj ;/* hi break point */
                                if ( t < lambdaL )    /* j is fixed */
                                {
                                    x [j] = loj ;
                                    *b -= aj*loj ;
                                }
                                else                  /* keep this variable*/
                                {
                                    next [k] = j ;
                                    k = j ;
                                    if ( t < *lambda )/* j is bound */
                                    {
                                        nbound++ ;
                                        bound_heap [nbound] = j ;
                                        breakpts [j] = t ;
                                    }
                                    else              /* j is free */
                                    {
                                        a2sum += aj*aj/dj ;
                                        nfree++ ;
                                        free_heap [nfree] = j ;
                                        breakpts [j] = s ;
                                    }
                                }
                            }
                        }
                        else
                        {
                            loj = lo [j] ;
                            s = (yj - loj*dj)/aj ;    /* lo break point */
                            if ( s >= *lambda )       /* j is fixed */
                            {
                                x [j] = loj ;
                                *b -= aj*loj ;
                            }
                            else
                            {
                                hij = hi [j] ;
                                t = (yj - hij*dj)/aj ;/* hi break point */
                                if ( t <= lambdaL )   /* j is fixed */
                                {
                                    x [j] = hij ;
                                    *b -= aj*hij ;
                                }
                                else                  /* keep this variable*/
                                {
                                    next [k] = j ;
                                    k = j ;
                                    if ( t < *lambda )/* j is bound */
                                    {
                                        nbound++ ;
                                        bound_heap [nbound] = j ;
                                        breakpts [j] = t ;
                                    }
                                    else              /* j is free */
                                    {
                                        a2sum += aj*aj/dj ;
                                        nfree++ ;
                                        free_heap [nfree] = j ;
                                        breakpts [j] = s ;
                                    }
                                }
                            }
                        }
                    }
                    else /* d [j] = 0 */
                    {
                        t = breakpts [j] ;
                        if ( *lambda <= t ) /* fix xj */
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = hi [j] ;
                            else             t = lo [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else if ( t < lambdaL )
                        {
                            aj = a [j] ;
                            if ( aj > ZERO ) t = lo [j] ;
                            else             t = hi [j] ;
                            *b -= aj*t ;
                            x [j] = t ;
                        }
                        else
                        {
                            nbound++ ;
                            bound_heap [nbound] = j ;
                            next [k] = j ;
                            k = j ;
                        }
                    }
                }
            }
        }
        next [k] = j ;
        Stats->nfree = nfree ;
        Stats->nbound = nbound ;
        if ( status == NAPHEAP_STATUS_OK ) /* lambda was optimal */
        {
            return ;
        }

        /* build maxheap of free variables */
        napmaxheap_build (free_heap, breakpts, nfree) ;

        /* build maxheap of bound variables */
        napmaxheap_build (bound_heap, breakpts, nbound) ;

        /* move through break points until finding 0 slope */
        if ( nfree > 0 )
        {
            jfree = free_heap [1] ;
            maxfree = breakpts [jfree] ;
        }
        else maxfree = -INF ;
        if ( nbound > 0 )
        {
            jb = bound_heap [1] ;
            maxbound = breakpts [jb] ;
            while ( (maxbound == *lambda) && nbound )
            {
                if ( d [jb] > ZERO )
                {
                    a2sum -= a [jb]*a [jb]/d [jb] ;
                }
                nbound = napmaxheap_delete(bound_heap, breakpts, nbound);
                if ( nbound )
                {
                    jb = bound_heap [1] ;
                    maxbound = breakpts [jb] ;
                }
                else maxbound = -INF ;
            }
        }
        else maxbound = -INF ;

        while ( 1 )
        {
            if ( maxfree > maxbound )
            {
                new_break = maxfree ;
                j = jfree ;
            }
            else
            {
                new_break = maxbound ;
                j = jb ;
            }

            t = *slope - a2sum*(new_break - *lambda) ;
            if ( (t > ZERO) || (new_break <= -INF) ) /* done */
            {
                if ( nfree && !refine ) *lambda += *slope/a2sum ;
                break ;
            }
            *lambda = new_break ;

            dj = d [j] ;
            if ( dj > ZERO ) *slope = t ;
            else
            {
                *slope = t + fabs (a [j])*(hi [j] - lo [j]) ;
                if ( *slope >= ZERO )
                {
                    break ;
                }
            }

            /* update the heaps */
            if ( *lambda == maxfree ) /* only happens when dj > 0 */
            {
                /* delete break point from heap */
                nfree = napmaxheap_delete (free_heap, breakpts, nfree) ;

                /* add break point to list of previously free variables */
                free_heap [nf] = jfree ;
                nf-- ;

                /* update a2sum */
                a2sum -= a [jfree]*a [jfree]/dj ;

                /* get the smallest entry in heap */
                if ( nfree == 0 )
                {
                    maxfree = -INF ;
                    if ( a2sum < ZERO ) a2sum = ZERO ;
                }
                else
                {
                    jfree = free_heap [1] ;
                    maxfree = breakpts [jfree] ;
                }
            }

            else
            {
                /* delete break point from heap */
                nbound = napmaxheap_delete(bound_heap, breakpts, nbound);

                /* add break point to list of previously bound variables */
                bound_heap [nb] = jb ;
                nb-- ;

                if ( dj ) /* only do the following when dj > 0 */
                {
                    aj = a [jb] ;
                    if ( K > 0 )
                    {
                        s = br_lo [jb] ; /* new break point in br_lo */
                    }
                    else                 /* compute break point */
                    {
                        if ( aj > ZERO ) s = (y [j] - hi [j]*dj)/aj ;
                        else             s = (y [j] - lo [j]*dj)/aj ;
                    }

                    /* store new break point */
                    breakpts [jb] = s ;

                    /* add new break point to free heap */
                    nfree = napmaxheap_add (free_heap, breakpts, jb, nfree);

                    /* get the smallest entry in free heap */
                    jfree = free_heap [1] ;
                    maxfree = breakpts [jfree] ;

                    /* update a2sum */
                    a2sum += aj*aj/dj ;
                }

                /* get the largest entry in bound heap */
                if ( nbound == 0 ) maxbound = -INF ;
                else
                {
                    jb = bound_heap [1] ;
                    maxbound = breakpts [jb] ;
                }
            }
        }
        Stats->nbrks = (n-nb) + (n-nf) ;
    }

    if ( refine )
    {
        /* evaluate slope at lambda */
        *slopehi  = ZERO ; /* upper bound on slope at given lambda */
        *slopelo  = ZERO ; /* lower bound on slope at given lambda */
        *slope = ZERO ;    /* slope ignoring any break point at lambda */
        j = -1 ;

        while ( (j = next [j]) < n )
        {
            aj = a [j] ;
            dj = d [j] ;
            if ( dj > ZERO )
            {
                t = (y [j] - *lambda*aj)/dj ;
                if      ( t < lo [j] ) t = lo [j] ;
                else if ( t > hi [j] ) t = hi [j] ;
                *slope += aj*t ;
            }
            else
            {
                t = breakpts [j] ;
                /* update the slope of dual function */
                if ( *lambda < t )
                {
                    if ( aj > ZERO ) *slope += aj*hi [j] ;
                    else             *slope += aj*lo [j] ;
                }
                else if ( *lambda > t )
                {
                    if ( aj > ZERO ) *slope += aj*lo [j] ;
                    else             *slope += aj*hi [j] ;
                }
                else /* lambda is a break point */
                {
                    if ( aj > ZERO )
                    {
                        *slopehi += aj*hi [j] ;
                        *slopelo += aj*lo [j] ;
                    }
                    else
                    {
                        *slopelo += aj*hi [j] ;
                        *slopehi += aj*lo [j] ;
                    }
                }
            }
        }
        *slope -= *b ;

        /* check if lambda is optimal */
        if ( (*slope+*slopelo <= ZERO) && (*slope+*slopehi >= ZERO) )
        {
            return ;
        }
        Nb = nb ;
        Nf = nf ;
        if ( goright )
        {
            if ( *slope + *slopelo >  err )
            {
                *slope += *slopelo ;
                /* lambda is to the left of the true multiplier, and the break
                   points are stored in a heap.  Search over the break points
                   to the right until finding 0 slope */
                if ( nfree > 0 )
                {
                    jfree = free_heap [1] ;
                    minfree = breakpts [jfree] ;
                }
                else minfree = INF ;
                if ( nbound > 0 )
                {
                    jb = bound_heap [1] ;
                    minbound = breakpts [jb] ;
                    while ( (minbound == *lambda) && nbound )
                    {
                        if ( d [jb] > ZERO )
                        {
                            a2sum -= a [jb]*a [jb]/d [jb] ;
                        }
                        nbound = napminheap_delete(bound_heap, breakpts,
                                                   nbound) ;
                        if ( nbound )
                        {
                            jb = bound_heap [1] ;
                            minbound = breakpts [jb] ;
                        }
                        else minbound = INF ;
                    }
                }
                else minbound = INF ;

                while ( 1 )
                {
                    if ( minfree < minbound )
                    {
                        new_break = minfree ;
                        j = jfree ;
                    }
                    else
                    {
                        new_break = minbound ;
                        j = jb ;
                    }

                    t = *slope - a2sum*(new_break - *lambda) ;
                    if ( (t <= ZERO) || (new_break >= INF) ) /* done */
                    {
                        if ( nfree ) *lambda += *slope/a2sum ;
                        break ;
                    }
                    *lambda = new_break ;

                    dj = d [j] ;
                    if ( dj ) *slope = t ;
                    else
                    {
                        *slope = t - fabs (a [j])*(hi [j] - lo [j]) ;
                        if ( *slope <= ZERO )
                        {
                            break ;
                        }
                    }

                    /* update the heaps */
                    if ( *lambda == minfree ) /* only happens when dj > 0 */
                    {
                        /* delete break point from heap */
                        nfree = napminheap_delete (free_heap, breakpts, nfree) ;

                        /* add break point to previously free list */
                        free_heap [nf] = jfree ;
                        nf-- ;

                        /* update a2sum */
                        a2sum -= a [jfree]*a [jfree]/dj ;

                        /* get the smallest entry in heap */
                        if ( nfree == 0 )
                        {
                            minfree = INF ;
                            if ( a2sum < ZERO ) a2sum = ZERO ;
                        }
                        else
                        {
                            jfree = free_heap [1] ;
                            minfree = breakpts [jfree] ;
                        }
                    }
                    else
                    {
                        /* delete break point from heap */
                        nbound  = napminheap_delete(bound_heap, breakpts,
                                                    nbound) ;

                        /* save break point in previously bound list */
                        bound_heap [nb] = jb ;
                        nb-- ;

                        if ( dj ) /* only do the following when dj > 0 */
                        {
                            aj = a [jb] ;
                            if ( K > 0 )
                            {
                                s = br_hi [jb] ;  /* new break point in br_hi */
                            }
                            else                  /* compute break point */
                            {
                                if ( aj > ZERO ) s = (y [j] - lo [j]*dj)/aj ;
                                else             s = (y [j] - hi [j]*dj)/aj ;
                            }

                            breakpts [jb] = s ;
                            nfree = napminheap_add (free_heap, breakpts, jb,
                                                    nfree);

                            /* get the smallest entry in free heap */
                            jfree = free_heap [1] ;
                            minfree = breakpts [jfree] ;

                            /* update a2sum */
                            a2sum += aj*aj/dj ;
                        }

                        /* get the smallest entry in bound heap */
                        if ( nbound == 0 ) minbound = INF ;
                        else
                        {
                            jb = bound_heap [1] ;
                            minbound = breakpts [jb] ;
                        }
                    }
                }
                Stats->nrefine = (Nb - nb) + (Nf - nf) ;
            }
            else if ( *slope + *slopehi < -err )
            {
                *slope += *slopehi ;
                Dfree = 0 ; /* change in number of free variable */
                if ( nb < n )
                {
                    maxbound = *lambda ;
                    while ( (maxbound == *lambda) && (nb < n) )
                    {
                        nb++ ;
                        jb = bound_heap [nb] ;
                        dj = d [jb] ;
                        if ( dj > ZERO )
                        {
                            aj = a [jb] ;
                            if ( K > 0 )
                            {
                                maxbound = br_lo [jb] ;
                            }
                            else
                            {
                                if ( aj > ZERO )
                                {
                                     maxbound = (y [jb]-hi [jb]*dj)/aj ;
                                }
                                else maxbound = (y [jb]-lo [jb]*dj)/aj ;
                            }
                            if ( maxbound == *lambda )
                            {
                                Dfree-- ;
                                a2sum -= aj*aj/dj ;
                            }
                        }
                        else maxbound = breakpts [jb] ;
                    }
                    if ( maxbound == *lambda ) maxbound = -INF ;
                }
                else maxbound = -INF ;

                if ( nf < n )
                {
                    nf++ ;
                    jfree = free_heap [nf] ;
                    maxfree = breakpts [jfree] ;
                }
                else maxfree = -INF ;

                Nb = nb ;
                Nf = nf ;

                while ( (maxfree > -INF) || (maxbound > -INF) )
                {
                    if ( maxfree >= maxbound )
                    {
                        new_break = maxfree ;
                        j = jfree ;
                    }
                    else
                    {
                        new_break = maxbound ;
                        j = jb ;
                    }
                    t = *slope - a2sum*(new_break - *lambda) ;
                    if ( t >= ZERO ) /* done */
                    {
                        break ;
                    }
                    *lambda = new_break ;

                    dj = d [j] ;
                    if ( dj > ZERO )
                    {
                        *slope = t ;
                    }
                    else
                    {
                        *slope = t + fabs (a [j])*(hi [j] - lo [j]) ;
                        if ( *slope >= ZERO )
                        {
                            *lambda = new_break ;
                            *slope = ZERO ;
                            break ;
                        }
                    }

                    if ( *lambda == maxfree )
                    {
                        Dfree++ ;
                        a2sum += a [j]*a [j]/dj ;   /* becomes free */
                        nf++ ;
                        if ( nf <= n )
                        {
                            jfree = free_heap [nf] ;
                            maxfree = breakpts [jfree] ;
                        }
                        else maxfree = -INF ;
                    }
                    else
                    {
                        if ( dj > ZERO )
                        {
                            Dfree-- ;
                            a2sum -= a [j]*a [j]/dj ; /* becomes bound */
                        }
                        nb++ ;
                        if ( nb <= n )
                        {
                            jb = bound_heap [nb] ;
                            if ( d [jb] )
                            {
                                if ( K > 0 )
                                {
                                    maxbound = br_lo [jb] ;
                                }
                                else
                                {
                                    aj = a [jb] ;
                                    if ( aj > ZERO )
                                    {
                                        maxbound = (y [jb] - hi [jb]*d [jb])/aj;
                                    }
                                    else
                                    {
                                        maxbound = (y [jb] - lo [jb]*d [jb])/aj;
                                    }
                                }
                            }
                            else maxbound = breakpts [jb] ;
                        }
                        else maxbound = -INF ;
                    }
                }
                if ( nfree + Dfree ) *lambda += *slope/a2sum ;
                Stats->nrefine = (nb-Nb) + (nf-Nf) ;
            }
        }
        else
        {
            if ( *slope + *slopehi < -err )
            {
                *slope += *slopehi ;
                /* lambda is to the right of the true multiplier, and the break
                   points are stored in a heap.  Search over the break points
                   to the right until finding 0 slope */
                if ( nfree > 0 )
                {
                    jfree = free_heap [1] ;
                    maxfree = breakpts [jfree] ;
                }
                else maxfree = -INF ;
                if ( nbound > 0 )
                {
                    jb = bound_heap [1] ;
                    maxbound = breakpts [jb] ;
                    while ( (maxbound == *lambda) && nbound )
                    {
                        if ( d [jb] > ZERO )
                        {
                            a2sum -= a [jb]*a [jb]/d [jb] ;
                        }
                        nbound = napmaxheap_delete(bound_heap, breakpts,
                                                   nbound);
                        if ( nbound )
                        {
                            jb = bound_heap [1] ;
                            maxbound = breakpts [jb] ;
                        }
                        else maxbound = -INF ;
                    }
                }
                else maxbound = -INF ;

                while ( 1 )
                {
                    if ( maxfree > maxbound )
                    {
                        new_break = maxfree ;
                        j = jfree ;
                    }
                    else
                    {
                        new_break = maxbound ;
                        j = jb ;
                    }

                    t = *slope - a2sum*(new_break - *lambda) ;
                    if ( (t > ZERO) || (new_break <= -INF) ) /* done */
                    {
                        if ( nfree ) *lambda += *slope/a2sum ;
                        break ;
                    }
                    *lambda = new_break ;

                    dj = d [j] ;
                    if ( dj > ZERO ) *slope = t ;
                    else
                    {
                        *slope = t + fabs (a [j])*(hi [j] - lo [j]) ;
                        if ( *slope >= ZERO )
                        {
                            break ;
                        }
                    }

                    /* update the heaps */
                    if ( *lambda == maxfree ) /* only happens when dj > 0 */
                    {
                        /* delete break point from heap */
                        nfree = napmaxheap_delete (free_heap, breakpts, nfree) ;

                        /* add break point to previously free list */
                        free_heap [nf] = jfree ;
                        nf-- ;

                        /* update a2sum */
                        a2sum -= a [jfree]*a [jfree]/dj ;

                        /* get the smallest entry in heap */
                        if ( nfree == 0 )
                        {
                            maxfree = -INF ;
                            if ( a2sum < ZERO ) a2sum = ZERO ;
                        }
                        else
                        {
                            jfree = free_heap [1] ;
                            maxfree = breakpts [jfree] ;
                        }
                    }

                    else
                    {
                        /* delete break point from heap */
                        nbound = napmaxheap_delete(bound_heap, breakpts,
                                                   nbound) ;

                        /* add break point to previously bound list */
                        bound_heap [nb] = jb ;
                        nb-- ;

                        if ( dj ) /* only do the following when dj > 0 */
                        {
                            aj = a [jb] ;
                            if ( K > 0 )
                            {
                                s = br_lo [jb] ; /* new break point in br_lo */
                            }
                            else                 /* compute break point */
                            {
                                if ( aj > ZERO ) s = (y [j] - hi [j]*dj)/aj ;
                                else             s = (y [j] - lo [j]*dj)/aj ;
                            }

                            /* store new break point */
                            breakpts [jb] = s ;

                            /* add new break point to free heap */
                            nfree = napmaxheap_add (free_heap, breakpts, jb,
                                                    nfree);

                            /* get the smallest entry in free heap */
                            jfree = free_heap [1] ;
                            maxfree = breakpts [jfree] ;

                            /* update a2sum */
                            a2sum += aj*aj/dj ;
                        }

                        /* get the largest entry in bound heap */
                        if ( nbound == 0 ) maxbound = -INF ;
                        else
                        {
                            jb = bound_heap [1] ;
                            maxbound = breakpts [jb] ;
                        }
                    }
                }
                Stats->nbrks = (Nb - nb) + (Nf - nf) ;
            }
            else if ( *slope + *slopelo >  err )
            {
                *slope += *slopelo ;
                Dfree = 0 ; /* change in number of free variables */
                if ( nb < n )
                {
                    minbound = *lambda ;
                    while ( (minbound == *lambda) && (nb < n) )
                    {
                        nb++ ;
                        jb = bound_heap [nb] ;
                        dj = d [jb] ;
                        if ( dj > ZERO )
                        {
                            aj = a [jb] ;
                            if ( K > 0 )
                            {
                                minbound = br_hi [jb] ;
                            }
                            else
                            {
                                if ( aj > ZERO )
                                {
                                     minbound = (y [jb]-lo [jb]*dj)/aj ;
                                }
                                else minbound = (y [jb]-hi [jb]*dj)/aj ;
                            }
                            if ( minbound == *lambda )
                            {
                                Dfree-- ;
                                a2sum -= aj*aj/dj ;
                            }
                        }
                        else minbound = breakpts [jb] ;
                    }
                    if ( minbound == *lambda ) minbound = INF ;
                }
                else minbound = INF ;

                if ( nf < n )
                {
                    nf++ ;
                    jfree = free_heap [nf] ;
                    minfree = breakpts [jfree] ;
                }
                else minfree = INF ;

                Nb = nb ;
                Nf = nf ;

                while ( (minfree < INF) || (minbound < INF) )
                {
                    if ( minfree <= minbound )
                    {
                        new_break = minfree ;
                        j = jfree ;
                    }
                    else
                    {
                        new_break = minbound ;
                        j = jb ;
                    }
                    t = *slope - a2sum*(new_break - *lambda) ;
                    if ( t <= ZERO ) /* done */
                    {
                        break ;
                    }
                    *lambda = new_break ;
                    dj = d [j] ;
                    if ( dj > ZERO )
                    {
                        *slope = t ;
                    }
                    else
                    {
                        *slope = t - fabs (a [j])*(hi [j] - lo [j]) ;
                        if ( *slope <= ZERO )
                        {
                            *lambda = new_break ;
                            *slope = ZERO ;
                            break ;
                        }
                    }

                    if ( *lambda == minfree )
                    {
                        Dfree++ ;
                        a2sum += a [j]*a [j]/dj ;     /* becomes free */
                        nf++ ;
                        if ( nf <= n )
                        {
                            jfree = free_heap [nf] ;
                            minfree = breakpts [jfree] ;
                        }
                        else minfree = INF ;
                    }
                    else
                    {
                        if ( dj > ZERO )
                        {
                            Dfree-- ;
                            a2sum -= a [j]*a [j]/dj ; /* becomes bound */
                        }
                        nb++ ;
                        if ( nb <= n )
                        {
                            jb = bound_heap [nb] ;
                            if ( d [jb] )
                            {
                                if ( K > 0 )
                                {
                                    minbound = br_hi [jb] ;
                                }
                                else
                                {
                                    aj = a [jb] ;
                                    if ( aj > ZERO )
                                    {
                                        minbound = (y [jb] - lo [jb]*d [jb])/aj;
                                    }
                                    else
                                    {
                                        minbound = (y [jb] - hi [jb]*d [jb])/aj;
                                    }
                                }
                            }
                            else minbound = breakpts [jb] ;
                        }
                        else minbound = INF ;
                    }
                }
                if ( nfree + Dfree ) *lambda += *slope/a2sum ;
                Stats->nrefine = (nb-Nb) + (nf-Nf) ;
            }
        }
    }
}

/*  =========================================================================
    === napfix0 =============================================================
    =========================================================================
    d is zero, the optimal dual solution lambda has been determined,
    fix all remaining unfixed components of x except those associated
    with any breakpoints coinciding with lambda.
    =========================================================================*/

PRIVATE void napfix0
(
    double        *x, /* size n. solution (output) */
    double const  *a, /* size n, linear constraint vector */
    double const *lo, /* size n, lower bounds on x */
    double const *hi, /* size n, upper bounds on x */
    NAPHEAP_INT    n, /* problem dimension */
    double        *b, /* b - ai*xi for fixed variables */
    double    lambda, /* optimal multiplier */
    NAPHEAP_INT *next,/* size n+1, next[-1..n-1] list of unbound var. */
    double const *breakpts /* break points */
)
{
    double aj, s, t ;
    NAPHEAP_INT j, k ;

    /* set all variables for which the break point != lambda */
    k = -1 ;
    j =  k ;
    while ( (j = next [j]) < n )
    {
        s = breakpts [j] ;
        aj = a [j] ;
        if ( s < lambda )
        {
            if ( aj > ZERO ) t = lo [j] ;
            else             t = hi [j] ;
            x [j] = t ;
            *b -= aj*t ;
        }
        else if ( s > lambda )
        {
            if ( aj > ZERO ) t = hi [j] ;
            else             t = lo [j] ;
            x [j] = t ;
            *b -= aj*t ;
        }
        else /* lambda is a break point, save for later */
        {
            next [k] = j ;
            k = j ;
        }
    }
    next [k] = j ;
}

/*  =========================================================================
    === napfix1 =============================================================
    =========================================================================
    d can have both positive and zero entries, the optimal dual multiplier
    lambda has been determined, fix all remaining unfixed components of
    x except those associated with any breakpoints coinciding with lambda.
    =========================================================================*/

PRIVATE void napfix1
(
    double        *x, /* size n. solution (output) */
    double const  *a, /* size n, linear constraint vector */
    double const  *y, /* size n. linear term in objective function */
    double const  *d, /* size n, diagonal of objective Hessian */
    double const *lo, /* size n, lower bounds on x */
    double const *hi, /* size n, upper bounds on x */
    NAPHEAP_INT    n, /* problem dimension */
    double        *b, /* b - ai*xi for fixed variables */
    double    lambda, /* optimal multiplier */
    NAPHEAP_INT *next,/* size n+1, next[-1..n-1] list of unbound var. */
    double const *breakpts /* break points */
)
{
    double aj, s, t ;
    NAPHEAP_INT j, k ;
    /* set all variables for which the break point != lambda */
    k = -1 ;
    j =  k ;
    while ( (j = next [j]) < n )
    {
        aj = a [j] ;
        if ( d [j] > ZERO )
        {
            t = (y [j] - lambda*aj)/d [j] ;
            if      ( t < lo [j] ) t = lo [j] ;
            else if ( t > hi [j] ) t = hi [j] ;
            *b -= aj*t ;
            x [j] = t ;
        }
        else
        {
            s = breakpts [j] ;
            /* update the slope of dual function */
            if ( lambda < s )
            {
                if ( aj > ZERO ) t = hi [j] ;
                else             t = lo [j] ;
                *b -= aj*t ;
                x [j] = t ;
            }
            else if ( lambda > s )
            {
                if ( aj > ZERO ) t = lo [j] ;
                else             t = hi [j] ;
                *b -= aj*t ;
                x [j] = t ;
            }
            else /* lambda is a break point, save for later */
            {
                next [k] = j ;
                k = j ;
            }
        }
    }
    next [k] = j ;
}

/*  =========================================================================
    === napsolution =========================================================
    =========================================================================
    d is positive and the optimal dual solution lambda has been determined.
    Evaluate the optimal primal solution.
    =========================================================================*/

PRIVATE void napsolution
(
    double       *x,  /* size n. solution (output) */
    double const *y,  /* size n. linear term in objective function */
    double const *d,  /* size n, diagonal of objective Hessian */
    double const *a,  /* size n, linear constraint vector */
    double const *lo, /* size n, lower bounds on x */
    double const *hi, /* size n, upper bounds on x */
    NAPHEAP_INT    n, /* problem dimension */
    double    lambda, /* optimal multiplier */
    NAPHEAP_INT const *next  /* size n+1, next[-1..n-1] list of unbound var. */
)
{
    double t ;
    NAPHEAP_INT j ;
    /* set solution */
    j = -1 ;
    while ( (j = next [j]) < n )
    {
        t = (y [j] - lambda*a [j])/d [j] ;
        if      ( t < lo [j] ) t = lo [j] ;
        else if ( t > hi [j] ) t = hi [j] ;
        x [j] = t ;
    }
}

/*  ==========================================================================
    === napsolution0 =========================================================
    ==========================================================================
    d is either zero or partly zero.  All the components of x are fixed
    except for the components corresponding to break points that are equal
    to lambda. These remaining components of x are determined by solving
    a least squares problem. Let X, A, LO, and HI denote the subvectors
    corresponding to these undetermined components.  We solve a problem
    of the form

                 min X'X subject to A'X = B, LO <= X <= HI

    to obtain the remaining components of x.
    ========================================================================= */

PRIVATE void napsolution0
(
    double        *x,    /* size n. solution (output) */
    double const  *a,    /* size n, linear constraint vector */
    double const *lo,    /* size n, lower bounds on x */
    double const *hi,    /* size n, upper bounds on x */
    NAPHEAP_INT    n,    /* problem dimension */
    double         B,    /* the constraint is A'X = B */
    double *breakpts,    /* size n, breakpts [j] is br_hi[j] or br_lo[j] */
    NAPHEAP_INT *next,   /* size n+1, next[-1..n-1] link list of unbound var */
    NAPHEAP_INT *free_heap,/* size n+1, heap of free variables */
    NAPHEAP_INT *bound_heap/* size n+1, heap of bound variables */
)
{
    NAPHEAP_INT j, jb, jfree, k, nbound, nfree ;
    double a2sum, aj, lambda, slope, s, t, maxbound, maxfree, minbound,
           minfree, new_break ;

    if ( next [-1] == n ) return ; /* no undetermined components */

    /* initialization */
    nbound = 0 ;   /* counts number of bound variables */
    nfree = 0 ;    /* counts number of free variables */
    a2sum = ZERO ; /* sum aj*aj for free variables */
    lambda = ZERO ;/* starting guess for multiplier */
    slope = ZERO ; /* slope initialization */

    /* evaluate the slope at lambda = 0 */
    k = -1 ;
    j =  k ;
    while ( (j = next [j]) < n )
    {
        if ( hi [j] < ZERO )
        {
            slope += a [j]*hi [j] ;
        }
        else if ( lo [j] > ZERO )
        {
            slope += a [j]*lo [j] ;
        }
        /* else min attained at x [j] = 0 */
    }
    slope -= B ;
    k = -1 ;
    j =  k ;
    if ( slope > ZERO ) /* move right */
    {
        /* d is positive and lambda = 0 is to the left of the true multiplier.
           Get ready to search over the break points to the right by
           constructing the heaps for currently bound and free variables,
           and by fixing variables whose break points are to the left. */
        while ( (j = next [j]) < n )
        {
            aj = a [j] ;
            if ( aj > ZERO )
            {
                s = -lo [j]/aj ;     /* hi break point */
                if ( s <= ZERO )     /* j is fixed */
                {
                    x [j] = lo [j] ;
                }
                else
                {
                    t = -hi [j]/aj ; /* lo break point */
                    next [k] = j ;
                    k = j ;
                    if ( t > ZERO )  /* j is bound */
                    {
                        nbound++ ;
                        bound_heap [nbound] = j ;
                        breakpts [j] = t ;
                    }
                    else             /* j is free */
                    {
                        a2sum += aj*aj ;
                        nfree++ ;
                        free_heap [nfree] = j ;
                        breakpts [j] = s ;
                    }
                }
            }
            else /* aj < 0 */
            {
                s = -hi [j]/aj ;     /* hi break point */
                if ( s <= ZERO )     /* j is fixed */
                {
                    x [j] = hi [j] ;
                }
                else
                {
                    t = -lo [j]/aj ; /* lo break point */
                    next [k] = j ;
                    k = j ;
                    if ( t > ZERO )  /* j is bound */
                    {
                        nbound++ ;
                        bound_heap [nbound] = j ;
                        breakpts [j] = t ;
                    }
                    else             /* j is free */
                    {
                        a2sum += aj*aj ;
                        nfree++ ;
                        free_heap [nfree] = j ;
                        breakpts [j] = s ;
                    }
                }
            }
        }
        next [k] = j ;

        /* build heap for free variables */
        napminheap_build (free_heap, breakpts, nfree) ;

        /* build heap for bound variables */
        napminheap_build (bound_heap, breakpts, nbound) ;

        /* d is positive, lambda is to the left of the true multiplier, and
           the break points are stored in a heap.  Search over the break points
           to the right until finding 0 slope */
        if ( nfree > 0 )
        {
            jfree = free_heap [1] ;
            minfree = breakpts [jfree] ;
        }
        else minfree = INF ;

        if ( nbound > 0 )
        {
            jb = bound_heap [1] ;
            minbound = breakpts [jb] ;
        }
        else minbound = INF ;

        while ( 1 )
        {
            new_break = MIN (minfree, minbound) ;
            t = slope - a2sum*(new_break - lambda) ;
            if ( (t <= ZERO) || (new_break >= INF) ) /* done */
            {
                /* When refine is TRUE, better to do last update in
                   refinement phase */
                if ( nfree ) lambda += slope/a2sum ;
                break ;
            }
            lambda = new_break ;
            slope = t ;

            /* update the heaps */
            while ( lambda == minfree )
            {
                /* delete break point from heap */
                nfree = napminheap_delete (free_heap, breakpts, nfree) ;

                /* update a2sum */
                a2sum -= a [jfree]*a [jfree] ;

                /* get the smallest entry in heap */
                if ( nfree == 0 )
                {
                    minfree = INF ;
                    if ( a2sum < ZERO ) a2sum = ZERO ;
                    break ;
                }
                else
                {
                    jfree = free_heap [1] ;
                    minfree = breakpts [jfree] ;
                }
            }

            while ( lambda == minbound )
            {
                /* delete break point from heap */
                nbound  = napminheap_delete(bound_heap, breakpts, nbound);

                /* compute new hi break point */
                aj = a [jb] ;
                if ( aj > ZERO ) s = -lo [jb]/aj ;
                else             s = -hi [jb]/aj ;
                breakpts [jb] = s ;

                /* add new break point to free heap */
                nfree = napminheap_add (free_heap, breakpts, jb, nfree) ;

                /* get the smallest entry in free heap */
                jfree = free_heap [1] ;
                minfree = breakpts [jfree] ;

                /* update a2sum */
                a2sum += aj*aj ;

                /* get the smallest entry in bound heap */
                if ( nbound == 0 ) minbound = INF ;
                else
                {
                    jb = bound_heap [1] ;
                    minbound = breakpts [jb] ;
                }
            }
        }
    }
    else if ( slope < ZERO ) /* move left */
    {
        /* lambda is to the right of the true multiplier. Get ready to
           search over the break points to the left by constructing the heaps
           for currently bound and free variables, and by fixing variables
           whose break points are to the right. */
        while ( (j = next [j]) < n )
        {
            aj = a [j] ;
            if ( aj > ZERO )
            {
                s = -hi [j]/aj ;     /* lo break point */
                if ( s >= ZERO )     /* j is fixed */
                {
                    x [j] = hi [j] ;
                }
                else
                {
                    t = -lo [j]/aj ; /* hi break point */
                    next [k] = j ;
                    k = j ;
                    if ( t < ZERO )  /* j is bound */
                    {
                        nbound++ ;
                        bound_heap [nbound] = j ;
                        breakpts [j] = t ;
                    }
                    else             /* j is free */
                    {
                        a2sum += aj*aj ;
                        nfree++ ;
                        free_heap [nfree] = j ;
                        breakpts [j] = s ;
                    }
                }
            }
            else
            {
                s = -lo [j]/aj ;     /* lo break point */
                if ( s >= ZERO )     /* j is fixed */
                {
                    x [j] = lo [j] ;
                }
                else
                {
                    t = -hi [j]/aj ; /* hi break point */
                    next [k] = j ;
                    k = j ;
                    if ( t < ZERO )  /* j is bound */
                    {
                        nbound++ ;
                        bound_heap [nbound] = j ;
                        breakpts [j] = t ;
                    }
                    else             /* j is free */
                    {
                        a2sum += aj*aj ;
                        nfree++ ;
                        free_heap [nfree] = j ;
                        breakpts [j] = s ;
                    }
                }
            }
        }
        next [k] = j ;

        /* build the free heap */
        napmaxheap_build (free_heap, breakpts, nfree) ;

        /* build the bound heap */
        napmaxheap_build (bound_heap, breakpts, nbound) ;

        /* d is positive, lambda is to the right of the true multiplier, and
           the break points are stored in a heap.  Search over the break points
           to the left until finding 0 slope */
        if ( nfree > 0 )
        {
            jfree = free_heap [1] ;
            maxfree = breakpts [jfree] ;
        }
        else maxfree = -INF ;

        if ( nbound > 0 )
        {
            jb = bound_heap [1] ;
            maxbound = breakpts [jb] ;
        }
        else maxbound = -INF ;

        while ( 1 )
        {
            new_break = MAX (maxfree, maxbound) ;
            t = slope - a2sum*(new_break - lambda) ;
            if ( (t >= ZERO) || (new_break <= -INF) ) /* done */
            {
                if ( nfree ) lambda += slope/a2sum ;
                break ;
            }
            lambda = new_break ;
            slope = t ;

            /* update the heaps */
            while ( lambda == maxfree )
            {
                /* delete break point from heap */
                nfree = napmaxheap_delete (free_heap, breakpts, nfree) ;

                /* update a2sum */
                a2sum -= a [jfree]*a [jfree] ;

                /* get the smallest entry in heap */
                if ( nfree == 0 )
                {
                    maxfree = -INF ;
                    if ( a2sum < ZERO ) a2sum = ZERO ;
                    break ;
                }
                else
                {
                    jfree = free_heap [1] ;
                    maxfree = breakpts [jfree] ;
                }
            }
            while ( lambda == maxbound )
            {
                /* delete break point from heap */
                nbound  = napmaxheap_delete(bound_heap, breakpts, nbound);

                /* compute new lo break point */
                aj = a [jb] ;
                if ( aj > ZERO ) s = -hi [jb]/aj ;
                else             s = -lo [jb]/aj ;
                breakpts [jb] = s ;

                /* add new break point to free heap */
                nfree = napmaxheap_add (free_heap, breakpts, jb, nfree) ;

                /* get the smallest entry in free heap */
                jfree = free_heap [1] ;
                maxfree = breakpts [jfree] ;

                /* update a2sum and evaluate new break point */
                a2sum += aj*aj ;

                /* get the largest entry in bound heap */
                if ( nbound == 0 ) maxbound = -INF ;
                else
                {
                    jb = bound_heap [1] ;
                    maxbound = breakpts [jb] ;
                }
            }
        }
    }

    /* evaluate the least squares solution */
    j = -1 ;
    while ( (j = next [j]) < n )
    {
        t = -lambda*a [j] ;
        if      ( t < lo [j] ) t = lo [j] ;
        else if ( t > hi [j] ) t = hi [j] ;
        x [j] = t ;
    }
}

/* ==========================================================================
   === napminheap_build =====================================================
   ==========================================================================
   build a min heap in heap [1..nheap]
   heap [i] is an index of an element of array X
   ========================================================================== */

PRIVATE void napminheap_build
(
    NAPHEAP_INT *heap,  /* on input, unsorted set of indices.  size nheap+1 */
    double const *X,    /* numbers to sort */
    NAPHEAP_INT nheap   /* number of elements to build into the heap */
)
{
    NAPHEAP_INT p ;

    for (p = nheap/2 ; p >= 1 ; p--)
    {
        napminheapify (heap, X, p, nheap) ;
    }
}

/* ==========================================================================
   === napminheap_delete ====================================================
   ==========================================================================
   delete the top element in a min heap
   ========================================================================== */

PRIVATE NAPHEAP_INT napminheap_delete  /* return new size of heap */
(
    NAPHEAP_INT *heap,  /* indices into X, 1..n on input.  size nheap+1 */
    double const *X,    /* not modified */
    NAPHEAP_INT nheap   /* number of items in heap */
)
{
    if (nheap <= 1)
    {
        return (0) ;
    }

    /* move element from the end of the heap to the top */
    heap [1] = heap [nheap] ;
    nheap-- ;
    napminheapify (heap, X, 1, nheap) ;
    return (nheap) ;
}

/* ========================================================================== */
/* === napminheap_add ======================================================= */
/* ========================================================================== */
/* ========================================================================== */

/* add a new leaf to a min heap */

PRIVATE NAPHEAP_INT napminheap_add      /* returns the new size of the heap */
(
    NAPHEAP_INT *heap,  /* size n, containing indices into X.  size nheap+2 */
    double const *X,    /* not modified */
    NAPHEAP_INT  leaf,  /* the new leaf */
    NAPHEAP_INT nheap   /* number of elements in heap not counting new one */
)
{
    double Xold, Xnew ;
    NAPHEAP_INT i, newindex, old ;

    nheap++ ;
    old = nheap ;
    heap [old] = leaf ;
    Xold = X [leaf] ;
    while ( old > 1 )
    {
        newindex = old/2 ;
        i = heap [newindex] ;
        Xnew = X [i] ;
        if ( Xnew > Xold )
        { /* swap new and old */
            heap [newindex] = leaf ;
            heap [old] = i ;
        }
        else return (nheap) ;
        old = newindex ;
    }
    return (nheap) ;
}

/* ========================================================================== */
/* === napminheapify ======================================================== */
/* ========================================================================== */

/* heapify starting at node p.  On input, the heap at node p satisfies the */
/* heap property, except for heap [p] itself.  On output, the whole heap */
/* satisfies the heap property. */

PRIVATE void napminheapify
(
    NAPHEAP_INT *heap,  /* size nheap+1, containing indices into X */
    double const *X,    /* not modified */
    NAPHEAP_INT     p,  /* start at node p in the heap */
    NAPHEAP_INT nheap   /* heap [1 ... n] is in use */
)
{
    double Xe, Xleft, Xright ;
    NAPHEAP_INT left, right, e, hleft, hright ;

    e = heap [p] ;
    Xe = X [e] ;

    while ( TRUE )
    {
        left = p * 2 ;
        right = left + 1 ;

        if ( right <= nheap )
        {
            hleft  = heap [left] ;
            hright = heap [right] ;
            Xleft  = X [hleft] ;
            Xright = X [hright] ;
            if (Xleft < Xright)
            {
                if (Xe > Xleft)
                {
                    heap [p] = hleft ;
                    p = left ;
                }
                else
                {
                    heap [p] = e ;
                    return ;
                }
            }
            else
            {
                if (Xe > Xright)
                {
                    heap [p] = hright ;
                    p = right ;
                }
                else
                {
                    heap [p] = e ;
                    return ;
                }
            }
        }
        else
        {
            if ( left <= nheap )
            {
                hleft = heap [left] ;
                Xleft = X [hleft] ;
                if (Xe > Xleft)
                {
                    heap [p] = hleft ;
                    p = left ;
                }
            }
            heap [p] = e ;
            return ;
        }
    }
}

/* ========================================================================== */
/* === napmaxheap_build ===================================================== */
/* ========================================================================== */

/* build a max heap in heap [1..nheap] */

PRIVATE void napmaxheap_build
(
    NAPHEAP_INT *heap,  /* on input, unsorted set of elements. size nheap+1 */
    double const *X,    /* not modified */
    NAPHEAP_INT nheap   /* number of elements to build into the heap */
)
{
    NAPHEAP_INT p ;

    for (p = nheap/2 ; p >= 1 ; p--)
    {
        napmaxheapify (heap, X, p, nheap) ;
    }
}

/* ========================================================================== */
/* === napmaxheap_delete ==================================================== */
/* ========================================================================== */

/* delete the top element in a max heap */

PRIVATE NAPHEAP_INT napmaxheap_delete  /* return new size of heap */
(
    NAPHEAP_INT *heap,  /* containing indices into X, size nheap+1 */
    double const *X,    /* not modified */
    NAPHEAP_INT nheap   /* number of items in heap */
)
{
    if (nheap <= 1)
    {
        return (0) ;
    }

    /* move element from the end of the heap to the top */
    heap [1] = heap [nheap] ;
    nheap-- ;
    napmaxheapify (heap, X, 1, nheap) ;
    return (nheap) ;
}

/* ========================================================================== */
/* === napmaxheap_add ======================================================= */
/* ========================================================================== */

/* add a new leaf to a max heap */

PRIVATE NAPHEAP_INT napmaxheap_add      /* returns the new size of the heap */
(
    NAPHEAP_INT *heap,  /* size nheap+2, containing indices into X */
    double const *X,    /* not modified */
    NAPHEAP_INT  leaf,  /* the new leaf */
    NAPHEAP_INT nheap   /* number of elements in heap not counting new one */
)
{
    double Xold, Xnew ;
    NAPHEAP_INT i, newindex, old ;

    nheap++ ;
    old = nheap ;
    heap [old] = leaf ;
    Xold = X [leaf] ;
    while ( old > 1 )
    {
        newindex = old/2 ;
        i = heap [newindex] ;
        Xnew = X [i] ;
        if ( Xnew < Xold )
        { /* swap new and old */
            heap [newindex] = leaf ;
            heap [old] = i ;
        }
        else return (nheap) ;
        old = newindex ;
    }
    return (nheap) ;
}

/* ========================================================================== */
/* === napmaxheapify ======================================================== */
/* ========================================================================== */

/* heapify starting at node p.  On input, the heap at node p satisfies the */
/* heap property, except for heap [p] itself.  On output, the whole heap */
/* satisfies the heap property. */

PRIVATE void napmaxheapify
(
    NAPHEAP_INT *heap,  /* size nheap+1, containing indices into X */
    double const *X,    /* not modified */
    NAPHEAP_INT p,      /* start at node p in the heap */
    NAPHEAP_INT nheap   /* heap [1 ... nheap] is in use */
)
{
    double Xe, Xleft, Xright ;
    NAPHEAP_INT left, right, e, hleft, hright ;

    e = heap [p] ;
    Xe = X [e] ;

    while ( TRUE )
    {
        left = p * 2 ;
        right = left + 1 ;

        if (right <= nheap)
        {
            hleft  = heap [left] ;
            hright = heap [right] ;
            Xleft  = X [hleft] ;
            Xright = X [hright] ;
            if (Xleft > Xright)
            {
                if (Xe < Xleft)
                {
                    heap [p] = hleft ;
                    p = left ;
                }
                else
                {
                    heap [p] = e ;
                    return ;
                }
            }
            else
            {
                if (Xe < Xright)
                {
                    heap [p] = hright ;
                    p = right ;
                }
                else
                {
                    heap [p] = e ;
                    return ;
                }
            }
        }
        else
        {
            if (left <= nheap)
            {
                hleft = heap [left] ;
                Xleft = X [hleft] ;
                if (Xe < Xleft)
                {
                    heap [p] = hleft ;
                    p = left ;
                }
            }
            heap [p] = e ;
            return ;
        }
    }
}

/* ========================================================================== */
/* === nap_infeasible ======================================================= */
/* ========================================================================== */

/* The problem is infeasible; final wrapup */

PRIVATE void nap_infeasible
(
    double        *x,   /* size n. solution (output) */
    double const  *a,   /* size n. linear constraint vector */
    double const *lo,   /* size n. lower bounds for x */
    double const *hi,   /* size n. upper bounds for x */
    NAPHEAP_INT    n,   /* problem dimension */
    NAPHEAP_INT *next,  /* size n+1, next[-1..n-1]: link list of unbound var */
    double minax,
    double maxax,
    double blo,         /* lower bound for a'x */
    double bhi          /* upper bound for a'x */
)
{
    NAPHEAP_INT j ;

    /* return x that minimizes |a'x-b| */
    if ( minax > bhi )
    {
        j = -1 ;
        while ( (j = next [j]) < n )
        {
            if ( a [j] > ZERO ) x [j] = lo [j] ;
            else                x [j] = hi [j] ;
        }
    }
    else if ( maxax < blo )
    {
        j = -1 ;
        while ( (j = next [j]) < n )
        {
            if ( a [j] > ZERO ) x [j] = hi [j] ;
            else                x [j] = lo [j] ;
        }
    }
}

/* ========================================================================== */
/* === nap_work ============================================================= */
/* ========================================================================== */

/* allocate double workspace */

PRIVATE int nap_work
(
    NAPHEAP_INT n,
    double *User_Work,          /* user-provide workspace */
    double **xw2,               /* return malloc'ed space */
    double **ada,               /* the 5 work arrays */
    double **ady,
    double **br_hi,
    double **br_lo,
    double **breakpts
)
{
    double *work ;

    /* determine how many workspaces to allocate */
    int m = (ada != NULL) + (ady != NULL) + (br_hi != NULL) +
            (br_hi != NULL) + (breakpts != NULL) ;

    if ( User_Work == NULL )
    {
        /* n will not be zero at this point */
        (*xw2) = (double *) malloc ((m*n)*sizeof (double)) ;
        if ( (*xw2) == NULL )
        {
            return (NAPHEAP_STATUS_OUT_OF_MEMORY) ;
        }
        work = (*xw2) ;
    }
    else
    {
        /* use the user-provided workspace */
        work = User_Work ;
        (*xw2) = NULL ;
    }

    /* allocate each array */
    if (breakpts != NULL)
    {
        (*breakpts) = work ;    work += n ;
    }
    if (ada != NULL)
    {
        (*ada) = work ;         work += n ;
    }
    if (ady != NULL)
    {
        (*ady) = work ;         work += n ;
    }
    if (br_hi != NULL)
    {
        (*br_hi) = work ;       work += n ;
    }
    if (br_lo != NULL)
    {
        (*br_lo) = work ;       work += n ;
    }

    return (NAPHEAP_STATUS_OK) ;
}
